"""
A small test example for channel flow
o"""

__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-11-16 -- 2011-10-21"
__license__  = "GNU LGPL Version 3.0 or later"

# Model arguments
ModelArguments(simple=True)

# Parameter
ModelParameters(sigma=RangeCheck(2.2e5, ge=0.1, le=1.0e7),
                V=RangeCheck(-82., ge=-150., le=150))

# Declare domains
mesh = Mesh("channel-mesh")
outer = MeshDomain("outer", mesh, 2)
inner  = MeshDomain("inner", mesh, 3)

# Declare a boundary on the mesh with its own parameters, sigma and dx
channel = MeshBoundary("channel", (inner, outer), (20, 10))

# Declare Species
Ca = Species("Ca", 2)
Na = Species("Na", 1)

# Declare domain specific species
ScalarFieldSpecies(outer, Ca, 1800., sigma)
ScalarFieldSpecies(inner, Ca, 10., sigma)

ScalarFieldSpecies(outer, Na, 140000., sigma/2)
ScalarFieldSpecies(inner, Na, 10000., sigma/2)

# Declare the NCX flux on the Boundary

# To get from pA/pF using intrinsic capacitance of 1F/cm^2
CurrentConv = 9.65e5

# Simple or complex NCX
if simple:
   # NCX parameters from:
   # Bondarenko, V. E.; Szigeti, G. P.; Bett, G. C. L.; Kim, S.-J. & Rasmusson, R. L.
   # Computer model of action potential of mouse ventricular myocytes.
   # Am. J. Physiol. Heart Circ. Physiol., 2004, 287, H1378-403
   # --------------------------------------------------------------------------------
   k_NaCa = 292.8
   Km_Na = 87500.
   Km_Ca = 1380.
   ksat = 0.1
   eta = 0.35
   F = 96.5
   R = 8.314
   T = 298.
   
   # The NCX current in pA/pF
   # ------------------------
   I_NCX = k_NaCa*1/(Km_Na**3+outer.Na**3)*1/(Km_Ca+outer.Ca)*\
           1/(1+ksat*exp((eta-1)*V*F/(R*T)))*\
           (inner.Na**3*outer.Ca*exp(eta*V*F/(R*T))-\
            outer.Na**3*inner.Ca*exp((eta-1)*V*F/(R*T)))

else:
    # Li, L.; Niederer, S. A.; Idigo, W.; Zhang, Y. H.; Swietach, P.;
    # Casadei, B. & Smith, N. P.
    # A mathematical model of the murine ventricular myocyte:
    # a data-driven biophysically based approach applied to mice
    # overexpressing the canine NCX isoform.
    # Am J Physiol Heart Circ Physiol, 2010, 299, H1045-H1063
    V_max = 3.939
    Km_Nao = 87500
    Km_Nai = 12300.
    Km_Cao = 1300.
    Km_Cai = 3.6
    ksat = 0.27
    eta = 0.35
    F = 96.5
    R = 8.314
    T = 298.
    
    # The NCX current in pA/pF
    # ------------------------
    I_NCX = V_max*1/(Km_Cao*inner.Na**3 + \
                     Km_Nao**3+inner.Ca + \
                     Km_Nai**3*outer.Ca*(1+inner.Ca/Km_Cai) + \
                     Km_Cai*outer.Na**3*(1+(inner.Na/Km_Nai)**3) + \
                     inner.Na**3*outer.Ca+outer.Na**3*inner.Ca)*\
                     1/(1+ksat*exp((eta-1)*V*F/(R*T)))*\
                     (inner.Na**3*outer.Ca*exp(eta*V*F/(R*T))-\
                      outer.Na**3*inner.Ca*exp((eta-1)*V*F/(R*T)))

# The sign of the NCX will correspond to the sign of the Na flux
Flux("NCX", channel, Na, 3*CurrentConv*I_NCX)
Flux("NCX", channel, Ca, -CurrentConv*I_NCX)
