# A small Markov model for Ca and Mg binding to the same RyR A-site
#
#   Kds varies from study to study but these values comes
#   from [1]. k_off for both Ca and Mg needs to be fitted. There are
#   some hints on the values though... Comming soon...
#
#   The binding of a single Ca and Mg alters the energy level of the open
#   and closed state of a RyR. To preserve energy this need to be reflected
#   in an altered Kd of binding. As the binding rate is diffusion limited,
#   we alter the off rate instead, see [2] and [3] for details on this.
#
#   State scheme:
#   0 - 1 - 2
#  
#   State | Value | Interpretation
#   ------------------------------
#       0 |     1 |       Ca bound
#       1 |     0 |   No ion bound
#       2 |     2 |       Mg bound

Parameters(Kd_Ca=1.2, k_off_Ca=.5, Kd_Mg=60.0, k_off_Mg=0.5,
           delta_E_Ca=0.5, delta_E_Mg=0.0)

Variables("Ca", "Mg")
Variables("RyR", discrete=True)

Rates[1,0] = Ca*k_off_Ca/Kd_Ca
Rates[0,1] = k_off_Ca * Conditional(Eq(RyR, 1), exp(-delta_E_Ca), 1)

Rates[1,2] = Mg*k_off_Mg/Kd_Mg
Rates[2,1] = k_off_Mg * Conditional(Eq(RyR, 1), exp(-delta_E_Mg), 1)

# The value of either Ca or Mg beeing bound needs to be interpreted by the
# receiving Markov model, the RyR in this case
StateValues[0] = 1
StateValues[2] = 2

#  References:
#  [1] Laver, D. R. & Honen, B. N.
#      Luminal Mg$^2+$, a key factor controlling RYR2-mediated Ca$^2+$ release:
#      cytoplasmic and luminal regulation modeled in a tetrameric channel.
#      J Gen Physiol, 2008, 132, 429-446
#
#  [2] Rice, J. J.; Stolovitzky, G.; Tu, Y. & de Tombe, P. P.
#      Ising model of cardiac thin filament activation with nearest-neighbor
#      cooperative interactions.
#      Biophys. J., 2003, 84, 897-909
#
#  [3] Bray, D. & Duke, T.
#      Conformational spread: the propagation of allosteric states in large
#      multiprotein complexes.
#      Annu Rev Biophys Biomol Struct, 2004, 33, 53-73
