"""
A try to actually do the coupling of the RyR and LCC markov model in
the dyadic cleft!
"""

__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-14 -- 2011-10-14"
__license__  = "GNU LGPL Version 3.0 or later"

# Parameter
ModelParameters(DCa=RangeCheck(2.2e5, ge=1.0e4, le=1.0e7))

# Three domains are possible:
# 1) InifiniteDomain : A domain who can only contain constant species
# 2) VolumeDomain : A domain who can include a scalar and constant species
# 3) MeshDomain : A domain who can contain scalar field, scalar and constant species
TT, Cyt, nSR = InfiniteDomains("TT", "Cyt", "nSR")

mesh = Mesh("cleft_mesh_3D_h12.0_r50_res_1")
Dyad = MeshDomain("Dyad", mesh, 2)
jSR  = MeshDomain("jSR", mesh, 3)

# Createing Markov models that can be used to gate GatedBoundaries
LCC_mm = DiscreteMarkovModel("LCC_mm")
RyR_mm = DiscreteMarkovModel("RyR_mm")

# Boundaries are created by connecting two domains. If one of the domain
# are a MeshDomain additional marker information is provided.

# Arbitrary parameters can be declared, here amp, later sigma and dx
LCC         = MeshBoundary("LCC", (TT, Dyad), 10, 1, amp=-0.12)
RyR_channel = MeshBoundary("RyR_channel", (Dyad, jSR), (100, 300), 4, 3,\
                            sigma=DCa/2)
Sarcolemma  = MeshBoundary("Sarcolemma", Dyad, 2)

Connection(LCC_mm, LCC)
Connection(RyR_mm, RyR_channel)

# Several Boundaris of one type can be declared. Here 4 cytosolic RyR receptors
# are declared. The Markers of these boundaries will then be 1000, 1001, 1002, 1003
RyR_rec_cyt = MeshBoundary("RyR_rec_cyt", Dyad, 1000, 16)
RyR_rec_jsr = MeshBoundary("RyR_rec_jsr", jSR, 3000, 16)
SR          = MeshBoundary("SR", (nSR, jSR), 500)
DyadCyt     = MeshBoundary("DyadCyt", (Dyad, Cyt), 3, 5)

# After the infrastucture is in place we declare what species can be used.
# Here the valence of the ion is also provided.
Ca = Species("Ca", 2)
Mg = Species("Mg", 2)

# Arbitrary variables can be declared which can then be described using its own dynamics
V = ValueClamp("V", [0., 10.], [-65.,0.])

# Now domain specific speices are added.
ConstSpecies(TT, Ca, 1000)

ScalarFieldSpecies(Dyad, Ca, 0.1, DCa)
ConstSpecies(Dyad, Mg, 1000)

ScalarFieldSpecies(jSR, Ca, 800, DCa)
ConstSpecies(jSR, Mg, 1000)

ConstSpecies(nSR, Ca, 800)
ConstSpecies(Cyt, Ca, 0.1)

# Fluxes can be added to Boundaries and connects species in two domains.
Flux("LCC", LCC, Ca, LCC.amp)
DiffusionFlux(RyR_channel, Ca, RyR_channel.sigma)
DiffusionFlux(DyadCyt, Ca, DCa)

# Create a submodels to the DiscreteMarkovModels
Ca_Mg_lum_bind = DiscreteMarkovSubModel(RyR_mm, "Ca_Mg_lum_bind", 4)
Ca_Mg_cyt_bind = DiscreteMarkovSubModel(RyR_mm, "Ca_Mg_cyt_bind", 4)

# Wire up the DiscreteMarkovModesl
Connection(jSR.Ca.RyR_rec_jsr, Ca_Mg_lum_bind)
Connection(jSR.Mg, Ca_Mg_lum_bind)
Connection(Dyad.Ca.RyR_rec_cyt, Ca_Mg_cyt_bind)
Connection(Dyad.Mg, Ca_Mg_cyt_bind)
Connection(Dyad.Ca.LCC, LCC_mm)
Connection(V, LCC_mm)
