"""
This module defines the declarative model language which can be used
in a C{.scm} file to declare a L{SubCell} model.

Declaring a model
=================

The language supplied by L{SubCell} define common entities which
describe subcellular diffusion and reaction:

Example:
--------

>>> ModelParameters(sigma=RangeCheck(2., ge=0.1, le=1.0e7))
>>> 
>>> # Declare domains
>>> mesh = Mesh("channel-mesh")
>>> lower = MeshDomain("lower", mesh, 2)
>>> upper  = MeshDomain("upper", mesh, 3)
>>> 
>>> # Declare a boundary on the mesh with its own parameters, sigma and dx
>>> channel = MeshBoundary("channel", (upper, lower), (20, 10), 1, sigma=sigma,
>>>                         dx=RangeCheck(1., gt=0))
>>> 
>>> # Boundary controller
>>> channel_ctrl = ValueClamp("channel_ctrl", [0.,5.,25.], [0.,1.,0.], True)
>>> 
>>> # Declare connection between channel controller and the channel
>>> Connection(channel_ctrl, channel)
>>> 
>>> # Declare Species
>>> Ca = Species("Ca", 2)
>>> 
>>> # Declare domain specific species
>>> ScalarFieldSpecies(lower, Ca, 1., sigma*5)
>>> ScalarFieldSpecies(upper, Ca, 10, sigma)
>>> 
>>> # Declare flux on the Boundary
>>> Flux(channel, Ca, channel.sigma*(upper.Ca-lower.Ca)/channel.dx)
 
SymPy
=====
  
In addition to the included functions and classes are also all
functions from sympy available. These functions can be used to declare
flux and reaction expressions.
"""

__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-05-02 -- 2011-05-03"
__license__  = "GNU LGPL Version 3.0 or later"

import subcell.model.modellanguage
from subcell.model.modellanguage import *

__all__ = []
__all__.extend(subcell.model.modellanguage.__all__)
