__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-22 -- 2012-07-23"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["DolfinBackend"]

# System imports
from dolfin import *
import ufl
import sympy
import numpy as np
import os

# SubCell imports
from subcell.common import *
from subcell.common.logger import *
from subcell.model.boundaries import *
from subcell.model.species import *
from subcell.model.subcellmodel import *
from subcell.model.domains import *
from subcell.model.fluxes import *
from subcell.model.reactions import *
from subcell.model.discretemarkovmodel import DiscreteMarkovModel
from subcell.model.connection import Connection
from subcell.model.variables import *
from subcell.backends.bases import *
from subcell.backends.dolfinmeshdata import *
from subcell.backends.dolfinsolvertools import *
from subcell.backends.dolfinhandlers import *
from subcell.backends.dolfindata import *
from subcell.backends.reactionsolver import *

# Check that dolfin is compiled with Epetra
if not has_linear_algebra_backend("Epetra"):
    error("dolfin need to be compiled with Epetra LA backend")

# DOLFIN parameters
parameters["graph_coloring_library"] = "Zoltan"
parameters["form_compiler"]["representation"] = "quadrature"
parameters["form_compiler"]["optimize"] = True
parameters["form_compiler"]["cpp_optimize"] = True
parameters["linear_algebra_backend"] = "Epetra"

class DolfinBackend(ContinuousBackend):
    "A FEniCS and DOLFIN specific backend to solve the coupled PDEs/ODEs"
    def __init__(self, subcellmodel):

        check_arg(subcellmodel, SubCellModel)

        timer = Timer("Create Dolfin Backend")
        # Init base class
        ContinuousBackend.__init__(self, subcellmodel)

        # Load mesh data
        self.mesh_data = load_and_check_mesh_data(subcellmodel)

        # Store the cell model
        self.subcellmodel = subcellmodel
        
        # ---- Initalize attributes ----

        # Runtime parameters (controlled by timestepper)
        self.runtime_params = ParameterDict(\
            dt=RangeCheck(1.0, gt=0),
            time=0.0,
            dt_expand=RangeCheck(1.2, gt=0),
            initial_steady_state_solve=TypeCheck(False),
            datadir=TypeCheck("data"),
            run_num=RangeCheck(0, ge=0),
            num_runs=RangeCheck(1, ge=0),
            field_plotted=TypeCheck(False),
            tstart=0.0,
            )
        
        # The domain handler. It keeps all the information of boundaries
        self.domain_handler = DomainHandler(subcellmodel, self.mesh_data)
        
        # The connection map
        self.connections = []
        self.constant_connections = []
        
        # Create solver
        self.solver = SubCellNewtonSolver()

        # Extract parameter
        self.params = self._extract_params()
        
    def _extract_params(self):
        "Extracts parameters"

        # Extract plot names
        plot_names = []
        for species_ind in xrange(self.subcellmodel.max_scalar_fields()):

            plot_species = []
            for domain in self.subcellmodel.mesh_domains():

                # Attach the ufl function to the scalar field species
                species = domain.scalar_field_species[species_ind]
                
                if species is not None:
                    # Append species for defining name and type for plotting
                    plot_species.append(species)

            plot_names.append("_".join([str(species).replace(".", "") \
                                        for species in plot_species]))
            
        # Extract plot names from fixed species
        fixed_plot_names = []
        for species in self.subcellmodel.fixed_scalar_field_species:
            fixed_plot_names.append(str(species).replace(".", ""))

        # IO related parameters
        io = dict(show_timings=True,
                  flux_as_integrals=True,
                  collect_domain_conc=True,
                  collect_conc_at_boundaries=True,
                  collect_fluxes=True,
                  conservation=[],
                  )
        
        # Field species dependent IO parameters
        field_species = {}
        
        for io_names in plot_names:
            field_species[io_names] = ParameterDict(\
                gradient=False,
                rescale=False,
                savevtk=False,
                savebinary=False,
                plot=False,
                vmin=RangeCheck(0, ge=0),
                vmax=RangeCheck(10, ge=0))

        # For fixed field species we do not allow plotting of gradient
        for io_names in fixed_plot_names:
            field_species[io_names] = ParameterDict(\
                gradient=ParConstant(False),
                rescale=False,
                savevtk=False,
                savebinary=False,
                plot=False,
                vmin=RangeCheck(0, ge=0),
                vmax=RangeCheck(10, ge=0))
        
        io["field_species"] = ParameterDict(**field_species)

        return ParameterDict(\
            newton_solver=self.solver.params,
            dolfin=to_parameterdict(parameters, ["krylov_solver"]),
            transient=True,
            params_to_numbers=False,
            const_species_to_numbers=False,
            constant_reaction_jacobian=False,
            operator_split_reactions=False,
            higher_order_split=True,
            nested_reaction_newton_solver=True,
            io=ParameterDict(**io))
        
    def reset_before_set_of_runs(self):
        "A function called for each set of runs"
        # Reset thread argument
        num_threads = self.params.dolfin.num_threads
        self.params.dolfin.num_threads = 0

        # Update dolfin parameters
        parameters.update(self.params.dolfin.copy(True))
        self.params.dolfin.num_threads = num_threads
        
        timer = Timer("Reset system")
        self.runtime_params.field_plotted = False
        
        # The function handelr. It keeps all the dolfin Functions and
        # Function spaces
        self.function_handler = FunctionHandler(self.domain_handler,\
                                                self.params)
        
        # Initialize any static electric potential
        self.function_handler.reset_system_static_electric_potential()
        
        # Create all integrals
        self.integral_handler = IntegralHandler(self.function_handler, \
                                    self.params.constant_reaction_jacobian,
                                    self.params.dolfin)
        
        # Create a ScalarDataCollector
        self.scalar_data_collector = ScalarDataCollector(
            self.integral_handler, self.params.io,\
            self.runtime_params)
        
        # Create a FieldDataCollector
        self.field_data_collector = FieldDataCollector(self.function_handler,\
                        self.params.io.field_species, self.runtime_params)
        
        # Create all tensors
        self.tensor_handler = TensorHandler(self.integral_handler)
        
        # Establish connections
        self._establish_connections(self.subcellmodel)
        
        # Create the Nonlinear problem
        self.problem = SubCellProblem(self.tensor_handler)
        
        # Calculate geometry
        self.scalar_data_collector.calculate_geometry()

        # Create a ReactionSolver if using operator splitting
        # NOTE: Use operator_split_reactions as stored in the function_handler
        # NOTE: as the parameter might get altered during construction
        if self.function_handler.operator_split_reactions:
            self.reaction_solver = ReactionSolver(self.function_handler,\
                                                  self.runtime_params,
                                                  self.params.newton_solver)
        else:
            self.reaction_solver = None
        
        # Store the original number of runs
        orig_num_runs = self.runtime_params.num_runs
        if self.runtime_params.num_runs == 0:
           self.runtime_params.num_runs = 1

    def get_geometry(self):
        """
        Return the calculated geometry
        """
        return self.scalar_data_collector.geometry
        
    def finalize(self):
        "Method called at the end of a set of runs"
        # Show times collected by DOLFIN's logging functionality
        if self.params.io.show_timings:
            list_timings()
            
    def reset_before_single_run(self):
        "Initalize the variables with their initial values"
        
        # Reset data collectors and handlers
        self.domain_handler.reset_before_run()
        self.function_handler.reset_before_run()
        self.field_data_collector.reset_before_run()
        self.tensor_handler.reset_before_run()
        
        # If we want to solve the continuous system to an initial steady state
        if self.runtime_params.initial_steady_state_solve:
            
            # Reset solver
            old_type = self.solver.params.type
            old_transient = self.params.transient
            self.solver.params.type = "direct"
            self.params.transient = False
            self.solver.reset()
            self.step(send_values=False)

            self.params.transient = old_transient
            self.solver.params.type = old_type
        
        # Reset solver
        self.solver.reset()
        
        # Collect all scalar values before they are visualized
        self.scalar_data_collector.collect_data(collect_initials=True)
        
        # Plot fields and collect values
        self.handle_io()
            
        # Send the values
        # NOTE: Send constant is only called once per run
        self.send_constant()
        self.send()
        
        # If we plot the first run turn into interactive mode
        if self.runtime_params.run_num == 0 and \
               self.runtime_params.num_runs > 0:
            self.interactive()

    def get_data_after_single_run(self):
        """
        Return the collected data after a single run
        """
        return self.scalar_data_collector.get_data_after_single_run()
        
    def step_back(self, offset):
        "Update the continuos solution with an offset"
        self.function_handler.step_back(offset)
        self.scalar_data_collector.step_back(offset)

    def diffusion_step(self, dt):
        """
        Perform one diffusion step
        """

        # dt variables
        dt_orig = dt
        ellapsed_time = 0.
        dt_expand = self.runtime_params.dt_expand
        
        # Flags
        converged = False
        transient = self.params.transient
        
        # Solution vector
        x = self.function_handler.vectors["u"]
        
        relaxation = 1.0

        while not converged or ellapsed_time<dt_orig:
            # If we over step the dt
            if ellapsed_time + dt>dt_orig:
                dt = dt_orig - ellapsed_time
            self.function_handler.update_u_solutions_before_solve()
            self.problem.set_time(dt, transient)
            num_iterations, converged = self.solver.solve(self.problem, x)
            info("Num iterations: %d", num_iterations)
            if (x<0).any():
                neg_sum = (x<0).sum()
                warning("Warning: Solution negative! Num print < 0:"\
                        " %d, (%.2f%%)"%(neg_sum, 100.*neg_sum/len(x)))
            if not converged:
                if relaxation <= 0.0625:
                    raise RuntimeError, "Newton Solver did not converge"
                
                relaxation /= 2
                self.solver.newton_solver.parameters.relaxation_parameter = \
                                                                    relaxation
                debug("No convergence. Trying relaxations %.3f.", relaxation)
                
                # NOTE: Disabling reduction of time step to improve 
                # NOTE: convergence. Now you need to make sure that the
                # solution converges
                #raise RuntimeError, "Newton Solver did not converge"
                #dt = dt/2
                #info("New dt: %.3f", dt)
            else:
                self.solver.newton_solver.parameters.relaxation_parameter = 1.0
                self.function_handler.update_solutions_after_solve()
                ellapsed_time += dt
                
                # Expand dt
                dt = dt*dt_expand
        
    def step(self, send_values=True):
        "Solve the Continuous reaction diffusion equations one step"
        timer = Timer("Stepping DolfinBackend")

        # FIXME: Find out how to update the time parameter.
        # FIXME: When we solve we solve for t + dt. But al variables do not
        # know it is t + dt
        # Only t....

        # Time variables
        dt = self.runtime_params.dt

        # Update u0 solutions (used in backsteppings)
        self.function_handler.update_u0_solutions_before_solve()
        
        # Update reaction solver if used and using higher order splitting
        if self.reaction_solver:
            self.function_handler.update_fixed_solutions_before_solve()
            if self.params.higher_order_split:

                debug("")
                debug("Half diffusion step")
                self.diffusion_step(dt/2)
                
                debug("")
                debug("Full reaction step")
                self.reaction_solver.step(dt)
                
                debug("")
                debug("Half diffusion step")
                self.diffusion_step(dt/2)
                
            else:
                debug("")
                debug("Full reaction step")
                self.reaction_solver.step(dt)
                
                debug("")
                debug("Full diffusion step")
                self.diffusion_step(dt)
        else:
            self.diffusion_step(dt)
            
        # Collect all scalar values before they are sent to receivers
        self.scalar_data_collector.collect_data()
        
        # Send values to receivers
        if send_values:
            self.send()

    def handle_io(self):
        "Plot any scalar field speices"
        timer = Timer("Handle io")
        self.field_data_collector.output_data()
        self.scalar_data_collector.output_collected_data()
        
    def interactive(self):
        "Call dolfin interactive"
        if self.runtime_params.field_plotted:
            info("") 
            interactive()

    def send(self):
        "Call all registered send functions"
        for con in self.connections:
            con()
        
    def send_constant(self):
        "Call all registered send functions for constant species"
        for con in self.constant_connections:
            con()

    def update_variables(self, variables):
        """
        Update passed list of variables
        """
        for variable in variables:
            assert(isinstance(variable, Variable))
            debug("Update variable: %s", variable)
            variable._func.value = variable.value
    
    def _establish_connections(self, subcellmodel):
        """
        Create a method for each connection and store it in the
        connection array"""

        # Connect the DiscreteMarkovModel to their receivers (GatedBoundaries)
        for connections in \
                subcellmodel.discrete_markov_model_connections.itervalues():
            assert(isinstance(connections, list))
            # First check if there is only one type of DiscreteMarkovModel
            # connection
            if len(connections)==1:
                receiver, sender = connections[0].receiver.receiver,\
                                   connections[0].sender.sender
                assert(receiver.isgated)
                assert(receiver.num == sender.num)
                for i in xrange(sender.num):
                    # Call the connectors with the recievers
                    sender.get_connector(i)(receiver.receiver(i))
            elif len(connections)>1:
                # If not check that the number of DiscreteMarkovModels equals
                # the number of receiving GatedBoundaries
                num = 0
                sender = connections[0].sender.sender
                receivers = []
                for con in connections:
                    if isinstance(con.receiver.receiver, BoundaryBase) and \
                           con.receiver.receiver.isgated:
                        num += con.receiver.receiver.num
                        receivers.extend(con.receiver.receiver.receiver(i)\
                                         for i in xrange(\
                                             con.receiver.receiver.num))
                if num != sender.num:
                    raise RuntimeError, "expected the number of registered "\
                          "'%s' Markovmodels to equal the numbers of "\
                          "receivers."
                
                # Establish connections
                for i, receiver in enumerate(receivers):
                    sender.get_connector(i)(receiver)

                # FIXME: A Terrible HACK which need to be removed!
                for con in connections:
                    if isinstance(con.receiver.receiver, Variable):
                        for i in xrange(num):
                            sender.get_connector(i)\
                                (con.receiver.receiver.receive)

        # We just add the receive function to the list of receivers in the
        # variable
        for con in subcellmodel.variable_connections:
            receiver = con.receiver.receiver
            sender = con.sender.sender
            if isinstance(receiver, DiscreteMarkovModel):
                sender.receivers.extend(\
                    [receiver.get_receiver(con.receiver.variable, i)\
                     for i in xrange(receiver.num)])
            elif isinstance(receiver, BoundaryBase) and receiver.isgated:
                assert(isinstance(sender, DiscreteVariable))
                sender.receivers.extend([receiver.receiver(i)\
                                         for i in xrange(receiver.num)])
                
        # Iterate over the species connections and establish them
        for con in subcellmodel.species_connections:
            senders, receivers = self._get_senders_and_receivers\
                                 (subcellmodel, con)
            
            # If we have an N-N connection
            if len(senders) == len(receivers):
                def function_closure(sender, receiver):
                    def send():
                        value = sender()
                        debug("Sending N-N: %.1f", value)
                        receiver(value)
                    return send
                
                # Append send function
                for sender, receiver in zip(senders, receivers):
                    self.connections.append(function_closure(sender, receiver))
            
            # If we have an 1-N situation
            elif len(senders) == 1:
                def function_closure(sender, receivers):
                    def send():
                        value = sender()
                        for receiver in receivers:
                            debug("Sending 1-N %.1f", value)
                            receiver(value)
                    return send
                
                # If we are sending a constant value
                if senders[0].func_name == "get_constant_value":
                    self.constant_connections.append(\
                        function_closure(senders[0], receivers))
                else:
                    self.connections.append(function_closure(senders[0],\
                                                             receivers))
            else:
                raise ValueError, "we only support 1-N and N-N connections"
        
    def _get_senders_and_receivers(self, subcellmodel, con):
        """
        A help method to retreive the actual sender and receivers
        in a Connection
        """
        assert(isinstance(con, Connection))
        #assert(isinstance(con.sender.sender, DomainSpecificSpecies))
        #assert(isinstance(con.receiver.receiver, DiscreteMarkovModel))
        #assert(hasattr(con.receiver.receiver, "instances"))

        debug("Connecting %s-%s", con.sender.sender, con.receiver.receiver)
        
        # Get the receiver functions from the connection information
        if isinstance(con.receiver.receiver, DiscreteMarkovModel):
            receivers = [receiver for receiver in con.receiver.receiver.\
                         iter_receivers(con.receiver.variable)]
        else:
            receivers = [con.receiver.receiver.receive]

        # Grab the data collectors
        collectors = self.scalar_data_collector.data_collectors
        
        # If a ConstSpecies just return the value of the init parameter
        if isinstance(con.sender.sender, ConstSpecies):
            def function_closure(const_species):
                param = subcellmodel.params.domains[const_species.domain.name]\
                        [str(const_species.species)]
                
                def get_value():
                    debug("Sending %s: %.2f", str(const_species), \
                          param["init"])
                    return param["init"]
                return get_value
            
            return [function_closure(con.sender.sender)], receivers
        
        if isinstance(con.sender.sender, (ScalarSpecies, FieldSpecies)):
            def function_closure(scalar_species):
                domain = str(scalar_species.domain)
                species = str(scalar_species.species)
                if not "domains" in collectors:
                    raise RuntimeError("No domain concentrations available "\
                                       "for establishing connection. Consider"\
                                       " setting 'collect_domain_conc' to"\
                                       " True")
                data = collectors["domains"].data
                def get_value():
                    value = max(0.0, data[domain][species][-1])
                    debug("Sending %s: %.3f", str(scalar_species), value)
                    return value
                return get_value
            
            return [function_closure(con.sender.sender)], receivers
        
        # If we have a ScalarFieldSpecies
        # Initiate the sender list
        senders = []

        def function_closure(boundary, field_species, i):
            domain = str(field_species.domain)
            species = str(field_species.species)
            boundary = str(boundary)
            if not "conc_at_boundary" in collectors:
                raise RuntimeError("No concentration at the boundary "\
                                   "available for establishing connection."\
                                   " Consider setting 'collect_domain_conc'"\
                                   " to True")
            data = collectors["conc_at_boundary"].data
            def get_value():
                value = max(0.0, data[domain][species][boundary][i][-1])
                debug("Sending %s.%s at %s[%d]: %.3f", domain, species, \
                      boundary, i, value)
                return value
            return get_value
        
        
        # Check if we are connecting to a BoundaryValue
        for ind in xrange(con.sender.boundary.num):
                
            # Append send function
            senders.append(function_closure(con.sender.boundary, \
                                            con.sender.sender, ind))
        
        # Check that we have a N-N connection
        assert(len(senders)==len(receivers))
        return senders, receivers
        
