__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-09-20 -- 2011-10-21"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["VariableCollector", "ScalarConcentrationCollector", \
           "TimeCollector", "ScalarConcentrationAtBoundaryCollector", \
           "ScalarFluxAtBoundaryCollector"]

# System imports
import numpy as np
import sympy
import ufl
from dolfin import dx, ds, assemble

# SubCell imports
from subcell.common import odict, scalar
from subcell.common.logger import *
from subcell.model.species import *
from subcell.model.boundaries import BoundaryValue
from subcell.backends.dolfinhandlers import Constant

class VariableCollector(object):
    def __init__(self, variables, not_include=None):
        """
        Initialize data structures
        """
        not_include = not_include or []
        self.variables = [var for var in variables \
                          if var.name not in not_include]

        # Data structures for the variables
        self.data = {}
        for var in self.variables:
            self.data[str(var)] = []

    def collect_data(self, collect_initials):
        """
        Iterate over the variables and collect the data
        """
        for var in self.variables:
            self.data[str(var)].append(var.value)

    def output_collected_data(self):
        """Dummy function as this is not outputed, yet..."""
        pass
    
    def return_and_reset_data(self):
        """
        Collect and reset data for the variables
        """
        return_data = {}
        
        # Collect and reset data for continuous variables
        for var, values in self.data.items():
            return_data[var] = np.array(values)
            self.data[var] = []
        
        return return_data
        
    def step_back(self, offset):
        """
        Step back collected values
        """
        for var, values in self.data.items():
            du = self.data[var][-1] -\
                 self.data[var][-2]
            self.data[var][-1] -= du*offset
        
class ScalarConcentrationCollector(object):
    def __init__(self, scalar_data_collector, conservation_logging):
        """
        Initalize structures for logging scalar domain concentrations
        """
        self.conservation_logging = conservation_logging
        self.subcellmodel = scalar_data_collector.subcellmodel
        self.domain_handler = scalar_data_collector.domain_handler
        self.function_handler = scalar_data_collector.function_handler
        self.assemble = scalar_data_collector.integral_handler.assemble
        self.init_data_structures()
        self.init_collect_functions()

    def init_data_structures(self):
        """
        Init data structures for collection
        """
        
        data = {}
        
        # Iterate over domains and init data structures
        for domain in self.subcellmodel.domains:
            data[str(domain)] = {}
            for species in domain.species.values():
                if isinstance(species, (ScalarSpecies, \
                                        ScalarFieldSpecies,\
                                        FixedScalarFieldSpecies)):
                    data[str(domain)][str(species.species)] = []

        
        
        # Store data
        self.data = data

    def init_collect_functions(self):
        """
        Initialize data collect functions
        """
        
        def domain_conc_closure(conc_form, volume):
            
            def domain_conc():
                # Assemble form and divide by volume
                return self.assemble(conc_form)/volume
            
            # Return domain concentration integrator
            return domain_conc

        def domain_scalar_conc_closure(value_ind):

            def domain_conc():
                # Grab the concentration from the global solution vector
                # FIXME: Make work in parallel
                # FIXME: It should just work when spaces of Reals are supported
                return self.function_handler.vectors["u"][value_ind]

            # Return domain concentration grabber
            return domain_conc
            
        # Create functions for domain concentration
        domain_funcs =  odict()
        domains = self.subcellmodel.mesh_domains() + \
                  self.subcellmodel.volume_domains()
        
        # Iterate over volume and mesh domains and create collect functions
        for domain in sorted(domains, lambda x, y: cmp(str(x), str(y))):
            
            domain_funcs[domain] = odict()
            
            # Iterate over the scalar field species
            for species in sorted(domain.iter_scalar_field_species(), \
                                  lambda x, y: cmp(str(x.species), \
                                                   str(y.species))):
                u = species.ufl_func["u"]
                
                # Check for a static electric potential
                if species.ufl_func["beta"] != 1:
                    debug("Adding beta func: %s", species.ufl_func["beta"])
                    u = species.ufl_func["beta"]*u

                conc_form = u*dx(domain.marker)
                debug("Conc form for %s: %s", species, conc_form)
                domain_funcs[domain][species.species] = \
                                domain_conc_closure(conc_form,\
                                                    domain.volume.get())
            
            # Iterate over the fixed scalar field species
            for species in sorted(domain.fixed_scalar_field_species, \
                                  lambda x, y: cmp(str(x.species), \
                                                   str(y.species))):
                u = species.ufl_func["u"]
                
                domain_funcs[domain][species.species] = \
                            domain_conc_closure(u*dx(domain.marker),\
                                                domain.volume.get())
            # Iterate over the scalar species
            for species in sorted(domain.scalar_species, \
                                  lambda x, y: cmp(str(x.species), \
                                                   str(y.species))):
                domain_funcs[domain][species.species] = \
                                domain_scalar_conc_closure(species.value_ind)

        # Store functions
        self.domain_funcs = domain_funcs

        # Calculate the maximal str output length
        max_length = 0
        for domain in self.domain_funcs.keys():
            for species in self.domain_funcs[domain]:
                max_length = max(max_length, len("%s.%s: "%(domain, species)))
                
        self.output_max_length = max_length
        

    def collect_data(self, collect_initials):
        """
        Iterate over the collect functions and collects the data
        """
        for domain in self.domain_funcs:
            for species in self.domain_funcs[domain]:
                self.data[str(domain)][str(species)].append(\
                    self.domain_funcs[domain][species]())
                
    def output_collected_data(self):
        """
        Ouput the collected domain concentrations
        """
        begin_log("Concentrations:")
        conserve = {}.fromkeys(self.conservation_logging, 0)
        for domain in self.domain_funcs.keys():
            for species in self.domain_funcs[domain]:
                conc = self.data[str(domain)][str(species)][-1]
                number = conc*domain.volume.get()/1e24*6e23*1e-6
                value_info = "%s.%s: "%(domain, species)
                info("%s%4.2f \xc2\xb5M (%4.2e#)",\
                     value_info.ljust(self.output_max_length), conc, number)
                for con_species in conserve:
                    if species in self.subcellmodel.containing_species\
                           [con_species]:
                        conserve[con_species] += number
        end_log()
        
        if self.conservation_logging:
            begin_log("Conservation:")
            for species, value in conserve.iteritems():
                info("%s: %4.2e#", species, value)
            end_log()
        
    def return_and_reset_data(self):
        """
        Collect and reset data for domain concentrations
        """
        return_data = {}
        for domain in self.data:
            return_data[domain] = {}
            for species in self.data[domain]:
                return_data[domain][species] = np.array(\
                    self.data[domain][species])
                self.data[domain][species] = []
        
        return return_data
        
    def step_back(self, offset):
        """
        Step back collected values
        """
        for domain in self.data.keys():
            for species in self.data[domain].keys():
                du = self.data[domain][species][-1] - \
                     self.data[domain][species][-2]
                self.data[domain][species][-1] -= du*offset
        

class TimeCollector(object):
    def __init__(self, runtime_params):
        """
        Data structures for the time variables
        """
        self.data = dict(times=[], dt=[])
        self.runtime_params = runtime_params

    def collect_data(self, collect_initials):
        """
        Iterate over the collect functions and collects the data
        """
        # Collect time data
        if collect_initials:
            self.data["times"].append(self.runtime_params.time) 
        else:
            self.data["times"].append(self.runtime_params.time + \
                                               self.runtime_params.dt)
            self.data["dt"].append(self.runtime_params.dt)

    def output_collected_data(self):
        pass
        
    def return_and_reset_data(self):
        """
        Collect and reset data for time values
        """
        return_data = {}
        return_data["times"] = np.array(self.data["times"])
        return_data["dt"] = np.array(self.data["dt"])

        self.data["times"] = []
        self.data["dt"] = []

        return return_data
        
    def step_back(self, offset):
        """
        Step back collected time values
        """

        # Offset time data
        old_dt = self.data["dt"][-1]
        self.data["dt"][-1] -= old_dt*offset
        self.data["times"][-1] -= old_dt*offset

class ScalarConcentrationAtBoundaryCollector(object):
    def __init__(self, scalar_data_collector):
        """
        Initalize structures for logging scalar domain concentrations
        """
        self.subcellmodel = scalar_data_collector.subcellmodel
        self.domain_handler = scalar_data_collector.domain_handler
        self.function_handler = scalar_data_collector.function_handler
        self.init_data_structures()
        self.init_collect_functions()

    def init_data_structures(self):
        """
        Init data structures for collection
        """
        
        data = {}
        for domain in self.subcellmodel.mesh_domains():
            data[str(domain)] = {}
            for species in domain.species.values():
                if not isinstance(species, FieldSpecies):
                    continue
                data[str(domain)][str(species.species)] = {}
                
                for boundary in domain.boundaries:
                    data[str(domain)][str(species.species)][str(boundary)] = \
                        [[] for j in xrange(boundary.num)]
        
        # Store data
        self.data = data

    def init_collect_functions(self):
        """
        Initialize data collect functions
        """
        
        # Grab usefull objects from domain_handler
        facet_domains = self.domain_handler.boundary_value_domains
        face_domain_indices = self.domain_handler.face_domain_indices
        mesh = self.domain_handler.mesh
        areas = self.domain_handler.areas
        
        def conc_at_boundaries_closure(boundary_form, m):
            
            def conc_at_boundaries():
                # Reset all facet values
                facet_domains.set_all(1)
                
                # Set faces at the given boundary to 0
                facet_domains.array()[face_domain_indices[m]] = 0
                
                # Assemble form and divide by the area
                return assemble(boundary_form,\
                                exterior_facet_domains=facet_domains,\
                                mesh=mesh)/areas[m]
            
            # Return value at boundary integrator
            return conc_at_boundaries
        
        # Collect functions
        conc_at_boundaries_funcs = odict()
            
        for value in self.subcellmodel.all_values_at_boundaries:

            # Grab field species
            species = value.domain.species[value.species]

            # Grab ufl function
            u = species.ufl_func["u"]
            
            # Check for a static electric potential
            if isinstance(species, ScalarFieldSpecies) and \
                   species.ufl_func["beta"] != 1:
                debug("Adding beta func: %s", species.ufl_func["beta"])
                u = species.ufl_func["beta"]*u
                
            # Adding an integral over subdomain 0, then update a
            # FacetFunction to integrate values at different boundaries
            conc_at_boundaries_funcs[value] = conc_at_boundaries_closure(u*ds,\
                                                                  value.marker)

        # Store collect functions
        self.conc_at_boundaries_funcs = conc_at_boundaries_funcs

        # Calculate the maximal str output length
        max_length = 0
        for value in self.conc_at_boundaries_funcs.keys():
            max_length = max(max_length, len("%s.%s at %s[%d]: "%\
                                             (value.domain, value.species, \
                                              value.boundary, value.ind)))

        self.output_max_length = max_length

    def collect_data(self, collect_initials):
        """
        Iterate over the collect functions and collects the data
        """
        
        for value, func in self.conc_at_boundaries_funcs.iteritems():
            self.data[str(value.domain)][str(value.species)]\
                                [str(value.boundary)][value.ind].append(func())
        
    def output_collected_data(self):
        """
        Output the collected domain concentrations
        """
        begin_log("Concentration at boundaries:")
        for value in self.conc_at_boundaries_funcs.keys():
            val = self.data[str(value.domain)][str(value.species)]\
                  [str(value.boundary)][value.ind][-1]
            value_info = "%s.%s at %s[%d]: "%(value.domain, value.species, \
                                              value.boundary, value.ind)
            info("%s%4.2f \xc2\xb5M", value_info.ljust(self.output_max_length),
                 val)
        end_log()
        
    def return_and_reset_data(self):
        """
        Collect and reset data for values at mesh boundaries
        """
        return_data = {}
        for domain in self.data.keys():
            return_data[domain] = {}
            for species in self.data[domain].keys():
                return_data[domain][species] = {}
                for boundary, values in self.data[domain][species].items():
                    return_data[domain][species][boundary] = \
                                [np.array(value) for value in values]
                    self.data[domain][species][boundary] = \
                                    [[] for j in xrange(len(values))]
        
        return return_data
        
    def step_back(self, offset):
        """
        Step back collected values
        """
        for domain in self.data.keys():
            for species in self.data[domain].keys():
                for boundary in self.data[domain][species].keys():
                    for i, values in enumerate(self.data[domain][species]\
                                               [boundary]):
                        du = values[-1]-values[-2]
                        self.data[domain][species][boundary][i][-1]-= du*offset

class ScalarFluxAtBoundaryCollector(ScalarConcentrationAtBoundaryCollector):
    
    def init_data_structures(self):
        """
        Init data structures for flux collection
        """
        data = {}
        for flux in self.subcellmodel.fluxes:
            if str(flux.boundary) not in data:
                data[str(flux.boundary)] = {}
            if str(flux.species) not in data[str(flux.boundary)]:
                data[str(flux.boundary)][str(flux.species)] = {}
            if flux.fluxname not in data[str(flux.boundary)]\
                   [str(flux.species)]:
                data[str(flux.boundary)][str(flux.species)][flux.fluxname] =\
                                    [[] for j in xrange(flux.boundary.num)]
        # Store data
        self.data = data

    def init_collect_functions(self):
        """
        Initialize data collect functions
        """
        
        # Grab usefull objects from domain_handler
        facet_domains = self.domain_handler.boundary_value_domains
        face_domain_indices = self.domain_handler.face_domain_indices
        mesh = self.domain_handler.mesh
        areas = self.domain_handler.areas
        
        # Create a dict of flux functions
        flux_funcs = {}
        for flux in self.subcellmodel.fluxes:
            debug("")
            debug("Creating function for %s flux %s at %s boundary", \
                  flux.species, flux.fluxname, flux.boundary)
            
            if str(flux.boundary) not in flux_funcs:
                flux_funcs[str(flux.boundary)] = {}
            if str(flux.species) not in flux_funcs[str(flux.boundary)]:
                flux_funcs[str(flux.boundary)][str(flux.species)] = {}

            # Flux functions are handled similar if it is a flux between
            # two Volume domains and if it is a flux between two meshdomains.
            # As the final flux will be a scalar function evaluation.
            # No assemble needed.
            if flux.num_scalar_field_species_domains() == 0 or \
                   flux.num_scalar_field_species_domains() == 2:
                funcs = self.create_scalar_flux_functions(flux)
            else:
                assert(flux.num_scalar_field_species_domains() == 1)
                funcs = self.create_flux_from_integrated_values(flux)

            flux_funcs[str(flux.boundary)][str(flux.species)][flux.fluxname]\
                                                                    = funcs
        
        self.flux_funcs = flux_funcs

        # Calculate the maximal str output length
        max_length = 0
        for flux in self.subcellmodel.fluxes:
            species = flux.species
            boundary = flux.boundary
            for i in xrange(flux.boundary.num):
                value_info = "%s: %s flux at %s[%d]: "%(flux.fluxname,\
                                                        species,\
                                                        boundary, i)
                max_length = max(max_length, len(value_info))
        
        self.output_max_length = max_length

    def create_scalar_flux_functions(self, flux):
        """
        Create a function which calculates the flux between two
        ScalarFieldSpecies or two ScalarSpecies
        """
        
        def scalar_value_closure(value):
            """
            Function closure to return the value of a ScalarSpecies
            or a BoundaryValue
            """
            assert(isinstance(value, (ScalarSpecies, BoundaryValue)))
            def scalar_value():
                return self.function_handler.vectors["u"][value.value_ind]
            return scalar_value

        def variable_value_closure(variable):
            """
            Function closure to return the value of a variable
            """
            def variable_value():
                return variable.value
            return variable_value

        # Create the namespace for the species defined in the flux
        flux_ns, subs = self.function_handler.create_flow_namespace(
            flux, print_debug=False, non_vars_to_scalars=True)
        
        # Grab the expression
        expr = flux.expr()

        # Create a dictionary with the str repr of every ScalarSpecies,
        # ScalarFieldSpecies and variable in the flux with the keys,
        # and a function which returns the scalar value of each guy
        scalar_var_functions = dict((str(subs[species]), \
                                     scalar_value_closure(species))\
                                    for species in flux.scalar_species)
        
        scalar_var_functions.update(dict((str(subs[variable]), \
                                          variable_value_closure(variable))\
                                         for variable in flux.variables))
        
        field_var_functions = dict((str(subs[species]),
                                    [scalar_value_closure(value) \
                                     for (sfs, ind), value in \
                                     flux.boundary.boundary_values.iteritems()\
                                     if sfs == species])\
                                   for species in flux.scalar_field_species)

        # Grab the areas of the boundary
        if flux.num_scalar_field_species_domains() == 2:
            def average_area(ind):
                m0 = flux.boundary.markers[0][ind]
                m1 = flux.boundary.markers[0][ind]
                return (self.domain_handler.areas[m0] + \
                        self.domain_handler.areas[m1])/2
            
            areas = [average_area(ind) for ind in xrange(flux.boundary.num)]
        else:
            areas = [flux.boundary.area.get()]
        
        def flux_func_closure(ind):
            def flux_func():
                for scalar_var, func in scalar_var_functions.iteritems():
                    flux_ns[scalar_var] = func()
                
                for field_var, funcs in field_var_functions.iteritems():
                    flux_ns[field_var] = funcs[ind]()

                # Evaluate the expression
                val = eval(flux_str, flux_ns)*areas[ind]

                # If we get a scalar just return it
                if isinstance(val, scalar):
                    return val

                # If it is an ufl expression we need to evaulate it and
                # hope there are not variables still un-defined
                elif isinstance(val, ufl.expr.Expr):
                    return val.evaluate(0, {}, 0, 0)
                
                raise RuntimeError("should not have reach here")
            
            return flux_func
        
        # Get the name of the species in the namespace
        expr = expr.subs(subs) if isinstance(expr, sympy.Basic) \
               else expr

        # Needed for the function closure
        flux_str = str(expr)

        # Debug
        flux_str_eval = eval(flux_str, flux_ns)
        debug("Flux function str: %s"%flux_str_eval)
        
        # Flux integral functions
        return  [flux_func_closure(ind) for ind in xrange(flux.boundary.num)]

    def create_flux_from_integrated_values(self, flux):
        """
        Create a function which calculates the flux where one species
        is a ScalarFieldSpecies
        """

        # Grab usefull objects from domain_handler
        facet_domains = self.domain_handler.boundary_value_domains
        face_domain_indices = self.domain_handler.face_domain_indices
        mesh = self.domain_handler.mesh

        def flux_func_closure(form, m):
            
            def flux_func():
                # Reset all facet values
                facet_domains.set_all(1)
                
                # Set faces at the given boundary to 0
                facet_domains.array()[face_domain_indices[m]] = 0
                
                # Assemble form and divide by the area
                return assemble(form,\
                                exterior_facet_domains=facet_domains,\
                                mesh=mesh)
            
            # Return value at boundary integrator
            return flux_func
        
        # Collect functions
        conc_at_boundaries_funcs = {}
            
        # Create the namespace for the species defined in the flux
        flux_ns, subs = self.function_handler.create_flow_namespace(flux, \
                                                                    False)
        
        # FIXME: Keep the possibility to no wrap with Constant?
        expr = flux.expr()
        expr = expr.subs(subs) if isinstance(expr, sympy.Basic) \
               else expr
        
        flux_str = str(expr)
        flux_str_eval = eval(flux_str, flux_ns)
        
        # NOTE: If flux is a scalar constant we need a hack because 
        # NOTE: 5*ds is not allowed in ufl
        if isinstance(flux_str_eval, scalar):
            form = Constant(flux_str_eval)*ds
        else:
            form = eval("(%s)*ds"%flux_str, flux_ns)
        
        debug("UFL form: %s", form)

        # Flux integral functions
        return  [flux_func_closure(form, marker) \
                 for marker in flux.boundary.markers[0]]
    
    def collect_data(self, collect_initials):
        """
        Iterate over the collect functions and collects the data
        """
        for flux in self.subcellmodel.fluxes:
            for ind, func in enumerate(self.flux_funcs[str(flux.boundary)]\
                                    [str(flux.species)][flux.fluxname]):
                if flux.boundary.isgated and not flux.boundary.active[ind]:
                    value = 0.
                else:
                    value = func()
                
                umoles_per_ms = value/1e24

                self.data[str(flux.boundary)][str(flux.species)]\
                                [flux.fluxname][ind].append(umoles_per_ms)
        

    def output_collected_data(self):
        """
        Ouput the collected fluxes
        """
        begin_log("Fluxes:")

        for flux in sorted(self.subcellmodel.fluxes, \
                           lambda flux0, flux1: cmp(str(flux0), str(flux1))):
            species = flux.species
            boundary = flux.boundary
            for i in xrange(flux.boundary.num):
                umoles_per_ms = self.data[str(boundary)]\
                                [str(species)][flux.fluxname][i][-1]
                numbers_per_ms = umoles_per_ms*6e23*1e-6
                pA = " (%9.2e pA)"%(umoles_per_ms*1e12*\
                                    species.valence*\
                                    0.096*1e3) if species.valence else ""
                value_info = "%s: %s flux at %s[%d]: "%(flux.fluxname, \
                                                        species,\
                                                        boundary, i)
                info("%s%9.2e #/ms%s", value_info.ljust(\
                    self.output_max_length), numbers_per_ms, pA)
        
        end_log()
    
    def return_and_reset_data(self):
        """
        Collect and reset data for values at mesh boundaries
        """
        return_data = {}
        for boundary in self.data:
            return_data[boundary] = {}
            for species in self.data[boundary]:
                return_data[boundary][species] = {}
                for flux in self.data[boundary][species]:
                    all_values = [np.array(values) for values in \
                                  self.data[boundary][species][flux]]
                    return_data[boundary][species][flux] = all_values
                    self.data[boundary][species][flux] = \
                                    [[] for value in xrange(len(all_values))]
        
        return return_data
        
    def step_back(self, offset):
        """
        Step back collected values
        """
        for boundary in self.data:
            for species in self.data[boundary]:
                for flux in self.data[boundary][species]:
                    for i, values in enumerate(self.data[boundary]\
                                               [species][flux]):
                        du = values[-1]-values[-2]
                        self.data[boundary][species][flux][i][-1]-= du*offset
                        
