__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-05-31 -- 2012-07-23"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["ScalarDataCollector", "FieldDataCollector"]

# System imports
from dolfin import (TestFunction, TrialFunction, Function, File, plot, \
                    Constant, inner, grad, dx, ds, assemble, solve, TimeSeries)
import sympy
import os
import copy
import numpy

# SubCell imports
from subcell.common import *
from subcell.common.logger import *
from subcell.model.species import *
from subcell.model.domains import MeshDomain, mesh_domains
from subcell.backends.dolfinversion import dolfin_version
from subcell.backends.dolfinhandlers import *
from subcell.backends.dolfincollectclasses import *

class ScalarDataCollector(object):
    """
    A class that handles all scalar data collection
    
    The object is created for each simulation, and the data structures
    are reseted after each run when the data is collected.
    """
    def __init__(self, integral_handler, io_params, runtime_params):
        check_arg(integral_handler, IntegralHandler, 0)

        self.integral_handler = integral_handler
        self.function_handler = integral_handler.function_handler
        self.subcellmodel = integral_handler.subcellmodel
        self.domain_handler = integral_handler.domain_handler
        self.params = self.default_params()
        self.params.update(io_params)
        
        conservation_logging = []

        # If logging for conservation
        if len(self.params.conservation) > 0:
            for species_name in self.params.conservation:
                if not isinstance(species_name, str):
                    raise TypeError, "expected a str for declaring Species "\
                          "for conservation logging"
                for species in self.subcellmodel.species:
                    if str(species) == species_name:
                        conservation_logging.append(species)
                        break
                if not species in conservation_logging:
                    raise TypeError, "expected the species name declared "\
                          "for conservation to be a registered Species"
            
            # If logging for conservation flag collect_domain_conc to True
            self.params.collect_domain_conc = True
        
        debug("")
        debug("Creating collect functions for scalar entities:")
        add_log_indent()
        
        # Store data collectors
        self.data_collectors = {}

        # Instantiate specific data collectors
        self.data_collectors["times"] = TimeCollector(runtime_params)

        if self.subcellmodel.continuous_variables:
            self.data_collectors["continuous_variables"] = \
                    VariableCollector(self.subcellmodel.continuous_variables)
            
        if self.subcellmodel.discrete_variables:
            self.data_collectors["discrete_variables"] = \
                    VariableCollector(self.subcellmodel.discrete_variables, \
                                      ["dt_max", "dt_expander", "tstop", \
                                       "time", "dt"])
            
        if self.params.collect_domain_conc:
            self.data_collectors["domains"] = \
                    ScalarConcentrationCollector(self, conservation_logging)

        if self.params.collect_conc_at_boundaries:
            self.data_collectors["conc_at_boundary"] = \
                    ScalarConcentrationAtBoundaryCollector(self)
            
        if self.params.collect_scalar_fluxes:
            self.data_collectors["scalar_fluxes"] = \
                    ScalarFluxAtBoundaryCollector(self)

        add_log_indent(-1)
            
    def default_params(self):
        """
        Return a ParameterDict with Default parameters.
        To be called and altered before instantiation
        
        """
        return ParameterDict(
            collect_domain_conc=True,
            collect_conc_at_boundaries=True,
            collect_scalar_fluxes=True,
            collect_grad_at_boundaries=False,
            conservation=[],
            )

    def collect_data(self, collect_initials=False):
        """
        Method that collects data using the the different methods generated
        by the init_collect_functions

        @type collect_initials : bool
        @param collect_initials: If we are collecting inital values.
        
        """
        debug("Collecting scalar data")

        for collector in self.data_collectors.values():
            collector.collect_data(collect_initials)

    def output_collected_data(self):
        """
        Ouput the collected domain concentrations
        """
        for collector in self.data_collectors.values():
            collector.output_collected_data()

    def step_back(self, offset):
        """
        Step back collected values
        """

        for collector in self.data_collectors.values():
            collector.step_back(offset)
                    
    def _flux_func_closure(self, form):
        """
        A helper function for assembling fluxes
        """
        def flux_func():
            return assemble(form, **self.domain_handler.assemble_kwargs())
        return flux_func

    def create_scalar_flux_functions(self, flux):
        """
        Create a function which calculates the flux between two VolumeDomains
        """
        assert(len(flux.scalar_species_domains())>0)
            
        # Find a host Boundary
        marker = self.domain_handler.marker_to_smallest_mesh_boundary
        
        if marker is None:
            raise NotImplementedError, "we need at least one "\
                  "MeshBoundary to be able to integrate a flux "\
                  "on a pure Boundary"
        
        # Fetch items nessesary to assemble flux
        area_scale = flux.boundary.area.get()/\
                     self.domain_handler.areas[marker]
        
        # Create the namespace for the species defined in the flux
        flux_ns, subs = self.function_handler.create_flow_namespace(flux, \
                                                                    False)
        
        # FIXME: Keep the possibility to no wrap with Constant?
        expr = flux.expr()
        expr = expr.subs(subs) if isinstance(expr, sympy.Basic) \
               else expr
        
        flux_str = str(expr)
        flux_str_eval = eval(flux_str, flux_ns)
        
        # NOTE: If flux is a scalar constant we need a hack because 
        # NOTE: 5*ds is not allowed in ufl
        if isinstance(flux_str_eval, scalar):
            form = Constant(area_scale*flux_str_eval)*ds(marker)
        elif self.function_handler.params_to_numbers:
            form = eval("%f*(%s)*ds(%d)"%(area_scale, flux_str,\
                                          marker), flux_ns)
        else:
            flux_ns["area_scale"] = Constant(area_scale)
            form = eval("area_scale*(%s)*ds(%d)"%(flux_str, marker),\
                        flux_ns)("  UFL form: %s", form)

        # Flux integral functions
        return  [self._flux_func_closure(form)]*flux.boundary.num
        
    def create_flux_from_gradient_1_meshdomain(self, flux):
        """
        Create a flux integral based on calculating the gradient when there
        are one mesh_domain
        """
        
        # Flux where one of the domains are a mesh domain
        domain = flux.scalar_field_species_domains()[0]
        species = domain.species[flux.species]

        # FIXME: Make Sigma a 0 order discontinuous lagrange function
        sigma = species.sigma.get() \
                if self.function_handler.params_to_numbers else \
                Constant(species.sigma.get())
        
        # Iterate over the markers and add forms
        forms = [sigma*inner(grad(species.ufl_func.u),\
                             tetrahedron.n)*ds(m) \
                 for m in flux.boundary.markers[0]]

        # NOTE: A negative gradient means that a species is flowing
        # NOTE: out of the domain. If the flux direction is from
        # NOTE: MeshDomain to OtherDomain then we need to reverse
        # NOTE: the sign
        if flux.sign(domain) == "-":
            forms = [-form for form in forms]
        
        for i, form in enumerate(forms):
            debug("  UFL form[%d]: %s", i, form)
        
        # Flux integral functions
        return [self._flux_func_closure(form) for form in forms]

    def create_flux_from_gradient_2_meshdomain(self, flux):
        """
        Create a flux integral based on calculating the gradient when there
        are one mesh_domain
        """
        
        # Flux where both of the domains are mesh domains
        assert(flux.boundary.num_mesh_domains == 2)
        domains = flux.boundary.domains
        all_species = [domain.species[flux.species] for domain in domains]
        
        # FIXME: Make Sigma a 0 order discontinuous lagrange function
        # Iterate over the field species
        sigmas = []
        for species in all_species:
            if species.sigma.is_scalar:
                sigma = species.sigma.get() if self.function_handler.\
                        params_to_numbers else \
                        Constant(species.sigma.get())
            elif species.sigma.is_array:
                sigma = Function(self.function_handler.Vc)
                values = species.sigma.get()
                if not all(values>0):
                    raise ValueError("expected all values of sigma for %s"\
                                     " to be positive"%species)
                sigma.vector()[species.domain.cell_inds] = values
            else:
                raise TypeError("expected sigma for %s to be a scalar or "\
                                "a CellParam"%species)
            sigmas.append(sigma)
            
        # Iterate over the markers and add an average of the boundary
        # integrals
        forms = [-sigmas[0]*inner(grad(all_species[0].ufl_func.u),
                                  tetrahedron.n)/2*ds(m0)+ \
                 sigmas[1]*inner(grad(all_species[1].ufl_func.u),
                                 tetrahedron.n)/2*ds(m1)\
                 for m0, m1 in zip(*flux.boundary.markers)]
        
        for i, form in enumerate(forms):
            debug("  UFL form[%d]: %s", i, form)
        
        return [self._flux_func_closure(form) for form in forms]

        
    def calculate_geometry(self):
        """
        Fetch and print all geometry data
        """
        data = {}
        data["mesh domains"] = {}
        data["volume domains"] = {}
        data["mesh boundaries"] = {}
        data["boundaries"] = {}
        info("")
        begin_log("Domain volumes:")
        for domain in self.subcellmodel.mesh_domains():
            info("%s: %4.2f nm\xc2\xb3", domain.name, domain.volume.get())
            data["mesh domains"][domain.name] = domain.volume.get()
        for domain in self.subcellmodel.volume_domains():
            info("%s: %4.2f nm\xc2\xb3", domain.name, domain.volume.get())
            data["volume domains"][domain.name] = domain.volume.get()
        end_log()
        
        begin_log("Boundary areas:")
        for boundary in self.subcellmodel.iter_mesh_boundaries():
            data["mesh boundaries"][boundary.name] = {}
            data["mesh boundaries"][boundary.name]["domains"] = \
                       [str(domain) for domain in boundary.domains]
            data["mesh boundaries"][boundary.name]["dx"] =boundary.dx.get()
            data["mesh boundaries"][boundary.name]["area"] = {}
            for domain in mesh_domains(boundary.domains):
                data["mesh boundaries"][boundary.name]["area"][domain.name] = \
                           []
                for ind, marker in enumerate(boundary.domain_markers(domain)):
                    area = self.domain_handler.areas[marker]
                    info("%s in %s[%d]: %4.2f nm\xc2\xb2", boundary, \
                         domain, ind, area)
                    data["mesh boundaries"][boundary.name]["area"]\
                               [domain.name].append(area)
        
        for boundary in self.subcellmodel.iter_none_mesh_boundaries():
            data["boundaries"][boundary.name] = {}
            data["boundaries"][boundary.name]["domains"] = \
                       [str(domain) for domain in boundary.domains]
            info("%s: %4.2f nm\xc2\xb2", boundary.name, boundary.area.get())
            data["boundaries"][boundary.name]["area"] = boundary.area.get()
            data["boundaries"][boundary.name]["dx"] = boundary.dx.get()
        end_log()
        
        # Store the geometry data
        self.geometry = data

    def get_data_after_single_run(self):
        """
        Return the collected data from each run and reset the data structure
        """
        return_data = {}
        
        for collection_type, collector in self.data_collectors.iteritems():
            return_data[collection_type] = collector.return_and_reset_data()
        
        return return_data
    
class FieldDataCollector(object):
    """
    A class that handles all field data collection and plotting
    """
    
    def __init__(self, function_handler, params, runtime_params):
        
        self.function_handler = function_handler
        self.domain_handler = function_handler.domain_handler
        self.params = params
        self.runtime_params = runtime_params
        
        # A dict holding vtk files for saving
        self.vtkfiles = odict()
        self.binaryfiles = odict()
        
        self.init_data()

    def reset_before_run(self):
        "Create vtk files for each new run"
        from os.path import isdir, join
        
        for io_names, plot_species, w, u_plot, u, u_grad in \
                self.function_handler.u.io_data:

            if self.params[io_names].savebinary:
                debug("Init file for saving '%s' field to DOLFIN "\
                      "binary formar", io_names)
                
                datadir = join(self.runtime_params.datadir,
                                str(self.runtime_params.run_num))
                if not isdir(datadir):
                    os.makedirs(datadir)
                filename = join(datadir, "%s"%io_names)
                self.binaryfiles[io_names] = TimeSeries(filename, True)
            
            if self.params[io_names].savevtk:
                debug("Init file for saving '%s' field to vtk", io_names)
                datadir = join(self.runtime_params.datadir,
                                str(self.runtime_params.run_num))
                if not isdir(datadir):
                    os.makedirs(datadir)
                filename = join(datadir, "%s.pvd"%io_names)
                self.vtkfiles[io_names] = File(filename, "compressed")
                if self.params[io_names].gradient:
                    debug("Init file for saving '%s' gradient field to vtk",\
                          io_names)
                    filename = join(datadir, "%s_gradient.pvd"%io_names)
                    self.vtkfiles[io_names+"gradient"] = \
                                            File(filename, "compressed")
                
    def init_data(self):
        # Init projection of gradient integrals
        self.gradient_funcs = odict()

        def grad_function_closure(u_grad, A, L_forms):
            
            def grad_function():
                # Project gradient of scalar field
                b = assemble(L_forms, **self.domain_handler.assemble_kwargs())
                solve(A, u_grad.vector(), b, "cg", "default")
            return grad_function
        
        vv = TestFunction(self.function_handler.Vv)
        uv = TrialFunction(self.function_handler.Vv)
        for io_names, plot_species, w, u_plot, u, u_grad in \
                self.function_handler.u.io_data:
            L_forms = []
            a_forms = []
            if not self.params[io_names]["gradient"]:
                continue
            for species in plot_species:
                # FIXME: Change sigma setting to a discontinuous lagrange
                # FIXME: function
                sigma = Constant(species.sigma.get())
                L_forms.append(-sigma*\
                               inner(vv, grad(u))*dx(species.domain.marker))
                a_forms.append(inner(vv,uv)*dx(species.domain.marker))
            info("Assemble initial gradient Matrix")
            A = assemble(add_forms(a_forms), \
                         **self.domain_handler.assemble_kwargs())
            L_forms = add_forms(L_forms)
            
            self.gradient_funcs[io_names] = grad_function_closure(\
                u_grad, A, L_forms)

    def output_data(self):
        """
        Update the field data structures
        """
        for io_names, plot_species, w, u_plot, u, u_grad in \
                self.function_handler.u.io_data:
            
            if not (self.params[io_names].plot or \
                    self.params[io_names].savevtk or \
                    self.params[io_names].savebinary):
                continue

            if not self.function_handler.using_mixed_space or \
                   all(isinstance(species, FixedScalarFieldSpecies)\
                       for species in plot_species):
                # No interpolation needed
                u_plot.vector()[:] = u.vector()
            else:
                # Interpolate the solution to the copied sub function
                u_plot.interpolate(u)
            
            # Create plot kwargs
            species_params = self.params[io_names]
            not_include = ["plot"]
            if dolfin_version > 1.0:
                not_include.extend(["vmax", "vmin", "gradient", "savebinary", \
                                    "savevtk"])
            kwargs = dict((name, species_params[name]) for name in \
                          species_params if name not in not_include and \
                          species_params[name] is not None)
            
            kwargs["rescale"] = species_params.rescale
            kwargs["title"] = io_names
            
            # Weight the plotted field with the weighting field
            x = u_plot.vector()
            x *= w.vector()
            
            if species_params.gradient:
                # Project gradient of scalar field
                self.gradient_funcs[io_names]()
            
            if species_params.plot:
                # Check for non interactivity in environment
                if not "SUBCELL_NOPLOT" in os.environ:
                    
                    # Plot the interpolated solution
                    plot(u_plot, **kwargs)
                    
                    # Flag that we have plotted a field
                    self.runtime_params.field_plotted = True
                    
                    if self.params[io_names].gradient:
                        plot(u_grad, title=io_names+" gradient")

            if species_params.savebinary:
                # Write the interpolated solution to TimeSeries
                debug("Saving '%s' field to dolfin binary format", io_names)
                if self.runtime_params.time == self.runtime_params.tstart:
                    self.binaryfiles[io_names].store(\
                        u_plot.function_space().mesh(),\
                        self.runtime_params.time)
                    
                self.binaryfiles[io_names].store(u_plot.vector(), \
                                                 self.runtime_params.time)
                
            if species_params.savevtk:
                # Write the interpolated solution to file
                self.vtkfiles[io_names] << (u_plot, self.runtime_params.time)
                debug("Saving '%s' field to vtk", io_names)
                
                if species_params.gradient:
                    debug("Saving '%s' gradient field to vtk", io_names)
                    self.vtkfiles[io_names+"gradient"] << \
                                (u_grad, self.runtime_params.time)
        
