__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-05-31 -- 2011-12-09"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["DomainHandler", "FunctionHandler", "IntegralHandler", \
           "TensorHandler", "add_forms"]

# System imports
from dolfin import *
from dolfin import Constant as dolfin_Constant
import ufl
from ufl.algorithms import expand_derivatives
import sympy
import numpy as np

# SubCell imports
from subcell.common import *
from subcell.common.logger import *
from subcell.model.boundaries import *
from subcell.model.species import *
from subcell.model.subcellmodel import *
from subcell.model.domains import *
from subcell.model.fluxes import *
from subcell.model.reactions import *
from subcell.backends.dolfinsolvertools import zero_integral, zero_forms

def Constant(value):
    """
    Overload dolfin.Constant to include common Cell
    """
    return dolfin_Constant(value, cell=tetrahedron)

def add_forms(forms):
    "Handy method to add a bunch of forms together"
    return add_iterable(forms, zero_integral)

# FFC parameters
ffc_parameters = dict(
    quadrature_degree = 2, 
    optimize = True, 
    eliminate_zeros = True, 
    precompute_basis_const = True
    )

class DomainHandler(object):
    """
    A class that handles the different domains in the model
    """
    def __init__(self, subcellmodel, mesh_data):

        # Store subcell model
        self.subcellmodel = subcellmodel
        
        # Get mesh
        self.mesh = mesh_data.mesh

        # Get boundary markers, areas and domain volumes
        self.markers = mesh_data.markers
        self.areas = mesh_data.areas
        self.volumes = mesh_data.volumes

        # A list with the current active boundaries given by its marker
        self.active_boundary_markers = []

        # A list with the current active constant boundaries
        self.active_constant_boundaries = []
        
        # A MeshFunction of active face domains
        self.active_face_domains = FaceFunction('uint', self.mesh)
        self.boundary_value_domains = FaceFunction('uint', self.mesh)
        
        # Initiate all non used faces to 0 => This means no boundaries can
        # have marker 0
        self.active_face_domains.set_all(0)
        self.boundary_value_domains.set_all(0)
        
        # Gather the domains into a dict for easier access during assemble
        self.all_domains = dict(exterior_facet_domains=\
                                self.markers.face.func,
                                cell_domains=self.markers.cell.func)
        
        # Gather the domains with the active_face_domain as the 
        # exterior facet domain for easier access during assemble
        self.all_active_domains = dict(\
            exterior_facet_domains=self.active_face_domains,\
            cell_domains=self.markers.cell.func)
        
        # Get the indices of the boundary domains
        self.face_domain_indices = self.markers.face.inds

        # Init marker to smallest mesh boundary
        self.marker_to_smallest_mesh_boundary = \
                                self.get_marker_to_smallest_mesh_boundary()

        # Init Functionspaces used to define geometric properties of
        # the domains
        self.Vs = FunctionSpace(self.mesh, "CG", 1)
        self.Vc = FunctionSpace(self.mesh, "DG", 0)

        # Initialize the coordinate parameters of each MeshDomain
        self._init_coordinates()
        
        # Initialize the coordinate parameters of each MeshDomain
        self._init_distance_to_boundaries()
        
        # Adjust the length on any array parameters based on the size of
        # the MeshDomains the parameters are defined on
        self._adjust_array_params()
            
    def _init_coordinates(self):

        # Grab the dofmaps
        vert_dofmap = self.Vs.dofmap()
        cell_dofmap = self.Vc.dofmap()

        # Create intermediate structures for tabulating coordinates
        vert_coords = np.zeros((4,3), 'd')
        cell_coord = np.zeros((1,3), 'd')
        
        # Create the coordinate arrays
        self.vertex_coords = np.zeros((self.Vs.dim(), 3), 'd')
        self.cell_coords = np.zeros((self.Vc.dim(), 3), 'd')
        
        # Iterate over all cells and tabulate coordinates
        for cell in cells(self.mesh):
            
            # Grab the cell mapping
            vert_inds = vert_dofmap.cell_dofs(cell.index())
            cell_ind = cell_dofmap.cell_dofs(cell.index())

            # Tabulate and store coordinates 
            vert_dofmap.tabulate_coordinates(cell, vert_coords)
            cell_dofmap.tabulate_coordinates(cell, cell_coord)
            self.vertex_coords[vert_inds, :] = vert_coords
            self.cell_coords[cell_ind[0], :] = cell_coord
            
        # Iterate over the MeshDomains and grab dof coordinates and assign 
        for domain in self.subcellmodel.mesh_domains():
            
            # Collect dofs 
            value_inds = set()
            cell_inds = set()
            
            # Iterate over cells in MeshDomain
            for cell_ind in self.markers.cell.inds[domain.marker]:
                value_inds.update(vert_dofmap.cell_dofs(cell_ind))
                cell_inds.update(cell_dofmap.cell_dofs(cell_ind))
            
            # Collect dof indices for each MeshDomain
            domain.value_inds = np.fromiter(sorted(value_inds), dtype="I")
            domain.cell_inds = np.fromiter(sorted(cell_inds), dtype="I")
            
            debug("Num value inds in %s: %d", domain, len(domain.value_inds))
            debug("Num cell inds in %s: %d", domain, len(domain.cell_inds))
            
            # Collect node coordinates
            for i, coord in enumerate("xyz"):
                param = getattr(domain, "%s_node"%coord)
                if param.used:
                    debug("Using %s NodeCoordinate"%str(param.sym))
                    param.resize(len(domain.value_inds))
                    param.setvalue(self.vertex_coords[domain.value_inds, i])

            # Collect cell coordinates
            for i, coord in enumerate("xyz"):
                param = getattr(domain, "%s_cell"%coord)
                if param.used:
                    debug("Using %s CellCoordinate"%str(param.sym))
                    param.resize(len(domain.cell_inds))
                    param.setvalue(self.cell_coords[domain.cell_inds, i])
    
    def _init_distance_to_boundaries(self):
        "Initialize the distances to the boundaries if any"
        for distance in self.subcellmodel.distance_to_boundaries:

            # If any of the distances have been used
            if distance.node_dist.used or distance.cell_dist.used:
                import os
                cache_file = "%s_%s.xml.gz"%(self.subcellmodel.meshes[0], distance)
                if os.path.isfile(cache_file):
                    debug("Loading %s_%s from file cache", \
                          self.subcellmodel.meshes[0], distance)
                    signed = Function(self.Vs)
                    File(cache_file) >> signed
                else:
                    
                    debug("Calculating %s", distance)
                    # Compute eikonal equation on the mesh, used to set 
                    # the distances to the included boundaries
                    signed = self._signed_distance_function(distance.markers)
                    File(cache_file) << signed
                    
                # FIXME: Does not work in parallel?
                # Get values from Function
                values = signed.vector().array()[distance.domain.value_inds]
                
                # Check for used node distances
                if distance.node_dist.used:

                    # Change size and set values
                    distance.node_dist.resize(len(distance.domain.value_inds))
                    distance.node_dist.setvalue(values)
                
                # Check for used cell distances
                if distance.cell_dist.used:
                    
                    # Project the signed distance function to a center of
                    # each cell
                    signed = project(signed, self.Vc)
                    values = signed.vector().array()[distance.domain.cell_inds]
                    
                    # Change size and set values
                    distance.cell_dist.resize(len(distance.domain.cell_inds))
                    distance.cell_dist.setvalue(values)
    
    def _signed_distance_function(self, markers):
        "Solve the eikonal equation to get the signed distance function"

        # Handle domains
        debug("  Boundary markers: %s", ",".join(str(m) for m in markers))
        faces = FacetFunction("uint", self.mesh, 0)
        for marker in markers:
            assert(marker in self.markers.face.inds)
            faces.array()[self.markers.face.inds[marker]] = 1
        
        # Define initial Poisson problem
        Vs = FunctionSpace(self.mesh, "CG", 1)
        v = TestFunction(Vs)
        u = TrialFunction(Vs)
        y = Function(Vs)
        f = Constant(1.0)

        # Define a DirichletBC 
        bc = DirichletBC(Vs, Constant(0.0), faces, 1)
        F = inner(grad(u), grad(v))*dx - f*v*dx + f*v*ds(1)
        debug("  Solving initial Poisson problem")
        solve(lhs(F)==rhs(F), y, bc, \
              solver_parameters={"linear_solver": "cg",
                                 "preconditioner": "ml_amg"})
        
        # Create a stabilizer
        # FIXME: Some what heuristic choice of this value...
        eps = Constant(self.mesh.hmax()/25.)
        F = sqrt(inner(grad(y), grad(y)))*v*dx - f*v*dx + \
            eps*inner(grad(y), grad(v))*dx
        
        # Solve Nonlinear problem
        debug("  Solving eikonal equation for the signed distance function")
        solve(F==0, y, bc, solver_parameters={
            "linear_solver": "lu",
            #"preconditioner": "ml_amg",
            "newton_solver": {"relative_tolerance":2e-5,
                              "absolute_tolerance":1e-4,
                              "maximum_iterations":30,
                              "error_on_nonconvergence":True}})
        
        # Return the signed distance function
        return y
        
    def _adjust_array_params(self):
        """
        Adjust the size of the any included ArrayParams
        """
        # Resize and set inds to ArrayParams connected to MeshDomains
        for domain in self.subcellmodel.mesh_domains():
            for param in domain.params.values():
                if isinstance(param, (NodeParam, CellParam)) and \
                       not isinstance(param, (NodeCoordinate, CellCoordinate)):
                    param.resize(len(domain.value_inds) \
                                 if isinstance(param, NodeParam)
                                 else len(domain.cell_inds))
        
        for species in self.subcellmodel.field_species:
            for param in species.params.values():
                if isinstance(param, (NodeParam, CellParam)) and \
                       not isinstance(param, (NodeCoordinate, CellCoordinate)):
                    param.resize(len(domain.value_inds) \
                                if isinstance(param, NodeParam)
                                else len(domain.cell_inds))
            
    def get_marker_to_smallest_mesh_boundary(self):
        "Used in the assembly of scalar species related quantities"
        # Calculate the marker to the smallest boundary
        all_boundary_markers = []
        for boundary in self.subcellmodel.iter_none_gated_mesh_boundaries():
            for markers in boundary.markers:
                all_boundary_markers.extend(markers)

        # If we have a mesh boundary
        if not all_boundary_markers:
            return None
        
        # Grab the MeshBoundary with fewest faces
        marker = all_boundary_markers[0]
        if len(all_boundary_markers) > 1:
            face_inds = self.markers.face.inds
            for new_marker in all_boundary_markers[1:]:
                if len(face_inds[new_marker]) < \
                       len(face_inds[marker]):
                    marker = new_marker
        return marker
    
    def assemble_kwargs(self, tensor=None, add_values=False,
                        active_domains=False):
        "Returns a dict used as kwargs for the assemble rutine"
        #tensor = tensor or self.matrices["A"]

        domain_kwargs = (self.all_active_domains if active_domains else \
                         self.all_domains).copy()
        
        return dict(tensor=tensor,
                    reset_sparsity=True if tensor is None else False,\
                    add_values=add_values, mesh=self.mesh,\
                    form_compiler_parameters = ffc_parameters,\
                    **domain_kwargs)
    
    def reset_before_run(self):
        # Do things that is now in _reset_boundary_actions
        # Reset any active face domains
        self.active_face_domains.set_all(0)

        # Reset GatedBoundaries
        for boundary in self.subcellmodel.iter_gated_boundaries():
            boundary.reset_activation()
        
        # Remove any activation marker from a previous run
        while self.active_boundary_markers:
            self.active_boundary_markers.pop()
            
        # Iterate over the non gated boundaries
        for boundary in self.subcellmodel.iter_none_gated_mesh_boundaries():
            for markers in boundary.markers:
                
                # Add any markers to the active boundary marker list
                self.active_boundary_markers.extend(markers)

                # Update the active_face_domains with the marker of non
                # gated boundaries
                for marker in markers:
                    self.active_face_domains.array()\
                        [self.face_domain_indices[marker]] = marker
        
class FunctionHandler(object):
    """
    A class that initialize and holds all data relating to the Functions
    """

    def __init__(self, domain_handler, params):

        # Store subcell model
        self.subcellmodel = domain_handler.subcellmodel

        self.domain_handler = domain_handler

        # Initalize the FunctionSpace and the subcellmodel structures
        self.init_function_space_and_initialize_structures()

        # Initialize variables
        self.init_variables()

        # Params
        self.params_to_numbers = params.params_to_numbers
        self.const_species_to_numbers = params.const_species_to_numbers
        
        if not params.operator_split_reactions and \
               len(self.subcellmodel.fixed_scalar_field_species):
            info("Using FixedScalarFieldSpecies will enforce operator "\
                 "splitting of reactions.")
            self.operator_split_reactions = True
        else:
            self.operator_split_reactions = params.operator_split_reactions
        
        # Initialize the UFL Functions and attach them to the proper
        # subcellmodel object
        self.init_ufl_functions_and_vectors()

    def init_function_space_and_initialize_structures(self):
        """
        1) Create a Mixed FunctionSpace based on the declared species
        2) Calculate areas and volumes for the different subdomains.
        3) Collect indices into the global solution vector for the
           different species.
        """
        subcellmodel = self.subcellmodel
        mesh = self.domain_handler.mesh
        
        # Collecting information usefull to create the MixedSpace
        self.num_mixed_field_spaces = subcellmodel.max_scalar_fields()
        self.num_scalars = len(subcellmodel.scalar_species)
        self.num_boundary_values = len(subcellmodel.boundary_values)
        self.num_fixed_fields = len(subcellmodel.fixed_scalar_field_species)
        
        # A scalar and a vector function space over the mesh
        self.Vs = self.domain_handler.Vs 
        self.Vc = self.domain_handler.Vc 
        self.Vv = VectorFunctionSpace(mesh, "CG", 1)
        Vr = FunctionSpace(mesh, "R", 0)
        
        # The whole Functionspace
        if self.num_mixed_field_spaces + self.num_scalars + \
               self.num_boundary_values>1:
            self.V = MixedFunctionSpace(\
                [self.Vs]*self.num_mixed_field_spaces +
                [Vr]*(self.num_scalars + self.num_boundary_values))
            self.Vcells = MixedFunctionSpace(\
                [self.Vc]*self.num_mixed_field_spaces +
                [Vr]*(self.num_scalars + self.num_boundary_values))
            self.using_mixed_space = True
        else:
            debug("No mixed spaces is used")
            self.V = self.Vs
            self.Vcells = self.Vc
            self.using_mixed_space = False
            
        # Iterate over the mesh domains and collect volumes
        for domain in subcellmodel.mesh_domains():
            getattr(domain, "__volume").setvalue(\
                self.domain_handler.volumes[domain.marker])

        # Add a constant dt Coefficient used
        self.dt = Expression("value", value=1.0, element=Vr.ufl_element())
        
        # Iterate over the scalar_field species and add information about
        # indices in full solution vector
        for species in subcellmodel.scalar_field_species:

            # Grab local species index
            species_ind = species.domain.scalar_field_species.index(species)

            # Set the indices in the global vector for the species
            vert_dofmap = self.V.sub(species_ind).dofmap() if \
                          self.using_mixed_space else self.V.dofmap()

            cell_dofmap = self.Vcells.sub(species_ind).dofmap() if \
                          self.using_mixed_space else self.Vcells.dofmap()
            
            value_inds = set()
            cell_inds = set()
            for cell_ind in self.domain_handler.markers.cell.\
                        inds[species.domain.marker]:
                value_inds.update(vert_dofmap.cell_dofs(cell_ind))
                cell_inds.update(cell_dofmap.cell_dofs(cell_ind))

            # Assign the indices to the values
            species.value_inds = np.fromiter(sorted(value_inds), dtype="I")
            species.cell_inds = np.fromiter(sorted(cell_inds), dtype="I")

            debug("Num dofs '%s': %d", species, len(species.value_inds))
                
            # Prepare a dict for ufl functions
            species.ufl_func = adict()

        # Iterate over the scalar species and collect indices
        for i, species in enumerate(subcellmodel.scalar_species):

            # Collect index in the global solution vector
            subspace = self.num_mixed_field_spaces+i
            species.value_ind = self.V.sub(subspace).dofmap().cell_dofs(0)[0]
        
            # Prepare a dict for ufl functions
            species.ufl_func = adict()
            
        # Iterate over const species and initalize structure
        for species in subcellmodel.const_species:
            
            # Prepare a dict for ufl functions
            species.ufl_func = adict()

        # Iterate over the boundary_values to collect areas and indices
        for i, value in enumerate(subcellmodel.boundary_values):
            
            # Collect areas and the index in the global solution vector 
            value.area = self.domain_handler.areas[value.marker]

            # Collect index in the global solution vector
            subspace = self.num_mixed_field_spaces + self.num_scalars + i
            value.value_ind = self.V.sub(subspace).dofmap().cell_dofs(0)[0]
            
            # Prepare a dict for ufl functions
            value.ufl_func = adict()

        # Iterate over the FixedScalarFieldSpecies and collect indices
        for species in subcellmodel.fixed_scalar_field_species:
                
            # Assign the indices to the values
            species.value_inds = species.domain.value_inds
            
            debug("Num dofs '%s': %d", species, len(species.value_inds))
                
            # Prepare a dict for ufl functions
            # FIXME: Not sure I want this, but include it so the electric
            # FIXME: potential initialization does not bang
            species.ufl_func = adict()

        for domain in subcellmodel.mesh_domains():
            for name, species in domain.species.iteritems():
                # Initialize electric potential scalings
                species.ufl_func["beta"] = 1
                
                if species.domain.static_electric_potential is not None:
                    species.numpy_beta = np.ones(len(\
                        self.domain_handler.vertex_coords))
                    domain.scale[name] = 1
        
        info("")
        fixed_dofs = sum(len(species.value_inds) for species \
                         in subcellmodel.fixed_scalar_field_species)
        sumdofs = self.V.dim() + fixed_dofs
        # FIXME: Add dofs from FixedScalarField species
        begin_log("Number of degrees of freedoms:")
        info("Field species:  %12s", "%d*%d"%(self.num_mixed_field_spaces, \
                                              self.Vs.dim()))
        info("Scalar species: %12d", self.num_scalars)
        info("Fixed species:  %12d", fixed_dofs)
        info("Boundary values:%12d", self.num_boundary_values)
        info("SUM:            %12d", sumdofs)
        end_log()


    def init_ufl_functions_and_vectors(self):
        "Create ufl functions and sub functions for the unknowns"

        # Vector dict for all the sollution vectors
        self.vectors = odict()

        subcellmodel = self.subcellmodel

        # Iterate over the different Function types
        local_inds = range(self.num_mixed_field_spaces) + \
                     range(self.num_scalars) + \
                     range(self.num_boundary_values)

        # Iterate over the different Functions
        for ftype in ["u", "u0", "u1", "v", "du"]:
            f = adict()
            
            # Instantiate the full Function
            f.full = {"v":TestFunction, "du":TrialFunction}.\
                     get(ftype, Function)(self.V)
            
            # Collect information from the different functions
            f.fields = []
            f.scalars = []
            f.boundary_values = odict()
            f.fluxes = odict()
            if ftype == "u":
                f.io_data = []
            
            # If we are dealing with a MixedFunctionSpace iterate over
            # the sub functions.
            if self.using_mixed_space:
                functions = split(f.full)
            else:
                functions = [f.full]
            for i, sub in enumerate(functions):
                if i < self.num_mixed_field_spaces:
                    f.fields.append(sub)
                    plot_species = []
                    for domain in subcellmodel.mesh_domains():
                        
                        # Attach the ufl function to the scalar field species
                        species = domain.scalar_field_species[i]
                        
                        # No species for this index
                        if species is None:
                            continue
                        
                        species.ufl_func[ftype] = sub
                        
                        # Append species for defining name and type for
                        # plotting
                        plot_species.append(species)
                    
                    # Append plotting functions
                    if ftype == "u":
                        
                        # Create a string representation of the speices
                        plot_name = "_".join([str(species).replace(".", "") \
                                              for species in plot_species])
                        
                        # Init plotting weights to 1
                        w = Function(self.Vs)
                        w.vector()[:] = 1
                        
                        # Create a list with plot information which is used
                        # to plot each species
                        if self.using_mixed_space:
                            f.io_data.append(\
                                (plot_name, plot_species, w,\
                                 f.full._sub(i, True), f.full._sub(i),\
                                 Function(self.Vv)))
                        else:
                            f.io_data.append(\
                                (plot_name, plot_species, w,\
                                 Function(self.V), f.full, Function(self.Vv)))
                    
                elif i < self.num_mixed_field_spaces + self.num_scalars:
                    f.scalars.append(sub)
                    
                    # Attach the ufl function to the scalar field species
                    subcellmodel.scalar_species[local_inds[i]].\
                                        ufl_func[ftype] = sub

                else:
                    f.boundary_values[subcellmodel.boundary_values\
                                      [local_inds[i]]] = sub
                    subcellmodel.boundary_values[local_inds[i]].\
                                      ufl_func[ftype] = sub
                    
            # Extract the vector
            if isinstance(f.full, Function):
                self.vectors[ftype] = f.full.vector()

            # Assign the function to self
            exec("self.%s = f"%ftype)

        # Copy a tmp vector
        self.vectors["tmp"] = self.vectors["u"].copy()

        # Init FixedFieldSpecies vectors and Functions
        self.vectors["fixed"] = odict()

        # Create Functions for any FixedSpecies
        # FIXME: Save spaces similare to what we do with ScalarFieldSpecies
        # FIXME: by putting several FixedSpecies into one Function, if that is
        # FIXME: possible
        if subcellmodel.fixed_scalar_field_species:
            
            for species in subcellmodel.fixed_scalar_field_species:
                func = Function(self.Vs)
                species.ufl_func["u"] = func
                self.vectors["fixed"][str(species)] = odict()
                self.vectors["fixed"][str(species)]["u"] = func.vector()
                self.vectors["fixed"][str(species)]["u0"] = \
                                                          func.vector().copy()
                
                plot_species = [species]
                        
                # Create a string representation of the speices
                plot_name = str(species).replace(".", "")
                
                # Init plotting weights to 1
                w = Function(self.Vs)
                w.vector()[:] = 1
                
                # Create a list with plot information which is used
                # to plot each species
                self.u.io_data.append((plot_name, plot_species, w,\
                                       Function(self.Vs), func, None))
            
            # A tmp vector
            self.vectors["fixed"]["tmp"] = func.vector().copy()

    def init_variables(self):
        """
        Initalize variables by adding an Expression to each one of them
        """
        for variable in self.subcellmodel.continuous_variables + \
                self.subcellmodel.discrete_variables:
            variable._func = Expression("value", value=0, cell=tetrahedron)
            
    def reset_system_static_electric_potential(self):
        "Help function for initalize the static electric potential"
        
        # Iterate over the mesh domains
        for domain in self.subcellmodel.mesh_domains():
            
            # If no potential return
            if domain.static_electric_potential is None:
                continue
            
            # Add an ufl expression of beta=exp(-z*psi) for each species
            # on the domain
            for species in domain.iter_scalar_field_species():
                if species.species.valence != 0.0:
                    
                    # Check for incidence domains
                    if self.domain_handler.markers.vertex.cell.\
                           incidence[domain.marker]:
                        for incidence_marker in self.domain_handler.\
                            markers.vertex.cell.incidence[domain.marker]:
                            if incidence_marker in self.subcellmodel.\
                                   marker_to_mesh_domain:
                                other_domain = self.subcellmodel.\
                                    marker_to_mesh_domain[incidence_marker]
                                domain_str = "'%s' shares vertices with '%s'"%\
                                             (str(domain), str(other_domain))
                                raise RuntimeError("Cannot declare an "\
                                    "electric potential on a MeshDomain "\
                                    "which shares vertices with another "\
                                    "MeshDomain. %s."%domain_str)
                    
                    # Create values and indices into these values
                    values = np.ones(len(self.domain_handler.vertex_coords))
                    inds = species.domain.value_inds
                    
                    # Assign the numpy values
                    values[inds] = np.exp(-species.species.valence*\
                                domain.static_electric_potential.expr.get())
                    species.numpy_beta = values
                    
                    species.ufl_func["beta"] = Function(self.Vs)
                    species.ufl_func["beta"].vector()[:] = values

                    # Store the scale factor at domain level
                    domain.scale[species.species] = species.ufl_func["beta"]

        debug("initializing plotting weights")
                    
        # Initialize plotting weights
        # We need to iterate over the plotting data and extract the "beta"
        # Coefficient for each species we plot.
        for i, (io_names, plot_species, w, u_plot, u, u_grad) in\
                enumerate(self.u.io_data):

            numpy_beta = reduce(\
                lambda x, y : x*y,\
                (species.numpy_beta for species in plot_species\
                 if species.domain.static_electric_potential is not None),\
                np.ones(self.domain_handler.mesh.num_vertices()))

            #debug(str(numpy_beta))
            w.vector()[:] = numpy_beta
            #print "Num negative values:", (w.vector().array()<0).sum()
            #print "Num below 1:", (w.vector().array()<1).sum()
            #plot(w, title="Beta weight", interactive=True)
            
    def reset_before_run(self):
        "Reset all values before a run"

        debug("")
        debug("Reseting values before run:")
        add_log_indent()
        # Apply initial values
        for species in self.subcellmodel.scalar_field_species:
            # FIXME: Make work in parallel
            # NOTE:  This should work in parallel :)
            value = species.init.get()
            valuestr = "=%.4f"%value if isinstance(value, float) else ""
            debug("ScalarFieldSpecies: %s%s", str(species), valuestr)
            self.vectors["u"][species.value_inds] = value
        
        for species in self.subcellmodel.fixed_scalar_field_species:
            # FIXME: Make work in parallel
            # NOTE:  This should work in parallel :)
            value = species.init.get()
            valuestr = "=%.4f"%value if isinstance(value, float) else ""
            debug("FixedScalarFieldSpecies: %s%s", str(species), valuestr)
            self.vectors["fixed"][str(species)]["u"][species.value_inds] =value
        
        for species in self.subcellmodel.scalar_species:
            # FIXME: Make work in parallel
            value = species.init.get()
            debug("ScalarSpecies: %s%s", str(species), "=%.4f"%value)
            self.vectors["u"][species.value_ind] = species.init.get()
            
        if self.subcellmodel.boundary_values:
            facet_domains = self.domain_handler.boundary_value_domains
            face_domain_indices = self.domain_handler.face_domain_indices
            assemble_kwargs = self.domain_handler.assemble_kwargs
            mesh = self.domain_handler.mesh
            areas = self.domain_handler.areas
            for value in self.subcellmodel.boundary_values:

                # Get marker
                m = value.marker

                # Reset all facet values
                facet_domains.set_all(1)
                
                # Set faces at the given boundary to 0
                facet_domains.array()[face_domain_indices[m]] = 0
                
                # Get ScalarFieldSpecies
                species = value.domain.species[value.species]
                u = species.ufl_func["u"]
                
                # Check for a static electric potential
                if species.ufl_func["beta"] != 1:
                    debug("Adding beta func: %s", species.ufl_func["beta"])
                    u = species.ufl_func["beta"]*u
                
                # Assemble form and divide by the area
                val = assemble(u*ds,\
                               exterior_facet_domains=facet_domains,\
                               mesh=mesh)/areas[m]
                
                debug("BoundaryValue: %s.%s at %s[%d]=%.4f", 
                      str(value.domain), str(value.species), \
                      str(value.boundary), value.ind, val)
                
                self.vectors["u"][value.value_ind] = val
            
        # Update io function
        self.vectors["u1"][:] = self.vectors["u"]
        debug("")
        add_log_indent(-1)
        
    def sympy_to_ufl_namespaces(self):
        # FIXME: Remove?
        # Populate namesspace with ufl's
        ns = {}
        ns.update(ufl.__dict__)
        
        return ns, {}
    
    def create_flow_namespace(self, flow, print_debug=True, \
                              non_vars_to_scalars=False):
        assert(isinstance(flow, (Reaction, Flux)))
        from subcell.common.logger import debug as _debug
        
        def fake_debug(*args):pass
        debug = _debug if print_debug else fake_debug

        debug("")
        if isinstance(flow, Reaction):
            debug("Creating namespace for reaction on '%s' domain"%flow.domain)
        else:
            debug("Creating namespace for flux: '%s'"%flow.fluxname)

        # Special treatment for Fluxes
        if isinstance(flow, Flux):
            
            debug("robin=%d, constant=%d, var_dep=%d, gated=%d", flow.robin,\
                  flow.constant, flow.variable_dependent,\
                  flow.boundary.isgated)

            # Check for epot_scale_all 
            epot_scale_all = flow.epot_scale_all

            if epot_scale_all:
                # Iterate over the scalar and const species and check for
                # existance of scale factor in domain dict
                for species in flow.scalar_species+flow.const_species:
                    if species.domain != flow.epot_domain:
                        if not species.species in flow.epot_domain.scale:
                            raise ValueError, "the species '%s' is not added"\
                                  " to the %s domain, which is needed for "\
                                  "scaling %s correctly."%\
                                  (str(species.species), \
                                   str(flow.epot_domain), str(species))
        else:
            epot_scale_all = False
        
        # Create namspaces and sympy subs dict
        flow_ns, subs = self.sympy_to_ufl_namespaces()
        
        # Fetch of MeshDomains to check how we should implement the electric
        # potential scale of scalar or constant species
        # The scaling of scalar and const species with an electric potential 
        # can only be done for these species if they are not defined on the
        # actuall MeshDomain
        mesh_domains = set()
        
        # Iterate over the ScalarFieldSpecies
        for species in flow.scalar_field_species:
            # Create a SymPy symbol with a unique name which 
            # will be bound to a ufl func through the name space
            sympy_name = str("_".join(str(species).split(".")))
            subs[species] = sympy.Symbol(sympy_name)

            debug("ScalarFieldSpecies: %s", str(species))
            
            # Weight any scalar field species with beta function
            flow_ns[sympy_name] = species.ufl_func["beta"]*\
                                  species.ufl_func["u"]
            
            mesh_domains.add(species.domain)
        
        # Iterate over the ScalarSpecies
        for species in flow.scalar_species:
            sympy_name = str("_".join(str(species).split(".")))
            
            # Populate namespaces
            subs[species] = sympy.Symbol(sympy_name)

            debug("ScalarSpecies: %s", species)
            
            # Check for scaling if electric potential
            if epot_scale_all and flow.epot_domain != species.domain:
                func = flow.epot_domain.scale[species.species]*\
                       species.ufl_func["u"]
            elif "beta" in species.ufl_func:
                func = species.ufl_func["u"]*species.ufl_func["beta"]
            else:
                func = species.ufl_func["u"]
            flow_ns[sympy_name] = func
        
        # Iterate over the ConstSpecies
        for species in flow.const_species:
            sympy_name = str("_".join(str(species).split(".")))
            subs[species] = sympy.Symbol(sympy_name)
            if self.const_species_to_numbers or non_vars_to_scalars:
                func = species.init.get()
            else:
                func = Constant(species.init.get())
            
            debug("ConstSpecies: %s=%.4f", str(species), species.init.get())
            
            # Check for scaling if electric potential
            if epot_scale_all and flow.epot_domain != species.domain:
                func = flow.epot_domain.scale[species.species]*func
            elif "beta" in species.ufl_func:
                func = func*species.ufl_func["beta"]
            
            flow_ns[sympy_name] = func
        
        # Create Constants/Functions of the Parameters
        for parsym in flow.parameters:
            # Need a name based on the name of the ParSlave instead of
            # the symbol
            if isinstance(parsym.param, ParSlave):
                sympy_name = str(parsym.param)
            else:
                sympy_name = str("_".join(str(parsym).split(".")))
            subs[parsym] = sympy.Symbol(sympy_name)
            param = parsym.param

            # Check for allowed ArrayParam
            if param.is_node_param or param.is_cell_param:

                assert(not non_vars_to_scalars)
                # Assign Fucntion based on type of ArrayParam
                if param.is_node_param:
                    func = Function(self.Vs)
                    inds = param.domain.value_inds
                else:
                    func = Function(self.Vc)
                    inds = param.domain.cell_inds
                
                v = func.vector()
                v.zero()
                
                # Assign values
                values = param.get()
                v[inds] = values
                ## FIXME: Continue here!!!!!
                out = " (%s)"%param.symbols if isinstance(param, ParSlave) \
                      else ""
                debug("%s: %s, sum=%.4f, min=%.4f, max=%.4f%s", \
                      param.__class__.__name__, param.name, \
                      values.sum(), values.min(), values.max(), out)
                
            elif param.is_scalar:
                out = " (%s)"%param.symbols if isinstance(param, ParSlave) \
                      else ""
                debug("%s: %s=%.4f%s", param.__class__.__name__, \
                      param.name, param.get(), out)

                if non_vars_to_scalars:
                    func = param.get()
                else:
                    func = Constant(param.get())
            else:
                raise RuntimeError, "Can only handle parameters that are "\
                      "expressions of Node- or CellParams or scalars."
            
            # Assign the function to the namespace
            flow_ns[sympy_name] = func

        # Create Expressions for the Variables
        for variable in flow.variables:
            # Get SymPy name
            sympy_name = str("_".join(str(variable).split(".")))
            subs[variable] = sympy.Symbol(sympy_name)
            
            debug("Variable: %s=%s", variable, sympy_name)
            
            # Map Expression stored at each Variable
            flow_ns[sympy_name] = variable._func
        
        return flow_ns, subs
    
    def step_back(self, offset):
        "Update the continuos solution with an offset"

        assert(isinstance(offset, float))
        assert(offset>0)
        
        # Calculate the difference in u0 and u1 and scale it with the offset
        du = self.vectors["tmp"]
        du._assign(self.vectors["u1"])
        du.axpy(-1, self.vectors["u0"])
        
        # Update the solution vector
        self.vectors["u1"].axpy(-offset, du)
        self.vectors["u"]._assign(self.vectors["u1"])

        # Iterate over the vectors and step back them
        if self.vectors["fixed"]:
            du = self.vectors["fixed"]["tmp"]
            for species, vecs in self.vectors["fixed"].items():
                if species == "tmp":
                    continue
                
                du._assign(vecs["u"])
                du.axpy(-1, vecs["u0"])
                vecs["u"].axpy(-offset, du)
                
    def update_fixed_solutions_before_solve(self):
        "Update u0 and u from previous solve"
        # FIXME: Consider move to solve
        for species, vecs in self.vectors["fixed"].items():
            if species == "tmp":
                continue
            vecs["u0"]._assign(vecs["u"])

    def update_solutions_after_solve(self):
        "Update the u1 vector from the solution vector"
        # FIXME: Consider move to solve
        self.vectors["u1"]._assign(self.vectors["u"])
        
    def update_u0_solutions_before_solve(self):
        "Update u0 from previous solve"
        # FIXME: Consider move to solve
        self.vectors["u0"]._assign(self.vectors["u1"])

    def update_u_solutions_before_solve(self):
        "Update u from previous solve"
        # FIXME: Consider move to solve
        self.vectors["u"]._assign(self.vectors["u1"])

class IntegralHandler(object):
    """
    A class that calculate the different integrals (UFL forms) and keep
    track of them
    """
    def __init__(self, function_handler, constant_reaction_jacobian, \
                 dolfin_params):
        self.function_handler = function_handler
        self.subcellmodel = function_handler.subcellmodel
        self.domain_handler = function_handler.domain_handler
        self.dolfin_params = dolfin_params

        # All common forms
        # (including constant fluxes that are not on a gated Boundary)
        self.forms = odict()
        
        # All gated constant flux J forms, the non gates are included 
        # in ufl_forms These are only used to assemble the
        # active_const_J_matrices
        self.flux_gated_constant_J_integrals = odict()
        
        # All active non linear flux F integrals
        self.flux_forms = odict()
        self.flux_forms["real F"] = zero_integral
        self.flux_forms["F"] = zero_integral
        
        # All active flux J forms that are going to be reassembled for
        # each Newton iteration
        self.flux_forms["real reassemble J"] = zero_integral
        self.flux_forms["reassemble J"] = zero_integral
        
        # All active flux J forms that are going to be reassembled each
        # timestep
        self.flux_forms["real reassemble once J"] = zero_integral
        self.flux_forms["reassemble once J"] = zero_integral
        
        # J integrals that are constant and active during the whole simulation
        self.flux_forms["real constant F form"] = zero_integral
        self.flux_forms["constant F form"] = zero_integral

        # Create the boundary values integrals and attach them to the
        # boundary values
        self.init_boundary_value_integrals()
        
        # Create integrals for the reactions and attach them to the reactions
        self.init_reactions()
        
        # Init integrals for the fluxes
        self.init_flux_integrals()
        
        # Add reaction terms to flux integrals
        self.handle_reaction_integrals(constant_reaction_jacobian)
        
        # Sort the integrals and add them to an appropriate form
        self.handle_flux_integrals()
        
        # Init ufl forms for easy access for assembly
        self.init_base_forms()
        

    def assemble(self, form, tensor=None, add_values=False,
                 active_domains=False, real=False):

        # First check for empty form
        # NOTE: Need double negation to avoid creating an Equation
        if zero_forms(form):
            if tensor is None:
                raise RuntimeError("Got empty form with no tensor")
            return

        form_data = form.compute_form_data()
        if not real:
            parameters.num_threads = self.dolfin_params.num_threads
            kwargs = self.domain_handler.assemble_kwargs(\
                tensor, add_values, active_domains)
            kwargs["exterior_facet_domains"] = None
            A = assemble(form, **self.domain_handler.assemble_kwargs(\
                tensor, add_values, active_domains))
            parameters.num_threads = 0
            return A
        else:
            return assemble(form, **self.domain_handler.assemble_kwargs(\
                tensor, add_values, active_domains))

    def init_boundary_value_integrals(self):
        """
        Create integrals for each boundary value
        """

        debug("")
        debug("Init BoundaryValue integrals:")
        add_log_indent()
        # Create the ufl forms for the boundaryvalues
        for value in self.subcellmodel.boundary_values:
            
            # Get the scalar field species
            species = value.domain.species[value.species]
            u = species.ufl_func["u"]
            
            # Check for a static electric potential
            if species.ufl_func["beta"] != 1:
                u = species.ufl_func["beta"]*u
            
            debug("Boundary value: %s", u)
            
            # Create and store the form for the right and left hand side
            area = value.area if self.function_handler.params_to_numbers \
                   else Constant(value.area)
            value.F_integral = (value.ufl_func["u"]-u)/\
                               area*value.ufl_func["v"]*ds(value.marker)
            value.J_integral = derivative(value.F_integral,\
                                          self.function_handler.u["full"],\
                                          self.function_handler.du["full"])
            debug("Boundary form: %s", value.F_integral)
            
        add_log_indent(-1)
        
    def init_reactions(self):
        """
        Create integrals for the reactions
        """
        for react in self.subcellmodel.reactions:

            # Continue if reaction is going to be handled by ReactionSolver

            if self.function_handler.operator_split_reactions and \
               react.only_field_species:
                continue
            
            # Create the namespace each reaction integral will be avaluated in
            react_ns, subs = self.function_handler.create_flow_namespace(react)

            # Get the expr
            expr = react.expr(self.function_handler.params_to_numbers)
            debug("Reaction expr: %s", expr)
            expr = expr.subs(subs) if isinstance(expr, sympy.Basic) else expr
            debug("Reaction expr: %s", expr)
            
            # Create the str representation of the expr which will be 
            # evaluated in the generated namespace
            react_str = str(expr)
            react_eval = ""
            
            # Iterate over the different reactants and products and create an
            # integral for each
            for i, (species, stoch) in enumerate(react.reactants.items() +\
                                                 react.products.items()):
                react_ns["v%d"%i] = react.domain.species[species].ufl_func["v"]
                sign = "+" if species in react.products else "-"
                
                # Check if the reactant/product is a scalar species
                if isinstance(react.domain.species[species], ScalarSpecies):

                    # If the reaction is defined on a MeshDomain
                    if isinstance(react.domain, MeshDomain):
                        mesh_domain = react.domain
                        domain_volume = react.domain.volume.get()
                        marker = mesh_domain.marker
                        integrand = "dx"
                        volume_scale = domain_volume/mesh_domain.volume.get()
                    else:
                        domain_volume = react.domain.volume.get()
                        marker = self.domain_handler.\
                                 marker_to_smallest_mesh_boundary

                        # No Mesh Boundary we just grab the first MeshDomain
                        if marker is None:
                            mesh_domain = self.subcellmodel.\
                                          scalar_field_species[0].domain
                            marker = mesh_domain.marker
                            volume_scale = domain_volume/\
                                           mesh_domain.volume.get()
                            integrand = "dx"
                        else:
                            boundary_area = self.domain_handler.areas[marker]
                            volume_scale = domain_volume/boundary_area
                            integrand = "ds"
                            #print boundary_value

                    # Converting to a Constant to avoid triggering
                    # recompilation
                    if not self.function_handler.params_to_numbers:
                        volume_scale = Constant(volume_scale)

                    # Update namespace
                    react_ns["volume_scale"] = volume_scale
                    
                    react_eval += "%s%d*volume_scale*(%s)*v%d*%s(%d)"%(\
                        sign, stoch, react_str, i, integrand, marker)
                else:
                    react_eval += "%s%d*(%s)*v%d*dx(%d)"%(sign, stoch,\
                                        react_str, i, react.domain.marker)

            
            # Reverse the sign of the whole integral statement
            react_eval = "-(%s)"%react_eval
            
            debug("  Reaction eval: %s"%react_eval)

            # Create the UFL forms
            F_integral = eval(react_eval, react_ns)
            
            react.J_integral = derivative(F_integral, \
                                          self.function_handler.u["full"], \
                                          self.function_handler.du["full"])
            react.F_integral = F_integral

    def init_flux_integrals(self):
        """
        Initialize the ufl forms for all fluxes
        """
        subcellmodel = self.subcellmodel

        # Create the ufl forms for the flux instances

        debug("")
        debug("Init flux integrals")
        add_log_indent()
        for flux in subcellmodel.fluxes:
            # In this loop we substitute the symbols in the sympy expression 
            # with symbols with name which we then associate ufl objects with.
            
            # Initialize structures
            flux.F_integrals = []
            flux.J_integrals = []

            # Create the namespace for the species defined in the flux
            flux_ns, subs = self.function_handler.create_flow_namespace(flux)

            # FIXME: More code re-use please!
            
            # If the flux is connects two FieldSpecies we need to create
            # an integral for each flux instance
            if flux.num_scalar_field_species_domains() == 2:
                
                # Check if both receiving species are ScalarFieldSpecies
                if all(isinstance(domain.species[flux.species], \
                                  ScalarFieldSpecies)
                       for domain in flux.boundary.domains):
                    markers = [[m0, m1] for m0, m1 in \
                               zip(*flux.boundary.markers)]
                
                # else assume one is scalar
                else:
                    markers = [[m, m] for m in flux.boundary.markers[0]]
                
                # Iterate over the different instances of the Boundary
                for ind in xrange(flux.boundary.num):
                    
                    # Iterate over the ScalarFieldSpecies and collect
                    # BoundaryValues (this will overwrite the values defined
                    # in the namespace already)
                    for species in flux.scalar_field_species:
                        boundary_value = flux.boundary.boundary_values[\
                            (species, ind)]
                        sympy_name = str(boundary_value)
                        
                        # Populate namespaces
                        subs[species] = sympy.Symbol(sympy_name)
                        flux_ns[sympy_name] = boundary_value.ufl_func["u"]
                    
                    # Create the integrals
                    self.create_and_append_boundary_integrals(flux, flux_ns,\
                                    subs, flux.boundary.domains, markers[ind])
                    
            # Just one FieldSpecies
            elif any(isinstance(domain.species[flux.species], \
                                ScalarFieldSpecies) \
                     for domain in flux.boundary.domains):
                
                # Grab the Domains
                domains = mesh_domains(flux.boundary.domains) + \
                          volume_domains(flux.boundary.domains)
                
                domains = [domain for domain in domains if \
                           not isinstance(domain.species[flux.species], \
                                          ConstSpecies)]
                
                # Grab the only marker list and integrate over it
                for marker in flux.boundary.markers[0]:
                    
                    markers = [marker]*len(domains)
                    
                    # Create the integrals
                    self.create_and_append_boundary_integrals(flux, flux_ns,\
                                                    subs, domains, markers)
            
            # All receiving domains are VolumeDomains
            else:
                
                # Find a host Boundary
                marker = self.domain_handler.marker_to_smallest_mesh_boundary
                if marker is None:
                    raise NotImplementedError, "we need at least one "\
                          "MeshBoundary to be able to integrate a flux "\
                          "on a pure Boundary"
                domains = volume_domains(flux.boundary.domains)
                
                domains = [domain for domain in domains if \
                           not isinstance(domain.species[flux.species], \
                                          ConstSpecies)]
                
                area_scale = flux.boundary.area.get()/\
                             self.domain_handler.areas[marker]
                markers = [marker]*len(domains)
                
                # Create the integrals
                for i in xrange(flux.boundary.num):
                    self.create_and_append_boundary_integrals(flux, flux_ns,\
                                                    subs, domains, markers,\
                                                    area_scale)
        add_log_indent(-1)

    def create_and_append_boundary_integrals(self, flux, flux_ns,\
                                              subs, domains, markers, \
                                              area_scale=1):
        """
        A helper method that _init_flux_ufl_forms used to create the
        flux forms"""
        # Create the new SymPy expression with substituted values for the
        # domain specific species
        expr = flux.expr(self.function_handler.params_to_numbers)

        expr = expr.subs(subs) if isinstance(expr, sympy.Basic) else expr
        
        flux_str = str(expr)
        
        # Start build the boundary integral for the ScalarFieldSpecies
        flux_eval = ""
        
        # Converting to a Constant to avoid triggering recompilation
        if area_scale != 1 and not self.function_handler.params_to_numbers:
            area_scale = Constant(area_scale)

        # Output area scale for debugging purposes
        flux_ns["area_scale"] = area_scale
        debug("area_scale: %.4f", area_scale)
        
        # The number of markers decides the number of integrals
        # For a flux that connects two FieldSpecies we need two integrals
        for i, (marker, domain) in enumerate(zip(markers, domains)):
            
            # Get the test function of the recieving species
            flux_ns["v%d"%i] = domain.species[flux.species].ufl_func["v"]
            
            # Get the sign of the flux
            sign = flux.sign(domain)
            
            # If only one marker we check flux sign if + get rid of it
            if i==0 and sign == "+":
                sign = ""
            
            # Create the scalar field species boundary integral
            flux_eval += "%s(%s)*v%d*area_scale*ds(%d)"%(sign, flux_str, i,\
                                                         marker)
            
            # Create the flux_log_eval (need only one version of this)
            # FIXME: Reuse this for pure scalar fluxes?
            #if i == 0:
            #    # NOTE: If flux is a scalar constant we need a hack because 
            #    # NOTE: 5*ds is not allowed in ufl
            #    flux_str_eval = eval(flux_str, flux_ns)
            #    if isinstance(flux_str_eval, scalar):
            #        form = Constant(flux_str_eval)*ds(marker)
            #    else:
            #        form = eval("(%s)*ds(%d)"%(flux_str, marker), flux_ns)
            #        
            #    # Attach the flux log forms
            #    self.forms["log"]["boundary_fluxes"][str(flux.boundary)]\
            #            [str(flux.species)].append(form)
        
        # Reverse the sign of the whole integral statement
        flux_eval = "-(%s)"%flux_eval

        debug("Flux eval: %s"%flux_eval)

        # Create the UFL forms
        F_integral = eval(flux_eval, flux_ns)

        debug("Flux ufl: %s"%F_integral)
        J_integral = derivative(F_integral, self.function_handler.u["full"],\
                                self.function_handler.du["full"])
        
        debug("Flux Jacobian ufl: %s"%expand_derivatives(J_integral))
        flux.J_integrals.append(J_integral)
        flux.F_integrals.append(F_integral)

    def handle_flux_integrals(self):
        """
        1) Create and attach callback methods for activation of fluxes
        """

        subcellmodel = self.subcellmodel
        
        # FIXME: Rearrange this. Collect all forms in the different sub
        # categorize, and controll which get assembled using a MeshFunction,
        # with altering active domain markers
        #
        # For this to work we need to instantiate a meshfunction for each 
        # type of form integral. We also need to fix a fast way of marking 
        # and unmarking a mesh function. This should be done in the callback
        # function The different type of 
        
        # Iterate over all fluxes
        if not subcellmodel.fluxes:
            return
        debug("")
        debug("Handling flux integrals")
        add_log_indent()
        for flux in subcellmodel.fluxes:

            # Get flux flags
            robin, constant, var_dep, gated, boundary = flux.robin, \
                                flux.constant, flux.variable_dependent, \
                                flux.boundary.isgated, flux.boundary
            
            # The nonlinear integral will be integrated each newton iteration
            # so lump all F_integrals into on big happy F_form
            # (Hopefully not too big...)
            # If the flux connects one or more ScalarSpecies it includes a
            # testfunction of a Real SubSpace
            if flux.num_scalar_species_domains() > 0:
                self.flux_forms["real F"] += add_forms(flux.F_integrals)
            else:
                self.flux_forms["F"] += add_forms(flux.F_integrals)

            # --- collect the integrals dependent on the flags  ---

            # If the flux is of robin type it will have a nonzero J integral
            if robin:
                
                # If we essentially are a constant flux we can just add it to the part that will get added to the 
                if constant and not var_dep:
                    
                    # If gated add the constant gated robin J integrals
                    # to a dict 
                    if gated:
                        debug("Adding '%s' flux '%s' on '%s' to gated forms",\
                              flux.species, flux, flux.boundary)
                        for ind in xrange(boundary.num):
                            # If an integral from this boundary is already
                            # register
                            if (boundary, ind) in \
                                   self.flux_gated_constant_J_integrals:
                                # Add it
                                self.flux_gated_constant_J_integrals\
                                    [(boundary, ind)] += flux.J_integrals[ind]
                            else:
                                # If not register the new one
                                self.flux_gated_constant_J_integrals\
                                    [(boundary, ind)] = flux.J_integrals[ind]
                    
                    # else add them to a form which will be integrated once
                    else:

                        debug("Adding '%s' flux '%s' on '%s' to constant "\
                              "F form", flux.species, flux, flux.boundary)

                        # Add all to real constant F form as they might turn
                        # into real one once they get differentiated
                        self.flux_forms["real constant F form"] += \
                                              add_forms(flux.F_integrals)
                            
                # If we are a semi constant flux
                elif constant and var_dep:
                    
                    debug("Adding '%s' flux '%s' on '%s' to reassemble "\
                          "once J", flux.species, flux, flux.boundary)
                    # Add constant but variable dependent robin
                    # integrals to form
                    # Add all to real reassemble once J form as they might turn
                    # into real one once they get differentiated
                    self.flux_forms["real reassemble once J"] += \
                                          add_forms(flux.J_integrals)
                        

                # If we are a non linear flux
                else:
                    # Add nonlinear robin integrals to form
                    debug("Adding '%s' flux '%s' on '%s' to reassemble J", \
                          flux.species, flux, flux.boundary)
                    # Add all to real reassemble J form as they might turn
                    # into real one once they get differentiated
                    self.flux_forms["real reassemble J"] += \
                                          add_forms(flux.J_integrals)
        
        add_log_indent(-1)

    def handle_reaction_integrals(self, constant_reaction_jacobian):
        """
        Add the reaction integrals to the flux forms
        """

        subcellmodel = self.subcellmodel
        
        # Put a flag for the use of reactions
        self.use_reactions = bool(subcellmodel.reactions)

        if not self.use_reactions:
            return

        debug("")
        debug("Handling reaction integrals")
        add_log_indent()

        # Iterate over all reactions
        for react in subcellmodel.reactions:

            if self.function_handler.operator_split_reactions and \
               react.only_field_species:
                continue
            
            # Get flux flags
            robin, constant, var_dep = react.robin, \
                            react.constant, react.variable_dependent
            
            debug("Reaction in '%s'", str(react.domain))
            debug("robin: %d; constant %d; vardep %d;", robin, \
                  constant, var_dep)
            
            # If the flux is of robin type it will have a nonzero J integral
            if robin:
                
                # If we are a semi constant (linear) reaction
                if constant_reaction_jacobian or (constant and var_dep):
                    
                    # Add constant but variable dependent robin
                    # integrals to form
                    if isinstance(react.domain, MeshDomain):
                        self.flux_forms["reassemble once J"] += \
                                                react.J_integral
                    else:
                        self.flux_forms["real reassemble once J"] += \
                                               react.J_integral
                                                
                # If we are a non linear reaction
                else:
                    # Add nonlinear robin integrals to form
                    if isinstance(react.domain, MeshDomain):
                        self.flux_forms["reassemble J"] += \
                                                react.J_integral
                    else:
                        self.flux_forms["real reassemble J"] += \
                                               react.J_integral

        add_log_indent(-1)
                    
    def init_base_forms(self):
        "Create the main ufl forms for the stiffnes and mass matrix"
        
        subcellmodel = self.subcellmodel
        
        # Collect the forms
        # forms[0]: The mass part of the lhs
        # forms[1]: The mass part of the rhs 
        # forms[2]: The stiffnes part of the rhs
        forms = [zero_integral]*3
        
        # Iterate over the field species
        for species in subcellmodel.scalar_field_species:
            sigma_zero = False
            if species.sigma.is_scalar:
                sigma_zero = species.sigma.get() == 0.0
                if not sigma_zero:
                    sigma = species.sigma.get() if self.function_handler.\
                            params_to_numbers else \
                            Constant(species.sigma.get())
            elif species.sigma.is_array:
                sigma = Function(self.function_handler.Vc)
                values = species.sigma.get()
                if not all(values>0):
                    raise ValueError("expected all values of sigma for %s"\
                                     " to be positive"%species)
                sigma.vector()[species.domain.cell_inds] = values
            else:
                raise TypeError("expected sigma for %s to be a scalar or "\
                                "a CellParam"%species)
            bases = [species.ufl_func["du"]*species.ufl_func["v"],\
                     (species.ufl_func["u"]-species.ufl_func["u0"])*\
                     species.ufl_func["v"]/self.function_handler.dt,\
                     ufl.constantvalue.Zero() if sigma_zero else
                     sigma*inner(grad(species.ufl_func["u"]),\
                                 grad(species.ufl_func["v"]))]
            
            # Iterate over the bases and add common part of intgral
            for i, base in enumerate(bases):
                # Weight the form with a the "beta" Function
                base = species.ufl_func["beta"]*base
                forms[i] += base*dx(species.domain.marker)

        # Collect the volume forms
        self.forms["J_M"], self.forms["F_K_transient"],\
                           self.forms["F_K"] = forms

        # Add facet integrals
        self.forms["real J_M"] = zero_integral
        self.forms["real F_K_transient"] = zero_integral
        self.forms["real F_K"] = zero_integral

        # Generate Mass for any scalar species
        if self.function_handler.num_scalars>0:
            # Grab the first scalar field species
            # When we integrate a scalar species we need to integrate over
            # a scalar field species. We therefore collect the marker and
            # volume of the domain of that species. The integral will then be
            # over that domain and we then scale the integral with the volume
            # of the corresponding mesh domain.
            if all(species is None for species in \
                   subcellmodel.scalar_field_species):
                raise NotImplementedError, "problem with no "\
                      "ScalarFieldSpecies is not yet implemented"

            # Collect forms from the scalar species
            for species in subcellmodel.scalar_species:
                
                domain_volume = species.domain.volume.get()
                marker = self.domain_handler.marker_to_smallest_mesh_boundary

                # If there are no mesh boundaries declared in the model
                if marker is None:
                    mesh_domain = self.subcellmodel.\
                                  scalar_field_species[0].domain
                    marker = mesh_domain.marker
                    volume_scale = domain_volume/mesh_domain.volume.get()
                    integrand = dx
                    scalar_species_uses_volume = True
                    
                else:
                    boundary_area = self.domain_handler.areas[marker]
                    volume_scale = domain_volume/boundary_area
                    integrand = ds

                # Converting to a Constant to avoid triggering recompilation
                if not self.function_handler.params_to_numbers:
                    volume_scale = Constant(volume_scale)

                # Add forms to real forms, so they wont be assembled using
                # the volume assembler, as OpenMP does not support assemble
                # of Global dofs
                self.forms["real J_M"] += volume_scale*\
                                           species.ufl_func.du*\
                                           species.ufl_func.v*integrand(marker)
                
                self.forms["real F_K_transient"] += volume_scale*\
                                    (species.ufl_func.u-species.ufl_func.u0)*\
                                    species.ufl_func.v/\
                                    self.function_handler.dt*\
                                    integrand(marker) 
        
        # Add the boundary value forms
        self.forms["real F_K"] += add_forms(
            value.F_integral for value in subcellmodel.boundary_values)

        # RHS to be differentiated
        # NOTE: constant (linear) forms are added here so they are included
        # NOTE: into the linearized forms
        real_F_K = self.forms["real F_K"] + \
                   self.flux_forms["real constant F form"]
        F_K = self.forms["F_K"] + \
                     self.flux_forms["constant F form"]

        # Differentiate the rhs, J_K now includes the form for the 
        # Stiffnes Matrix together with the form for the Boundary Values
        # and constant fluxes
        self.forms["real J_K"] = derivative(real_F_K, \
                                       self.function_handler.u.full, \
                                       self.function_handler.du.full)
        
        self.forms["J_K"] = derivative(F_K, \
                                       self.function_handler.u.full, \
                                       self.function_handler.du.full)
        
        # Continue building rhs form (Now adding all flux forms)
        self.forms["F_K"] += self.flux_forms["F"]
        self.forms["real F_K"] += self.flux_forms["real F"]
        
        # Add the reaction forms
        # NOTE: Only adding if react has F_integral, which is doesn't if
        # NOTE: the reaction is operator splitted
        for react in subcellmodel.reactions:
            if not hasattr(react, "F_integral"):
                continue
            if isinstance(react.domain, MeshDomain):
                self.forms["F_K"] += react.F_integral
            else:
                self.forms["real F_K"] += react.F_integral
        
        # Add the rest
        self.forms["F_K_transient"] += self.forms["F_K"]
        self.forms["real F_K_transient"] += self.forms["real F_K"]
        
        debug("")
        debug("Basic forms:")
        add_log_indent()
        debug("Stiffness form: %s", expand_derivatives(\
            self.forms["J_K"]))
        debug("Stiffness form real part: %s", expand_derivatives(\
            self.forms["real J_K"]))
        debug("")
        debug("Mass form: %s", self.forms["J_M"])
        debug("Mass form real part: %s", self.forms["real J_M"])
        debug("")
        debug("F form: %s", self.forms["F_K"])
        debug("F form real part: %s", self.forms["real F_K"])
        debug("")
        debug("Transient F form: %s", self.forms["F_K_transient"])
        debug("Transient F form real part: %s", self.forms["real F_K_transient"])
        
        add_log_indent(-1)
        
        # NOTE: Both F_K and FK_transient now consists of every thing without 
        # the time discretization part. By controlling the markers in the 
        # MeshFunction providing the external_facet_domains we controll what 
        # parts of this big form are assembled for each time step.


class TensorHandler(object):
    """
    A class that holds the assembled tensors
    """
    def __init__(self, integral_handler):
        self.subcellmodel = integral_handler.subcellmodel
        self.domain_handler = integral_handler.domain_handler
        self.integral_handler = integral_handler
        self.function_handler = integral_handler.function_handler
        self.init_tensors()
        
    def init_tensors(self):
        """
        Create some base tensors.

        A:  Is the matrix that is used to build the jacobian matrix each
            Newton iterations. Summing into an existing matrix saves time.

        K: Is the stiffness matrix used to build the jacobian 
        M: The mass matrix used to build the time discretized part of
           the jacobian

        b: Is the right hand side vector. We prealloacte memory for this
            vector, speeding up the assembly process.
            (probably not that much...)
        
        tmp: A tmp vector for scratch calculations
        
        Note that both K and M are constant during the whole simulation.
        """

        # Start timer
        timer = Timer("Assemble of base tensors")

        self.matrices = {}
        self.vectors = {}

        info("")
        info("Assemble base tensors:")
        add_log_indent()
        info("Initial Jacobian (Stiffnes) matrix")
        A = self.integral_handler.assemble(\
            self.integral_handler.forms["J_K"])
        self.integral_handler.assemble(\
            self.integral_handler.forms["real J_K"], A, \
            add_values=True, real=True)
        
        self.matrices["A"] = A
        
        # Copy the stiffness matrix
        self.matrices["K"] = self.matrices["A"].copy()

        info("Assemble initial load vector")
        self.vectors["b"] = self.integral_handler.assemble(\
                            self.integral_handler.forms["F_K"])
        
        info("Assemble Mass matrix")
        M = self.integral_handler.assemble(\
            self.integral_handler.forms["J_M"], A.copy())

        self.integral_handler.assemble(\
                self.integral_handler.forms["real J_M"], M, \
            add_values=True, real=True)
        
        self.matrices["M"] = M
        
        # Initate a matrix keeping all the constant flux matrices
        self.matrices["active constant fluxes"] = A.copy()
        self.matrices["active constant fluxes"].zero()
        self.matrices["inactivation fluxes"] = A.copy()
        self.matrices["inactivation fluxes"].zero()
        add_log_indent(-1)
        
    def reset_before_run(self):
        # Reset constant flux matrix
        
        self.matrices["active constant fluxes"].zero()
        
        def function_closure(boundary):
            """
            Function closure to generate the activation callback
            """
            
            def activation_callback(ind, value):
                """
                Callback function for Activation of gated boundary

                ind is the sub indice of the boundary and value is the
                value that either opens or closes the channel
                """
                # If activating
                if bool(value):
                    # Print that we are activating the boundary
                    info(wrap_log_message("Activating: %s %d"%(boundary.name,\
                                                               ind), "#"))
                    # First check if some action should be done wrt
                    # constant fluxes
                    if (boundary, ind) in self.integral_handler.\
                           flux_gated_constant_J_integrals:

                        debug("Adding constant flux contribution to "\
                              "Stiffnes part: %s[%d]", boundary, ind)
                        
                        # Add key to list
                        self.domain_handler.active_constant_boundaries.\
                                                    append((boundary, ind))

                        # Get the matrix
                        A = self.matrices["active constant fluxes"]

                        # Assemble the contribution of the constant flux
                        form = self.integral_handler.\
                               flux_gated_constant_J_integrals\
                               [(boundary, ind)]
                        assemble(form, **self.domain_handler.\
                                 assemble_kwargs(A, add_values=True))
                        
                    # Iterate over the marker lists
                    for markers in boundary.markers:

                        # Add marker to the list of active boundary markers
                        self.domain_handler.active_boundary_markers.append(\
                            markers[ind])
                        
                        debug("Adding '%d' to active face markers",\
                              markers[ind])
                        
                        # Add the indices in the active_face_domains
                        # MeshFunction for the correspondning marker
                        self.domain_handler.active_face_domains.array()\
                                    [self.domain_handler.face_domain_indices\
                                     [markers[ind]]] = markers[ind]
                    
                # If de-activating 
                else:
                    info(wrap_log_message("De-Activating: %s %d"%\
                                          (boundary.name, ind), "#"))
                    # First check if some action should be done wrt constant
                    # fluxes
                    if (boundary, ind) in self.integral_handler.\
                           flux_gated_constant_J_integrals:
                        
                        debug("Removing constant flux contribution to "\
                              "Stiffnes part: %s[%d]", boundary, ind)
                        
                        # Remove key from list
                        if (boundary, ind) in self.domain_handler.\
                               active_constant_boundaries:
                            self.domain_handler.active_constant_boundaries.\
                                remove((boundary, ind))
                        
                        # Get the matrix
                        tmp = self.matrices["inactivation fluxes"]
                        
                        # Assemble the contribution of the constant flux
                        form = self.integral_handler.\
                               flux_gated_constant_J_integrals\
                               [(boundary, ind)]
                        assemble(form, **self.domain_handler.\
                                 assemble_kwargs(tmp))
                        
                        # Remove contribution from the constant matrix
                        self.matrices["active constant fluxes"].\
                                              axpy(-1, tmp, True)
                        
                    # Iterate over the marker lists
                    for markers in boundary.markers:
                        
                        # Check if marker is registered in the active boundary
                        # marker list
                        if markers[ind] in self.domain_handler.\
                               active_boundary_markers:

                            debug("Removing '%d' from active boundary "\
                                  "markers"%markers[ind])

                            # Remove marker from the list of active boundary
                            # markers
                            self.domain_handler.active_boundary_markers.\
                                                         remove(markers[ind])
                            
                        # Nullify the indices in the active_face_domains
                        # MeshFunction
                        self.domain_handler.active_face_domains.array()\
                                    [self.domain_handler.face_domain_indices\
                                    [markers[ind]]] = 0
                
            return activation_callback
        
        # Iterate over the gated boundaries and create a callback method
        # and attach them to the Boundary
        for boundary in self.subcellmodel.iter_gated_mesh_boundaries():
            # Attach the function to the boundary
            boundary.activation_callback = function_closure(boundary)
        
