__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-16 -- 2012-07-23"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["load_mesh_data", "load_and_check_mesh_data"]

# System imports
from dolfin import (Mesh, MeshFunction, VertexFunction,
                    CellFunction, Cell, Face, cells, Timer)
import os
import numpy as np

# subcell imports
from subcell.common import adict, debug
from subcell.model.subcellmodel import SubCellModel

def load_mesh_data(filename):
    assert(isinstance(filename, str))
    
    debug("Loading Mesh")
    # Check the mesh filenames
    if not os.path.isfile(filename):
        if not os.path.isfile(filename+".xml"):
            if not os.path.isfile(filename+".xml.gz"):
                raise IOError, "mesh file '%s'{.xml{.gz}} does "\
                      "not excist"%filename
            else:
                filename = filename+".xml.gz"
        else:
            filename = filename+".xml"

    facemarker_complement = "_face_markers"
    cellmarker_complement = "_tetrahedron_attributes"
    
    # Extract filename information
    basename = filename.split(".xml")[0]
    suffix = filename.replace(filename.split(".xml")[0], "")

    # Load mesh
    t=Timer("Load Mesh")
    mesh_data = adict()
    mesh = Mesh(filename)
    mesh_data.mesh = mesh

    # Check dimensions
    assert(mesh.topology().dim()==3)

    # Create return structure
    mesh_data.markers = adict(cell=None, face=None, vertex=None)
    mesh_data.markers.vertex = adict(cell=None, face=None)
    mesh_data.areas = None
    mesh_data.volumes = None

    # Check for available face mesh function
    face_name = "%s%s%s"%(basename, facemarker_complement, suffix)
    if os.path.isfile(face_name):

        # Load face markers and remove 0 from distinct
        t=Timer("Load face markers")
        debug("Loading FaceMarkers")
        face_markers = MeshFunction("uint", mesh, face_name)

    else:
        # Face markers not in own file check if contained in mesh
        face_markers = mesh.domains().facet_domains(mesh)
        if face_markers is None:
            raise IOError("No face markers found. Either provide them in "\
                          "a Mesh file or as a face marker file with name:"\
                          " '%s'." % face_name)
        
        # Make max value equals 0
        face_markers.array()[\
            face_markers.array() == face_markers.array().max()] = 0
        
    # FIXME: Collect _all_ values also them on different processors
    distinct = list(sorted(set(face_markers.array())))
    if 0 in distinct:
        distinct.remove(0)

    # FIXME: Add check when getting indices of off processor marker,
    # FIXME: should then return []
    inds = dict([(m, (face_markers.array()==m).nonzero()[0]) \
                 for m in distinct])
    mesh_data.markers.face = adict(inds=inds, func=face_markers, \
                                   distinct=distinct)
    
    # Initialize area structure
    areas = {}.fromkeys(distinct, 0.0)
    mesh_data.areas = areas

    # Compute vertex markers belonging to given faces
    # Avoid marker = 0
    vertex_markers = VertexFunction("uint", mesh)
    vertex_markers.set_all(0)
    inds = dict((m, set()) for m in distinct)
    con20 = mesh_data["mesh"].topology()(2,0)
    for face in face_markers.array().argsort():
        marker = face_markers.array()[face]

        # Skip internal faces
        if marker == 0:
            continue

        inds[marker].update(con20(face))

        # NOTE: If a face share vertices with another one these will be
        # NOTE: overwritten by the highest marker number
        vertex_markers.array()[con20(face)] = marker
        areas[marker] += Face(mesh, face).area()

    # Create numpy array of the indices
    for marker in inds.keys():
        inds[marker] = np.array(list(sorted(inds[marker])))
    
    mesh_data.markers.vertex.face = adict(func=vertex_markers, inds=inds)

    # Check for available cell mesh function
    cell_name = "%s%s%s"%(basename, cellmarker_complement, suffix)
    if os.path.isfile(cell_name):

        # Load cell markers
        debug("Loading CellMarkers")
        t=Timer("Load cell markers")
        try:
            cell_markers_double = MeshFunction("double", mesh, cell_name)
            cell_markers = CellFunction("uint", mesh)
            cell_markers.array()[:] = cell_markers_double.array()
        except:
            try:
                cell_markers = MeshFunction("uint", mesh, cell_name)
            except Exception, e:
                raise IOError, "could not open CellFunction:", cell_name

    else:
        # Cell markers not in own file check if contained in mesh
        cell_markers = mesh.domains().cell_domains(mesh)
        if cell_markers is None:
            raise IOError("No cell markers found. Either provide them in "\
                          "the Mesh file or as a cell marker file with name:"\
                          " '%s'." % cell_name)
        
        # Make max value equals 0
        # FIXME: Do not do this for CELL Functions
        #cell_markers.array()[\
        #    cell_markers.array() == cell_markers.array().max()] = 0
        

    distinct = list(sorted(set(cell_markers.array())))

    inds = dict([(m, (cell_markers.array()==m).nonzero()[0]) \
                 for m in distinct])
    mesh_data.markers.cell = adict(func=cell_markers, distinct=distinct, \
                                   inds=inds)
    
    # Initialize area structure
    volumes = {}.fromkeys(distinct, 0.0)
    mesh_data.volumes = volumes

    # Compute vertex markers belonging to given cells
    vertex_markers = VertexFunction("uint", mesh)
    vertex_markers.set_all(0)
    inds = dict((m, set()) for m in distinct)
    con30 = mesh_data.mesh.topology()(3,0)
    for cell in cell_markers.array().argsort():
        marker = cell_markers.array()[cell]
        inds[marker].update(con30(cell))

        # NOTE: If a cell share vertices with another one these will be
        # NOTE: overwritten by the highest marker number
        vertex_markers.array()[con30(cell)] = marker
        volumes[marker] += Cell(mesh, cell).volume()

    # Create an incidence dict for later checking if two domains share vertices
    incidence_markers = {}.fromkeys(distinct, [])
    markers = distinct[:]

    while markers:
        marker0 = markers.pop()
        for marker1 in markers:
            # If the two cell domains share vertices
            if inds[marker0].intersection(inds[marker1]):
                debug("domain with cell marker %d share vertices with domain"\
                      " with cell marker %d"%(marker0, marker1))
                incidence_markers[marker0].append(marker1)
                incidence_markers[marker1].append(marker0)
            
    # Create numpy array of the indices
    for marker in inds.keys():
        inds[marker] = np.array(list(sorted(inds[marker])))

    mesh_data.markers.vertex.cell = adict(func=vertex_markers, inds=inds, \
                                          incidence=incidence_markers)
    #else:
    #    volumes = {0:0.0}
    #    for cell in cells(mesh):
    #        volumes[0] += cell.volume()
    #    
    #    mesh_data.volumes = volumes
        
    return mesh_data

def load_and_check_mesh_data(subcellmodel):
    "Check if defined subcellmodel is contained in the mesh data"

    t = Timer("Load and checking of Mesh")
    assert(isinstance(subcellmodel, SubCellModel))
    if subcellmodel.num_mesh_domains() == 0:
        # Non mesh to be loaded!
        raise ValueError, "expected at least one MeshDomain"
    
    filename = subcellmodel.mesh_domains()[0].mesh.filename
    mesh_data = load_mesh_data(filename)
    
    distinct_domain_markers = mesh_data["markers"]["cell"]["distinct"][:]
    for domain in subcellmodel.mesh_domains():

        # Check cell domains
        if domain.marker not in distinct_domain_markers:
            # FIXME: This might not work in parrallell because some 
            # FIXME: markers might disappear.
            # FIXME: Make some 'or' function over the processes
            raise ValueError, "marker '%d' is not registered as a cell "\
                  "marker in '%s'"%\
                  (domain.marker, domain.mesh.filename)
        
        # Pop distinct domain marker if any left after checking
        # domains raise error
        distinct_domain_markers.remove(domain.marker)
        
        # Create a vertex mesh functions with vertices maching the cell mumbers
        cell_vertex_markers = set(mesh_data["markers"]["vertex"]
                                  ["cell"]["inds"][domain.marker])
        
        # Check boundary domains
        for boundary in domain.boundaries:
            
            # Iterate over all face markers a boundary has
            for face_marker in boundary.domain_markers(domain):
                
                # Check if it excists in the mesh data
                if face_marker not in mesh_data["markers"]["face"]["distinct"]:
                    raise ValueError, "marker '%d' is not registered as a "\
                          "face marker in '%s'"%\
                          (face_marker, subcellmodel.meshes[0].filename)
                
                # Check that the vertex indices in each boundary is contained
                # in the vertex indices for the cells
                face_vertex_markers = set(mesh_data["markers"]["vertex"]
                                          ["face"]["inds"][face_marker])
                if len(face_vertex_markers.difference(cell_vertex_markers))\
                       != 0:
                    raise ValueError, "Boundary marker '%d' is not contained "\
                          "in the '%s' cell domain."%(face_marker, domain.name)

    # Check for remaining distinct domain markers
    if distinct_domain_markers:
        raise ValueError, "expected all registered domains in loaded mesh to "\
              "be used in declared SubCell model. Not in use: %s"%\
              (", ".join("%d"%m for m in distinct_domain_markers))
    
    return mesh_data

