__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-05-16 -- 2012-07-23"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["SubCellNewtonSolver", "SubCellProblem", "to_parameterdict", \
           "zero_integral"]

# System imports
from dolfin import *
import ufl

# SubCell imports
from subcell.common import *
from subcell.common.logger import *
from subcell.backends.dolfinversion import dolfin_version


# Initiate a zero (Empty) form
zero_integral = ufl.Form([])

def zero_forms(*forms):
    "Help function to determine if forms are zero"
    return not any(bool(form.integrals()) for form in forms)

def to_parameterdict(param, exclude=None):
    """Convert a dolfin parameters to a ParametersDict"""
    assert(isinstance(param, Parameters))
    ret = {}
    if exclude is None:
        exclude = []
    p = ParameterDict()
    for key, values in param.iterdata():
        
        if hasattr(p, key) or key in exclude:
            continue
        if isinstance(values, Parameters):
            ret[key] = to_parameterdict(values)
        else:
            (value, value_range, access_count, change_count) = values
            if isinstance(value, str) and value_range is not None:
                ret[key] = OptionCheck(value, value_range)
            elif isinstance(value, (float, int)) and value_range is not None:
                ret[key] = RangeCheck(value, ge=value_range[0], \
                                      le=value_range[1])
            else:
                ret[key] = TypeCheck(value)
    
    return ParameterDict(**ret)

class SubCellNewtonSolver(object):
    "A place holder for a dolfin.NewtonSolver"
    def __init__(self):
        accepted_krylov_solvers = ["default", "cg", "gmres", "minres",
                                   "tfqmr", "richardson", "bicgstab"]
        
        accepted_lu_solvers = ["default", "umfpack", "mumps", "spooles",
                               "superlu_dist", "superlu"]
        
        accepted_preconditioners = ["default", "none", "additive_schwarz",
                                    "ilu", "bjacobi", "jacobi", "sor", "icc",
                                    "hypre_amg", "hypre_euclid",
                                    "hypre_parasails", "hypre_pilut",
                                    "amg_hypre", "ml_amg"]
        
        krylov_solver = KrylovSolver("bicgstab","ml_amg")
        lu_solver = LUSolver("default")
        newton_solver = NewtonSolver("cg")
        self.linear_solver = None
        self.newton_solver = None

        self.params = ParameterDict(**dict(\
            krylov_solver = OptionCheck("bicgstab", accepted_krylov_solvers),
            lu_solver = OptionCheck("default", accepted_lu_solvers),
            preconditioner = OptionCheck("ml_amg", accepted_preconditioners),
            type = OptionCheck("iterative", ["iterative", "direct"]),
            lu_params = to_parameterdict(lu_solver.parameters),
            krylov_params = to_parameterdict(krylov_solver.parameters),
            **to_parameterdict(newton_solver.parameters)))

        # Set maximum krylov solver iterations
        self.params.krylov_params.maximum_iterations = 10000
        
        # Default to no error on convergence
        self.params.error_on_nonconvergence = False
        
    def reset(self):
        "Reset the solver based on stored parameters"
        # Create the linear solver based on the stored parameters
        if self.params.type == "iterative":
            self.linear_solver = KrylovSolver(self.params.krylov_solver,
                                              self.params.preconditioner)
        
            # Update krylov solvers parameters
            self.linear_solver.parameters.update(\
                self.params.krylov_params.copy(True))
        else:
            self.linear_solver = LUSolver(self.params.lu_solver)

            # Update lu solvers parameters
            self.linear_solver.parameters.update(\
                self.params.lu_params.copy(True))
        
        # Create a Newton solver and assign parameters
        if dolfin_version == 1.0:
            self.newton_solver = NewtonSolver(self.linear_solver,
                                              EpetraFactory_instance())
        else:
            self.newton_solver = NewtonSolver(self.linear_solver,
                                              EpetraFactory.instance())

        # Update the NewtonSolver parameters from stored parameters
        for key in self.newton_solver.parameters.keys():
            self.newton_solver.parameters[key] = self.params[key]

    def solve(self, problem, vector):
        "Solve the coupled reaction diffusion problem for a timestep dt"
        assert(isinstance(problem, SubCellProblem))

        # Initiate matrices before solving
        problem.before_solve()

        # Call newton solvers solve and return result
        return self.newton_solver.solve(problem, vector)
        
class SubCellProblem(NonlinearProblem):
    "The fully coupled non-linear problem"
    def __init__(self, tensor_handler):
        from subcell.backends.dolfinbackend import DolfinBackend
        NonlinearProblem.__init__(self)
        self.tensor_handler = tensor_handler
        domain_handler = tensor_handler.domain_handler
        integral_handler = tensor_handler.integral_handler
        function_handler = integral_handler.function_handler
        
        # Get tensors
        # Jacobian matrix which is built for each Newton iteration
        self.A = tensor_handler.matrices["A"]
        
        # Part of Jacobian tensor which is constant for each time step
        self.A0 = tensor_handler.matrices["A"].copy()
        
        # The Stiffness and Mass matrices are fixed for the duration of
        # the simulation
        self.K = tensor_handler.matrices["K"]
        self.M = tensor_handler.matrices["M"]
        
        # A tensor that contains the contributions from all active
        # constant boundaries
        self.active_constant_flux_tensor = \
                        tensor_handler.matrices["active constant fluxes"]

        # Some relevant vectors
        #self.b = tensor_handler.vectors["b"]
        self.tmp = function_handler.vectors["tmp"]
        self.u0 = function_handler.vectors["u0"]

        # The flux forms
        self.flux_forms = integral_handler.flux_forms

        # The nonlinear forms
        #self.F_form = integral_handler.forms["F_K"]
        #self.F_form_transient = integral_handler.forms["F_K_transient"]
        self.forms = integral_handler.forms

        # Assembler function
        self.assemble = integral_handler.assemble

        # Set flags for the time solution
        self.dt = function_handler.dt
        self.transient = True

        # Get help method for assemble kwargs 
        self.assemble_kwargs = domain_handler.assemble_kwargs

        # Get list of active constant boundaries
        self.active_constant_boundaries = \
                                domain_handler.active_constant_boundaries

        # Get the active boundary markers
        self.active_boundary_markers = domain_handler.active_boundary_markers
        
        # Flag for using reactions
        # FIXME: Fix another way to do this, more parameterized...
        self.use_reactions = integral_handler.use_reactions
    
    def F(self, b, x):
        "Computes the Non-linear part for Newton iteration"
        
        # Check size of rhs
        if len(b) != len(x):
            b.resize(len(x))

        # Zero out tensor
        b.zero()
        
        # If transient
        if self.transient:
            
            # Assemble the transient part of the nonlinear form
            timer = Timer("Assemble of runtime tensors")
            self.assemble(self.forms["real F_K_transient"], b,\
                          active_domains=True, real=True, add_values=True)
            self.assemble(self.forms["F_K_transient"], b,\
                          active_domains=True, add_values=True)
            return

        # Assemble the non transient part of the nonlinear form
        timer = Timer("Assemble of runtime tensors")
        self.assemble(self.forms["real F_K"], b, active_domains=True, \
                      add_values=True, real=True)
        self.assemble(self.forms["F_K"], b, active_domains=True, add_values=True)

    def J(self, A, x):
        "Computes the Jacobian for each Newton iteration"

        # Assign the constant part
        A.assign(self.A0)

        # If we have a contribution from the Nonlinear part
        if self.use_reactions or (bool(self.active_boundary_markers) and \
                not zero_forms(self.flux_forms["real reassemble J"],
                               self.flux_forms["reassemble J"])):

            debug("Reassemble J")
            # Assemble into the A matrix and use the active boundary
            # mesh function
            timer = Timer("Assemble of runtime tensors")
            self.assemble(self.flux_forms["real reassemble J"], A, \
                          add_values=True, active_domains=True, real=True)
            self.assemble(self.flux_forms["reassemble J"], A, \
                          active_domains=True, add_values=True)

    def before_solve(self):
        "Called once for each solve. Initiate a Jacobian matrix"
        
        # Zero the matrix
        self.A0.zero()
        
        # If transient add mass matrix
        if self.transient:
            self.A0.axpy(1/self.dt.value, self.M, True)

        # Add the stiffness matrix
        self.A0.axpy(1, self.K, True)

        if self.active_boundary_markers:
            debug("Active boundaries: %s", self.active_boundary_markers)

        # Add the constant flux matrix if we have any active constant
        # boundaries
        if self.active_constant_boundaries:
            debug("Adding active constant boundaries: %s",
                  ", ".join("%s[%d]"%active for active \
                            in self.active_constant_boundaries))
            self.A0.axpy(1, self.active_constant_flux_tensor, True)
        
        # If we have a contribution from the semi constant fluxes
        if self.use_reactions or (bool(self.active_boundary_markers) and \
            not zero_forms(self.flux_forms["real reassemble once J"],
                       self.flux_forms["reassemble once J"])):
            
            # Assemble into the self.A0 matrix and use the active boundary
            # mesh function
            debug("Reassemble J once")
            timer = Timer("Assemble of runtime tensors")
            self.assemble(self.flux_forms["real reassemble once J"],\
                          self.A0, add_values=True, active_domains=True, \
                          real=True)
            self.assemble(self.flux_forms["reassemble once J"], \
                          self.A0, active_domains=True, add_values=True)
        
    def set_time(self, dt, transient=True):
        "Update the time for the Non linear problem"
        if dt <= 0:
            raise ValueError, "provide a positive dt"
        self.dt.value = dt
        self.transient = transient
        debug("Setting time step: %.3f, transient: %d", dt, transient)

