__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-11-07 -- 2011-08-15"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["GillstepBackend"]

# System import
from gillstep import GillespieStepper, jit, load_model

# subcell imports
from subcell.common import *
from subcell.common.logger import *
from subcell.model.discretemarkovmodel import *
from subcell.model.variables import *

# local imports
from bases import StochasticBackend

class GillstepBackend(StochasticBackend):
    "A gillstep specific backend to solve the discrete and stochastic problem"
    def __init__(self, subcellmodel):
        
        # Init base class
        StochasticBackend.__init__(self, subcellmodel)
        
        # Initiate the GillespiStepper
        self.stepper = GillespieStepper()
        
        # Storage for the DiscreteMarkovModels
        self.discrete_markov_models = {}
        
        # Instantiate DiscreteMarkovModels and check receivability
        self._instantiate_markov_models(subcellmodel)
        
        # Establish connection from Variables to the DiscreteMarkovModels
        #self._establish_variable_connections(subcellmodel)
        
        # Make the GillespieStepper time available
        self.time = self.stepper.time
        
        # Reset only time
        self.reset_time = self.stepper.reset_time
        
        # Create parameters
        self.params = self._extract_parameters()

    def _instantiate_markov_models(self, subcellmodel):
        """
        Instantiate the DiscreteMarkovModels and store then in the
        SubCellModel"""
        
        def check_receivability(dmm_inst, variables):
            # Check that the MarkovModel can receive a variables 
            # (Not to be confused with Variable)
            for variable in variables:
                if not dmm_inst.can_receive(variable):
                    raise ValueError, "The DiscreteMarkovModel '%s' "\
                          "cannot receive the variable '%s' given from "\
                          "the subcellmodel"%(dmm.name, variable)
        
        def receiver_closure(instances):
            def receiver(var, num):
                assert(num<len(instances))
                return instances[num].receive(var)
            return receiver
        
        def receivers_closure(instances):
            def iter_receiver(var):
                return (instance.receive(var) for instance in instances)
            return iter_receiver
        
        def connector_closure(instances):
            def connector(num):
                assert(num<len(instances))
                return instances[num].connect
            return connector

        # Instantiate the DiscreteMarkovModels
        for dmm in subcellmodel.discrete_markov_models:
            assert(dmm.num>0)
            instances = listwrap(jit(load_model(dmm.name), self.stepper, \
                                     dmm.num))
            self.discrete_markov_models[dmm.name] = instances
            dmm.instances = instances
            
            # Attach callbacks
            dmm.receivers_callback = receivers_closure(instances)
            dmm.receiver_callback = receiver_closure(instances)
            dmm.connector_callback = connector_closure(instances)
            
            # Get an instance
            dmm_inst = instances[0]
            check_receivability(dmm_inst, dmm.variables)
            
            for name, submodel in dmm.submodels.iteritems():
                submodel.instances = []
                
                # Get number of submodels in instantiated MarkovModel
                gillstep_subnum = len(dmm_inst.submodels(name))
                if gillstep_subnum!=submodel.subnum:
                    raise ValueError, "The number of the submodel '%s' in the"\
                          " DiscreteMarkovModel '%s' does not match the one "\
                          "given in the subcellmodel: %d!=%d"%(name, dmm.name,
                                                               gillstep_subnum,
                                                               submodel.subnum)
                check_receivability(dmm_inst.submodels(name)[0],
                                    submodel.variables)
                
                # Collect submodels
                for dmm_inst in instances:
                    submodel.instances.extend(dmm_inst.submodels(name))
                
                # Attach callbacks
                submodel.receivers_callback = \
                         receivers_closure(submodel.instances)
                submodel.receiver_callback = \
                         receiver_closure(submodel.instances)
                submodel.connector_callback = \
                         connector_closure(submodel.instances)
    
    def _establish_variable_connections(self, subcellmodel):
        "Establish connections from variables to receiver"
        
    def _extract_parameters(self):
        "Extract parameters from the MarkovModels"
        dmms_params = {}
        for dmm in self.subcellmodel.discrete_markov_models:
            dmm_params = {}
            for smname, submodel in dmm.submodels.iteritems():
                subinst = submodel.instances[0]
                sub_dmm_params = dict((name, \
                            TypeCheck(subinst.get_parameter(name)))\
                            for name in subinst.get_parameter_names())
                sub_dmm_params["init_state_values"] = \
                                        TypeCheck(submodel._init_state_values)
                sub_dmm_params["fixed_state_values"] = \
                                        TypeCheck(submodel._fixed_state_values)
                dmm_params[smname] = ParameterDict(**sub_dmm_params)
            inst = dmm.instances[0]
            dmm_params.update(dict((name, TypeCheck(inst.get_parameter(name)))
                                   for name in inst.get_parameter_names()))
            dmm_params["init_state_values"] = TypeCheck(dmm._init_state_values)
            dmm_params["fixed_state_values"] = \
                                             TypeCheck(dmm._fixed_state_values)
            dmms_params[dmm.name] = ParameterDict(**dmm_params)
        
        # Construct the stochastic backend ParameterDict
        return ParameterDict(\
            substep=TypeCheck(True),
            max_fractional_rate_change=RangeCheck(0.1, gt=0, lt=1), 
            collect_values=TypeCheck(False),
            collect_states=TypeCheck(False),
            models=ParameterDict(**dmms_params),
            )

    def init_state_values(self):
        "Apply any user specific init state values"

        for dmm in self.subcellmodel.discrete_markov_models:
            for smname, submodel in dmm.submodels.iteritems():

                # Default all state to -1
                for subinst in submodel.instances:
                    subinst.init_state = -1
                
                # Set any prescribed init state values
                for index, value in submodel._init_state_values.iteritems():
                    if index >= len(submodel.instances):
                        raise IndexError, "index: '%d' too large in "
                    "'init_state_values' for '%s'"%(index, smname)
                    submodel.instances[index].init_state = value

            # Default all state to -1
            for inst in dmm.instances:
                inst.init_state = -1
                
            # Set any prescribed init state values
            for index, value in dmm._init_state_values.iteritems():
                if index >= len(dmm.instances):
                    raise IndexError, "index: '%d' too large in "\
                "'init_state_values' for '%s'"%(index, dmm.name)
                dmm.instances[index].init_state = value

    def apply_parameters(self):
        # Use the stored parameters to set the MarkovModels parameter
        dmms_params = self.params["models"]
        for dmm in self.subcellmodel.discrete_markov_models:
            for smname, submodel in dmm.submodels.iteritems():

                # Get Submodel instance
                subinst = submodel.instances[0]
                for parname in subinst.get_parameter_names():
                    subinst.set_parameter(parname, dmms_params[dmm.name]\
                                          [smname][parname])
                submodel._init_state_values = \
                    dmms_params[dmm.name][smname]["init_state_values"]
                submodel._fixed_state_values = \
                    dmms_params[dmm.name][smname]["fixed_state_values"]
                
                # Set any prescribed init state values
                for index, value in submodel._init_state_values.iteritems():
                    if index >= len(submodel.instances):
                        raise IndexError, "index: '%d' too large in "\
                              "'init_state_values' for '%s'"%(index, smname)
                    submodel.instances[index].init_state = value
                    debug("Init %s[%d] state's to %d", smname, index, value)
                
                # Set any prescribed fixed state values
                for index in submodel._fixed_state_values:
                    if index >= len(submodel.instances):
                        raise IndexError, "index: '%d' too large in "\
                              "'fixed_state_values' for '%s'"%(index, smname)
                    submodel.instances[index].fix_state = True
                    debug("Fixing %s[%d] state", smname, index)
            
            # Get MarkovModel instance
            inst = dmm.instances[0]
            for parname in inst.get_parameter_names():
                inst.set_parameter(parname, dmms_params[dmm.name][parname])

            dmm._init_state_values = \
                    dmms_params[dmm.name]["init_state_values"]
            dmm._fixed_state_values = \
                    dmms_params[dmm.name]["fixed_state_values"]
            
            # Set any prescribed init state values
            for index, value in dmm._init_state_values.iteritems():
                if index >= len(dmm.instances):
                    raise IndexError, "index: '%d' too large in "\
                "'init_state_values' for '%s'"%(index, dmm.name)
                dmm.instances[index].init_state = value
                debug("Init %s[%d] state's to %d", dmm, index, value)

            
            # Set any prescribed fixed state values
            for index in dmm._fixed_state_values:
                if index >= len(dmm.instances):
                    raise IndexError, "index: '%d' too large in "\
                "'fixed_state_values' for '%s'"%(index, dmm.name)
                dmm.instances[index].fix_state = True
                debug("Fixing %s[%d] state", dmm, index)
                
        # Set save state/value parameters
        self.stepper.collect_values = self.params.collect_values
        self.stepper.collect_states = self.params.collect_states
        self.stepper.set_max_fractional_rate_change(\
            self.params.max_fractional_rate_change)

    def reset_before_set_of_runs(self):
        """
        Reset the stochastic backend before a set of runs
        """

        # Apply parameters
        self.apply_parameters()
        
    def reset_before_single_run(self, time=0.0):
        """
        Reset the stochastic backend before a single run
        """

        # Reset the MarkovModels and send any updated value to receivers
        self.stepper.reset(time)

    def step_dt(self, dt):
        """
        Step all MarkovModels and return any possible offset
        """
        debug("Next_transition: %.3f", self.stepper.next_transition())
        return self.stepper.step_dt_unless_transition(dt, \
                                     self.params["substep"])

    def get_data_after_single_run(self):
        """
        Collect any data and return them
        """
        
        # If no collection
        if not any([self.params.collect_values, self.params.collect_states]):
            return
        
        collect = []

        if self.params.collect_values:
            collect.extend(["values", "value_times"])

        if self.params.collect_states:
            collect.extend(["states", "state_times"])
        
        debug("Collecting %s from Gillstepper", quote_join(collect))
        
        value_getters = dict((value_type,\
                              getattr(self.stepper, "get_%s"%value_type))\
                             for value_type in collect)
        
        data = {}
        
        for dmm in self.subcellmodel.discrete_markov_models:
            data[str(dmm)] = {}
            for value_type in collect:
                data[str(dmm)][value_type] = []
                for inst in dmm.instances:
                    data[str(dmm)][value_type].append(\
                        value_getters[value_type](inst))
            
            for smname, submodel in dmm.submodels.iteritems():
                data[smname] = {}
                for value_type in collect:
                    data[smname][value_type] = []
                    for subinst in submodel.instances:
                        data[smname][value_type].append(\
                            value_getters[value_type](subinst))
        return data

    def set_minimize_dt_callback(self, func):
        "Connect Markov models to minimize dt callback function"
        assert(callable(func))
        for dmm in self.subcellmodel.discrete_markov_models:
            if dmm.minimize_dt:
                for instance in dmm.instances:
                    instance.connect(func)

    def set_end_sim_callback(self, func):
        "Connect Markov models to end sim callback function"
        assert(callable(func))
        for dmm in self.subcellmodel.discrete_markov_models:
            if dmm.end_sim:
                for instance in dmm.instances:
                    instance.connect(func)


