__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-06-09 -- 2011-09-20"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["ReactionSolver"]

# System imports
from dolfin import Vector
from numpy import array, arange
import sympy

# Subcell imports
from subcell.common import *
from subcell.backends.dolfincompilereaction import compile_reaction
from subcell.backends.reactionnewtonsolversnippets import *
from subcell.model.species import *

class SingleReactionSolver(object):
    """
    Class that holds everything needed to solve a single reaction
    """
    def __init__(self, reactions, function_handler, newton_params):
        self.function_handler = function_handler
        self.reactions = reactions
        self.domain = self.reactions[0].domain
        self.species = odict()
        self.equations = odict()
        self.newton_params = newton_params
        
        # A map between the indices for the field species in global and
        # local vectors
        self.global_value_inds = odict()
        self.local_value_inds = odict()

        # A dict holding references to the global vectors keeping the solutions
        self.global_vecs = odict()
        self.global_vecs["u"] = odict()
        self.global_vecs["u0"] = odict()
        
        # Parameters includes parameters, variables and species that are not
        # involved as reactants or products
        self.parameters = odict()
        self.parameters["field_species"] = []
        self.parameters["scalar_species"] = []
        self.parameters["const_species"] = []
        self.parameters["variables"] = []
        self.parameters["params"] = []
        self.parameters["node_params"] = []

        # Analyse all reactions and extract info
        self.extract_reaction_info()
        
        # Create code!
        code = self.create_code()
        
        # Compile code
        self.solver = compile_reaction(code)(self.num_single_dofs)
        
        # Update parameter and initial variables parameters
        self.update_params()
        for param in self.solver.parameters:
            if param in self.newton_params:
                self.solver.parameters[param] = self.newton_params[param]

    def extract_reaction_info(self):
        # Collect information of the participating reactions
        for reaction in self.reactions:
            
            # Extend parameters
            self.parameters["field_species"].extend(\
                reaction.scalar_field_species)
            self.parameters["field_species"].extend(\
                reaction.fixed_scalar_field_species)
            self.parameters["scalar_species"].extend(reaction.scalar_species)
            self.parameters["const_species"].extend(reaction.const_species)
            self.parameters["variables"].extend(reaction.variables)
            self.parameters["params"].extend(\
                param for param in reaction.parameters \
                if param.param.is_scalar)
            self.parameters["node_params"].extend(\
                param for param in reaction.parameters if \
                param.param.is_node_param)
            
            # Grab information about reactants and products
            for species in reaction.reactants.keys()+reaction.products.keys():
                if not isinstance(self.domain.species[species], FieldSpecies):
                    raise TypeError, "expected all species involved as "\
                          "products or reactants in a reaction to be "\
                          "FieldSpecies"
                
                # Get domains species
                domain_species = self.domain.species[species]
                
                # Remove any registered parameters which is either in
                # products or reactants
                if domain_species in self.parameters["field_species"]:
                    self.parameters["field_species"].remove(domain_species)
                    
                # Check if we have already registered the species
                if species in self.species:
                    continue
                
                # Initialize equation
                self.equations[species] = 0
                
                # Add Field species together with indice into global dofs
                self.species[species] = domain_species
                self.global_value_inds[species] = domain_species.value_inds

                # Get vectors
                if isinstance(domain_species, FixedScalarFieldSpecies):
                    self.global_vecs["u"][species] = self.function_handler.\
                                                     vectors["fixed"]\
                                                     [str(domain_species)]["u"]
                    self.global_vecs["u0"][species] = self.function_handler.\
                                                vectors["fixed"]\
                                                [str(domain_species)]["u0"]
                else:
                    self.global_vecs["u"][species] = self.function_handler.\
                                                     vectors["u"]
                    self.global_vecs["u0"][species] = self.function_handler.\
                                                      vectors["u0"]
            
            # Grab equations
            for species, stoich in reaction.reactants.iteritems():
                self.equations[species] -= stoich*reaction.expr()
            
            for species, stoich in reaction.products.iteritems():
                self.equations[species] += stoich*reaction.expr()

        # Gather all parameters
        self.parameters["all"] = self.parameters["field_species"] + \
                                 self.parameters["scalar_species"] + \
                                 self.parameters["const_species"] + \
                                 self.parameters["variables"] + \
                                 self.parameters["params"] + \
                                 self.parameters["node_params"]
        
        # Get number of involved species and parameters
        self.num_species = len(self.global_value_inds)
        self.num_params = len(self.parameters["all"])
        
        # Get length of single dof vector
        self.num_single_dofs = len(self.global_value_inds.values()[0])
        if not all(self.num_single_dofs == len(inds) \
                   for inds in self.global_value_inds.values()):
            raise RuntimeError("expected the same number of degrees of "\
                               "freedom for each FieldSpecies used in the "\
                               "reaction. This is probably a bug as it "\
                               "should not happen.")
        
        # Get indices into local vector
        for i, species in enumerate(self.species.keys()):
            self.local_value_inds[species] = arange(\
                i*self.num_single_dofs, (i+1)*self.num_single_dofs, dtype="I")
        
        # Get length of all dof vectors
        self.num_total_dofs = self.num_single_dofs*self.num_species
        
        # Create the solution vectors
        self.u = Vector(self.num_total_dofs)
        self.u0 = Vector(self.num_total_dofs)

    def create_subs(self):
        """
        Create a substitution dict for the equations
        """
        subs = {}

        # Create subs for species
        for i, species in enumerate(self.species.values()):
            #print i, species.species
            subs[species] = sympy.Symbol("u[%d]"%i)
        
        # Create subs for params
        for i, param in enumerate(self.parameters["all"]):
            subs[param] = sympy.Symbol("params[%d]"%i)

        return subs
        
    def create_code(self):
        """
        Create code for the single reaction
        """
        subs = self.create_subs()
        variables = [subs[species] for species in self.species.values()]
        #print variables

        # Create symbols for equations definitions
        dt = sympy.Symbol("dt")
        init_species = odict((species, sympy.Symbol("u0[%d]"%i)) \
                             for i, species in enumerate(self.species.keys()))
        
        # Add time dependent part of equation
        for species in self.species.keys():
            self.equations[species] = self.equations[species].subs(subs)
            self.equations[species] -= (subs[self.species[species]]-\
                                        init_species[species])/dt
        
        # Get ready to role residual functions
        Fu = self.equations.values()

        # Create jacobian
        Ju = [Fi.diff(var) for Fi in Fu for var in variables]
        
        # Start fill the form
        reaction_form["SYSTEMSIZE"] = str(self.num_species)
        reaction_form["NUMPARAMS"] = str(self.num_params)

        # Generate param name map
        param_map = ["      _param_map[\"%s\"] = %d;"%(str(param), i) \
                     for i, param in enumerate(self.parameters["all"])]
        reaction_form["PARAMNAMES"] = "\n".join(param_map)
        
        # If we have a small system create a solver using direct solver
        debug("")
        debug("Code generation")
        if self.num_species <= 4:
            solve_code = []
            N = self.num_species
            f_code = ["F[%d] = %s;"%(i, Fi) for i, Fi in enumerate(Fu)]
            
            JuMat = sympy.Matrix(N,N, lambda i, j: \
                    sympy.Symbol("J%d%d"%(i, j)) if Ju[i*N+j] !=0 else 0)
            FuVec = sympy.Matrix(N,1, lambda i, j: \
                                 sympy.Symbol("F[%d]"%i))

            # Calculate the determinant (used to simplify the code)
            detJuMat = JuMat.det()
            detJuMatSub = sympy.Symbol("detJuMat")
            for i in xrange(N):
                for j in xrange(N):
                    if Ju[i*N+j] !=0:
                        solve_code.append("const double J%d%d = %s;"%\
                                          (i, j, str(Ju[i*N+j])))
            solve_code.append("const double detJuMat = %s;"%str(detJuMat))

            # Inverse the Jacobian
            JuMatInv = JuMat.inverse_ADJ()
            JuMatInv = JuMatInv.expand().subs({detJuMat:detJuMatSub})

            # Write the solution
            for i, value in enumerate(JuMatInv*FuVec):
                solve_code.append("dx[%d] = %s;"%\
                                  (i, str(value.subs(\
                                      {detJuMatSub:sympy.Symbol(\
                                          "detJuMat")}))))
            debug("\n".join(solve_code))
            debug("\n".join(f_code))
            reaction_form["SOLVE_U"] = "\n".join(" "*6+line \
                                                 for line in solve_code)
            reaction_form["COMPUTE_F"] = "\n".join(" "*6+line \
                                                   for line in f_code)
            reaction_form["PRIVATEFUNCS"] = \
                            direct_snippets["PRIVATEFUNCS"]%reaction_form
            
            reaction_form["SOLVE"] = direct_snippets["SOLVE"]%reaction_form
            reaction_form["SOLVE_ONCE"] = \
                                direct_snippets["SOLVE_ONCE"]%reaction_form
        else:
            jacobian_code = ["J.at(%d,%d) = %s;"%\
                             (i/self.num_species, i%self.num_species, Ji)\
                             for i, Ji in enumerate(Ju)]
            f_code = ["F.at(%d,0) = %s;"%(i, Fi) for i, Fi in enumerate(Fu)]
            
            debug("\n".join(jacobian_code))
            debug("\n".join(f_code))

            reaction_form["COMPUTE_JACOBIAN"] = "\n".join(" "*6+line \
                                                   for line in jacobian_code)
            reaction_form["COMPUTE_F"] = "\n".join(" "*6+line \
                                                   for line in f_code)
            reaction_form["PRIVATEFUNCS"] = \
                            nondirect_snippets["PRIVATEFUNCS"]%reaction_form
            reaction_form["SOLVE"] = nondirect_snippets["SOLVE"]%reaction_form

        # Create the code
        return common_snippet%reaction_form
        
        # Generate parameters out of all Parameters and none
        # receiving species
        # Add ParameterVector!
        # Generate a symbol namespace based on the string repr of the
        # FieldSpecies
        # Create symbols for the involved species, including u and u0
        # Count involved species not more than 4-5 use symbolic
        # inversion of Jacobian!
        # Generate F and J
        # Symbolicly solve J*dx=F
        # Populate code formats and generate code!
        # Compile!
            
    def step(self, dt):
        """
        Step solution
        """
        
        # Update paramters that might vary each time step
        self.update_variable_params()
        
        for species, fixed_species in self.species.iteritems():

            # Init solution vector
            self.u[self.local_value_inds[species]] = \
                                        self.global_vecs["u"][species]\
                                        [self.global_value_inds[species]]
            
        # Step the solver
        self.solver.solve(dt, self.u)
        
        # Put back solution
        for species, fixed_species in self.species.iteritems():
            
            # Put back solution vector
            self.global_vecs["u"][species][self.global_value_inds[species]] = \
                        self.u[self.local_value_inds[species]]
    
    def step_once(self, dt):
        """
        Step solution
        """
        
        for species, fixed_species in self.species.iteritems():

            # Init solution vector
            self.u[self.local_value_inds[species]] = \
                                        self.global_vecs["u"][species]\
                                        [self.global_value_inds[species]]

        # Step the solver
        self.solver.solve_once(dt, self.u0, self.u)
        
        # Put back solution
        for species, fixed_species in self.species.iteritems():
            
            # Put back solution vector
            self.global_vecs["u"][species][self.global_value_inds[species]] = \
                        self.u[self.local_value_inds[species]]
        
    def update_once_before_step_once(self):
        
        # Update paramters that might vary each time step
        self.update_variable_params()
        
        for species, fixed_species in self.species.iteritems():
            
            # Init initial value vector
            self.u0[self.local_value_inds[species]] = \
                                            self.global_vecs["u0"][species]\
                                            [self.global_value_inds[species]]
            
    def update_params(self):
        """
        Update parameters in the model
        """
        for species in self.parameters["const_species"]:
            self.solver.set_model_param(str(species), species.init.get())
        
        for param in self.parameters["params"]:
            self.solver.set_model_param(str(param), param.param.get())
        
        for param in self.parameters["node_params"]:
            self.solver.set_model_vector_param(str(param), param.param.get())

    def update_variable_params(self):
        """
        Update parameters in the model
        """
        # FIXME: Check that these works...
        # FIXME: They do not...
        for species in self.parameters["field_species"]:
            debug("Updating field species: '%s'", str(species))
            values = species.ufl_func["u"].vector()[species.value_inds]
            self.solver.set_model_vector_param(str(species),
                                               values.array())
        for species in self.parameters["scalar_species"]:
            debug("Updating scalar species: '%s'", str(species))
            value = species.ufl_func["u"].vector()[species.value_inds]
            self.solver.set_model_vector_param(str(species), value)

        for variable in self.parameters["variables"]:
            debug("Updating variable: '%s'", str(variable))
            self.solver.set_model_vector_param(str(variable), variable.value)
        
class ReactionSolver(object):
    """
    Class that holds everything needed to solve an operator splitted
    reaction of FieldSpecies
    """
    def __init__(self, function_handler, runtime_params, newton_params):

        self.function_handler = function_handler
        self.runtime_params = runtime_params
        
        domain_reactions = odict()
        for reaction in function_handler.subcellmodel.reactions:
            # If the reaction does not only include field species
            if not reaction.only_field_species:
                continue
            if reaction.domain not in domain_reactions:
                domain_reactions[reaction.domain] = [reaction]
            else:
                domain_reactions[reaction.domain].append(reaction)
            
            # If a static electric potential is defined
            if reaction.domain.static_electric_potential is not None:
                raise NotImplementedError(\
                    "reaction on domains with static electric potentials "\
                    "cannot be solved using a operator splitted approach. "\
                    "Either unset that option or use ScalarFieldSpecies "\
                    "instead of FixedScalarFieldSpecies")
        
        self.reaction_solvers = [
            SingleReactionSolver(reactions, function_handler, newton_params)\
            for reactions in domain_reactions.itervalues()]
        
            
    def step(self, dt):
        """
        Step the reaction solver one time step
        """
        debug("Stepping reaction solver with dt: %g ms", dt)
        for reaction_solver in self.reaction_solvers:
            reaction_solver.step(dt)
        
        self.function_handler.update_solutions_after_solve()

    def step_once(self, dt):
        """
        Step the reaction solver one Newton iteration
        """
        debug("Solve one Newton iteration of reaction solver "\
              "with dt: %g ms", dt)
        for reaction_solver in self.reaction_solvers:
            reaction_solver.step_once(dt)
        
    def update_once_before_step_once(self):
        for reaction_solver in self.reaction_solvers:
            reaction_solver.update_once_before_step_once()
