__author__ = "Johan Hake <hake.dev@gmail.com>"
__date__ = "2008-06-22 -- 2011-10-21"
__copyright__ = "Copyright (C) 2008-2010 " + __author__
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["TypeCheck", "RangeCheck", "OptionCheck", "ParConstant", \
           "inf", "VALUE_JUST", "ParSlave", "ParSymbol", "ArrayParam", \
           "exchange_symbols_with_ParSlaves"]

# System imports
import numpy as np
from numpy import inf
import sympy
import types
import operator
from string import ljust, rjust, center

# SubCell imports
from subcell.common.logger import *

VALUE_JUST = rjust

def value_formatter(value, str_length = 0):
    " Return a formated string of a value"
    ret = None
    if isinstance(value, (np.ndarray, list)) and len(value)>4:
        if isinstance(value[0], int):
            formatstr = "[%d, %d, ..., %d, %d]"
        elif isinstance(value[0], float):
            formatstr = "[%.2g, %.2g, ..., %.2g, %.2g]"
        else:
            formatstr = "[%s, %s, ..., %s, %s]"
        ret = formatstr%(value[0], value[1], value[-2], value[-1])
    
    if isinstance(value, float):
        if value == inf:
            ret = "\xe2\x88\x9e"
        elif value == -inf:
            ret = "-\xe2\x88\x9e"
    elif isinstance(value, str):
        ret = repr(value)
    if ret is None:
        ret = str(value)
    if str_length == 0:
        return ret
    return VALUE_JUST(ret, str_length)

scalars = (int, float)
option_types = scalars + (str,)

def exchange_symbols_with_ParSlaves(expr):
    """
    Substitution dict to remove all generated sympy.Symbols which
    does not play with ParSymbols
    """
    d = {}
    for atom in expr.atoms():
        if str(atom) in ParSlave._all_objects:
            d[atom] = ParSlave._all_objects[str(atom)].sym
    return expr.subs(d)

def extract_original_expression(expr):
    """
    Subsitution dict to remove all parslaves with the symbolic representation 
    of them
    """
    d = {}
    for atom in expr.atoms():
        if str(atom) in ParSlave._all_objects:
            d[atom] = ParSlave._all_objects[str(atom)].symbols
    return expr.subs(d)

def parsymbolify_expression(expr):
    """
    If the expression only consists of ParSymbols then we make a
    ParSlave out of it
    """
    if all(isinstance(atom, ParSymbol) for atom in expr.atoms()):
        count = TypeCheck._combined_count
        TypeCheck._combined_count += 1
        return ParSlave(expr, "combined_%d"%count)
    return expr

class ParSymbol(sympy.Symbol):
    "A sympy Symbol holding a TypeChecker"
    _par_symbol_count = 0
    def __new__(cls, param, symname=None):
        if not isinstance(param, TypeCheck):
            raise TypeError, "expected a TypeCheck when initalizing ParSymbol"
        #if isinstance(param, ParSlave):
        #    return param.sym
        if symname is None:
            name = "%s%d"%(param.name, ParSymbol._par_symbol_count) \
                   if param.name is not None else \
                   "p%d"%ParSymbol._par_symbol_count
        else:
            if not isinstance(symname, str):
                raise TypeError, "expected a str for symname"
            name = symname
        
        self = sympy.Symbol.__new__(cls, name)
        self.param = param
        ParSymbol._par_symbol_count += 1
        return self

    def name_subs(self):
        "Return a tuple of the sym and the name of the parameter"
        return (self, sympy.Symbol(str(self)) if self.param.name \
                is None else sympy.Symbol(self.param.name))

    def symb_subs(self):
        "Return a tuple of the sym and the name of the parameter"
        return (self, sympy.Symbol(str(self)))

    def value(self):
        "Return a tuple of the sym and the value of the parameter"
        if not (self.param.is_scalar or self.param.is_array):
            raise TypeError, "expected the parameter type to be a scalar "\
                  "or array"
        return self.param.get()

    def repr_subs(self):
        "Return a tuple of the sym and the repr of the parameter"
        return (self, sympy.Symbol(repr(self.param)))

class TypeCheck(object):
    _combined_count = 0
    "A simple type checking class for a single value"
    def __init__(self, value, name=None, symname=None):
        if not isinstance(name, (types.NoneType, str)):
            raise TypeError, "expected a 'str' for the name argument"
        self._value = value
        self.value_type = type(value)
        self._repr_str = "TypeCheck(%s)"
        self.is_array = isinstance(value, np.ndarray)
        self.is_scalar = isinstance(value, scalars)
        self._not_in_str = None
        self._in_str = None
        self._in_range = None
        self._name = name
        self.sym = ParSymbol(self, symname)

        # Init some flags
        self._used = False
        self.is_node_param = False
        self.is_cell_param = False
        self.domain = None
        
    def _get_name(self):
        return self._name
    
    def _set_name(self, name):
        "Set name"
        symname=None
        if isinstance(name, tuple):
            assert(len(name)==2)
            assert(all(isinstance(n, str) for n in name))
            name, symname = name
            
        if not isinstance(name, str):
            raise TypeError("expected a str for the name argument")

        if self._name:
            raise AttributeError("Cannot set name attribute of Parameter, "\
                                 "it is already set to '%s'"%self._name)
        self._name = name
        
        # Create a new symbol
        self.sym = ParSymbol(self, symname)
    
    name = property(_get_name, _set_name)
    
    def _algebra_op(self, other, op, exchange=False):
        "A generic method handling all algebraic methods"
        for arg in [self, other]:
            if not (isinstance(arg, (int, float, bool, sympy.Basic)) or \
                        (isinstance(arg, TypeCheck) and \
                         (arg.is_scalar or arg.is_array))):
                raise TypeError("algebraic operation only possible between "\
                                "scalars, scalar or array parameters, "\
                                "or sympy value")
        
        # Mark the parameter as used
        self._used = True

        # Other is a scalar value
        if isinstance(other, (int, float, bool)):
            args = (other, self.sym) if exchange else (self.sym, other)
            count = TypeCheck._combined_count
            TypeCheck._combined_count += 1
            return ParSlave(op(*args), "combined_%d"%count)

        # Other is a TypeCheck, grab the symbolic representation of it
        if isinstance(other, (TypeCheck, ParSymbol)):
            other._used = True
            args = (other.sym, self.sym) if exchange else (self.sym, other.sym)
            
            # If both are arrays we need to check compatability
            if self.is_array and other.is_array:

                # If not same domain, or not the same param type
                # just return the operator without wrapping in ParSlave
                if not(self.domain == other.domain) or \
                       (self.is_node_param != other.is_node_param) or \
                       (self.is_cell_param != other.is_cell_param):
                    return op(*args)
            
            count = TypeCheck._combined_count
            TypeCheck._combined_count += 1
            return ParSlave(op(*args), "combined_%d"%count)

        # Exchange any Symbols generated through SymPy magic back to
        # ParSlaves
        other = exchange_symbols_with_ParSlaves(other)
        
        # Other is a sympy.Basic but consists of only ParSymbols
        # For example the result of exp(parameter)
        parameter_types = (int, float, sympy.Number, ParSymbol)
        if all(isinstance(atom, parameter_types) for atom in other.atoms()):
            args = (other, self.sym) if exchange else (self.sym, other)
            count = TypeCheck._combined_count
            TypeCheck._combined_count += 1
            return ParSlave(op(*args), "combined_%d"%count)

        args = (other, self.sym) if exchange else (self.sym, other)
        return op(*args)
    
    def __neg__(self, other):
        return self._algebra_op(other, operator.neg)

    def __pow__(self, other):
        return self._algebra_op(other, operator.pow)

    def __mul__(self, other):
        return self._algebra_op(other, operator.mul)

    def __add__(self, other):
        return self._algebra_op(other, operator.add)

    def __sub__(self, other):
        return self._algebra_op(other, operator.sub)

    def __div__(self, other):
        return self._algebra_op(other, operator.div)

    def __rpow__(self, other):
        return self._algebra_op(other, operator.pow, True)

    def __rmul__(self, other):
        return self._algebra_op(other, operator.mul, True)

    def __radd__(self, other):
        return self._algebra_op(other, operator.add, True)

    def __rsub__(self, other):
        return self._algebra_op(other, operator.sub, True)

    def __rdiv__(self, other):
        return self._algebra_op(other, operator.div, True)

    def __neg__(self):
        return self._algebra_op(-1, operator.mul, True)

    @property
    def used(self):
        "Return true if the parameter has been used in an expression"
        return self._used

    def setvalue(self, value):
        "Try to set the value using the check"
        self._value = self.check(value)
        
    def get(self):
        "Return the value"
        return self._value
    
    def check(self, value):
        "Check the value using the type and any range check"
        # Fist check if the value is an int or float.
        # (Treat these independently)
        name_str = "" if self.name is None else \
                   " for the parameter '%s'"%self.name
        if self.value_type in [int, float] and \
               isinstance(value, (int, float)):
            if isinstance(value, float) and self.value_type == int:
                info("Converting %s to %d%s", str(value), int(value), \
                     name_str)
            value = self.value_type(value)
        if self.value_type in [bool] and isinstance(value, int) and \
               not isinstance(value, bool):
            info("Converting %s to '%s'%s", str(value), \
                 "True" if value else "False", name_str)
            value = self.value_type(value)
            
        if not isinstance(value, self.value_type):
            raise TypeError, "Please provide a '%s'%s"%\
                  (self.value_type.__name__, name_str)
        if self._in_range is not None:
            if not self._in_range(value):
                raise ValueError, "Illegal value%s: %s"%\
                      (name_str, self.format_data(value, True))
        return value
    
    def format_data(self, value = None, not_in = False, str_length = 0):
        "Print a nice formated version of the value and its range"
        if value is None:
            value = self._value

        # If no '_in_str' is defined
        if self._in_str is None:
            return value_formatter(self._value, str_length)
        
        if not_in:
            return self._not_in_str%(value_formatter(value, str_length))
        else:
            return self._in_str%(value_formatter(value, str_length))

    def length_value_format(self):
        "Return the length of the formated str of value"
        return len(value_formatter(self._value))

    def __repr__(self):
        "Print a executable version of the Check"
        return self._repr_str%value_formatter(self._value)
    
    def __str__(self):
        "Print a nice representation of the value and Check"
        return self.format_data()

class ParConstant(TypeCheck):
    " A Constant parameter which prevent any change of values" 
    def __init__(self, value, name=None, symname=None):
        TypeCheck.__init__(self, value, name, symname)
        self._repr_str = "ParConstant(%s)"
        
        # Define a 'check function'
        self._in_range = lambda x:x==self._value
        
        # Define some string used for pretty print
        self._in_str = "%s - Constant"
        self._not_in_str = "%%s != %s"%str(self._value)
        self.setvalue(value)

class ParSlave(TypeCheck):
    "A slave parameter defined by other parameters"
    _all_objects = {}
    def __init__(self, value, name=None):
        if not isinstance(value, (TypeCheck, sympy.Basic)):
            raise TypeError, "expected a scalar, or expression of "\
                  "other parameters"
        if isinstance(value, TypeCheck):
            value = value.sym

        # Exchange any Symbols generated through SymPy magic back to
        # ParSlaves
        value = exchange_symbols_with_ParSlaves(value)
            
        if not all(isinstance(atom, (sympy.Number, ParSymbol))\
                   for atom in value.atoms()):
            raise TypeError, "expected expression of other parameters"
        TypeCheck.__init__(self, value, name, symname=name)

        # Store the original expression used to evaluate the value of
        # the ParSlave
        self.symbols = extract_original_expression(value)
        self.is_array = any(isinstance(atom.param, ArrayParam) for atom in \
                            self.iter_par_symbols())
        
        # If the parameter is an array we need to check that all array
        # parameters are of the same legal type
        if self.is_array:
            from subcell.model.species import NodeParam, CellParam
            all_array = [atom.param for atom in self.iter_par_symbols() \
                          if isinstance(atom.param, ArrayParam)]
            self.is_node_param = all(param.is_node_param \
                                     for param in all_array)
            self.is_cell_param = all(param.is_cell_param \
                                     for param in all_array)
            if not (self.is_node_param or self.is_cell_param):
                raise TypeError("expected all ArrayParams in an expression"\
                                " to be of either 'CellParam' or "\
                                "'NodeParam'")
            if not all(all_array[0].domain == param.domain \
                       for param in all_array):
                raise TypeError("expected all '%s' in an expression "\
                                "to be defined on the same domain"%\
                                ("NodeParam" if self.is_node_param \
                                 else "CellParam"))
            
            # Assign the domain the ParSlave is defined on
            self.domain = all_array[0].domain
            
        # If the parameter is not array it is scalar
        self.is_scalar = not self.is_array
        self._repr_str = "ParSlave(%s)"

        # Store object
        ParSlave._all_objects[str(self)] = self
        
    def check(self, value):
        "A check function which always fails"
        raise TypeError, "Cannot assign a value to a 'ParSlave'"

    def _str_repr(self, op="repr"):
        assert(op in ["name", "repr", "symb"])
        return str(self.sym.subs(dict(eval("atom.%s_subs()"%op)\
                                      for atom in self.sym.atoms()\
                                      if isinstance(atom, ParSymbol))))
    def __str__(self):
        return self._str_repr("name")
    
    def __repr__(self):
        return self._str_repr("name")

    def iter_par_symbols(self):
        """
        Return an iterator over all original parameter symbols of the
        expression
        """
        for atom in self.symbols.atoms():
            if not isinstance(atom, ParSymbol):
                continue
            if isinstance(atom.param, ParSlave):
                for atom in atom.param.iter_par_symbols():
                    yield atom
            yield atom
                
    def get(self):
        """
        Return a computed value of the Parameters
        """
        par_symbols = [atom for atom in self.iter_par_symbols()]
        
        # Create name space which the expression will be evaluated in
        ns = dict((str(sym), sym.value()) for sym in par_symbols)
        ns.update(np.__dict__)
        all_length = [len(sym.param.get()) for sym in par_symbols if \
                      sym.param.is_array]
        same_length = all(all_length[0] == l for l in all_length)
        if not same_length:
            raise RuntimeError, "expected all ArrayParams in an expression "\
                  "to be of equal size."

        return eval(str(self.symbols), globals(), ns)
    
    def format_data(self, value=None, not_in=False, str_length=0):
        "Print a nice formated version of the value and its range"

        # If no '_in_str' is defined
        return "%s - ParSlave(%s)"%(value_formatter(self.get(), str_length), \
                                    self._str_repr("symb"))

class RangeCheck(TypeCheck):
    """
    A simple type and range checking class for a single value

    Example:
    ========
    
    >>> sigma = RangeCheck(5, ge=0, le=10)
    >>> sigma_slave = sigma*5

    Here is sigma_slave an immutable parameter, which can only change when
    sigma is changed.

    @type value : scalar
    @param value : The initial value of the RangeCheck parameter
    @type ge : scalar
    @param ge : Defines the lower closed limit of the check (greater or
    equal than).
    @type gt : scalar
    @param gt : Defines the lower open limit of the check (greater than).
    @type le : scalar
    @param le : Defines the upper closed limit of the check (lesser or
    equal than).
    @type lt : scalar
    @param lt : Defines the upper open limit of the check (lesser than).

    If none of C{gt} or C{ge} are set, are -\inf set as the lower
    limit. Likewise, if none of C{lt} or C{le} are set, are \inf
    set as the upper limit.

    """
    def __init__(self, value, ge=None, le=None, gt=None, lt=None, \
                 name=None, symname=None):
        TypeCheck.__init__(self, value, name, symname)
        
        # If no range is provided
        if ge is None and le is None and gt is None and lt is None:
            raise TypeError, "Provide at least one range"
        
        # Checking valid combinations of kwargs
        if not ge is None and not gt is None:
            raise TypeError, "Cannot create a 'RangeCheck' including "\
                  "both 'ge' and 'gt'"
        if not le is None and not lt is None:
            raise TypeError, "Cannot create a 'RangeCheck' including "\
                  "both 'le' and 'lt'"

        # Checking valid types for 'range checks'
        wrong_range_check_type = False
        for arg in [value, ge, le, gt, lt]:
            if not arg is None:
                wrong_range_check_type |= not isinstance(arg, scalars)
                
        if wrong_range_check_type:
            raise TypeError, "Can only provide 'range check' for"\
                  " 'int' or 'float'"
        
        # Pick the min and max values
        if not ge is None:
            min_val = ge
        elif not gt is None:
            min_val = gt
        else:
            min_val = -inf
        
        if not le is None:
            max_val = le
        elif not lt is None:
            max_val = lt
        else:
            max_val = inf

        # Check the relation between the min and max value
        if min_val > max_val:
            raise ValueError, "Please provide a 'max value' that is "\
                  "larger than the 'min value'."
        
        range_check_dict = {}

        # Dict for test and repr
        range_check_dict["min_op"] = ">=" if gt is None else ">"
        range_check_dict["max_op"] = "<=" if lt is None else "<"
        range_check_dict["min_value"] = str(min_val)
        range_check_dict["max_value"] = str(max_val)
        
        # Dict for pretty print
        range_check_dict["min_op_format"] = "[" if gt is None else "("
        range_check_dict["max_op_format"] = "]" if lt is None else ")"
        range_check_dict["min_format"] = value_formatter(min_val)
        range_check_dict["max_format"] = value_formatter(max_val)

        # Define a 'check function'
        self._in_range = eval(("lambda value:value %(min_op)s %(min_value)s "\
                               "and value %(max_op)s %(max_value)s")%\
                              range_check_dict)

        # Define some string used for pretty print
        self._in_str = ("%%s \xe2\x88\x88 %(min_op_format)s%(min_format)s, "\
                        "%(max_format)s%(max_op_format)s")%range_check_dict
        
        self._not_in_str = ("%%s \xe2\x88\x89 %(min_op_format)s%(min_format)s"\
                            ", %(max_format)s%(max_op_format)s")%\
                            range_check_dict

        # Define a 'repr string'
        op_list = [ge, gt, le, lt]
        op_str_list = ["ge", "gt", "le", "lt"]
        self._repr_str = "RangeCheck(%%s, %s)"%(", ".join(\
            "%s=%s"%(op_str_list[i], str(v)) \
            for i, v in enumerate(op_list) if not v is None))
        
        # Set the value using the check functionality
        self.setvalue(value)
    
class OptionCheck(TypeCheck):
    " A simple type and options checking class for a single value"
    def __init__(self, value, options, name=None):
        TypeCheck.__init__(self, value, name)
        
        if not isinstance(options, list):
            raise TypeError, "Provide a 'list' with the valid options."
        
        # Check valid types for an 'option check'
        if not isinstance(value, option_types) or isinstance(value, bool):
            raise TypeError, ("Can only provide 'option check' for 'str', "\
                              " 'int' and 'float', not for '%s'")%\
                              (type(value).__name__)
        
        # Check that all values in options has the same type
        for val in options:
            if not isinstance(val, self.value_type):
                raise TypeError, ("All values of the 'option check' " +\
                      "need to be of type: '%s'")%(type(value).__name__)
        
        # Define a 'check function'
        self._in_range = eval("lambda value:value in %s"%repr(options))
        
        # Define some string used for pretty print
        self._in_str = "%%s \xe2\x88\x88 %s"%repr(options)
        
        self._not_in_str = "%%s \xe2\x88\x89 %s"%repr(options)
        
        # Define a 'repr string'
        self._repr_str = "OptionCheck(%%s, %s)"%repr(options)
        
        # Set the value using the check functionality
        self.setvalue(value)
    
class ArrayParam(TypeCheck):
    """
    A numpy Array based parameter
    """
    def __init__(self, value, size=None, name=None, symname=None):
        if size is not None:
            # If a size is provided a scalar is expected for the value
            if not isinstance(value, scalars):
                raise TypeError("expected a scalar as the first argument")
            if not isinstance(size, int) and size<=0:
                raise TypeError("expected a positive int as the second "\
                                "argument")
            value = np.array([value]*size, dtype="d" \
                             if isinstance(value, float) \
                             else "i")
        
        # Checking Array argument
        if isinstance(value, np.ndarray):
            if value.dtype.char == "f":
                value = value.astype("d")
            elif value.dtype.char in ["I", "u", "b"]:
                value = value.astype("i")
            elif value.dtype.char not in ["i", "d"]:
                raise TypeError("expected dtype of passed NumPy array to "\
                                "be a scalar")
        elif isinstance(value, scalars):
            value = np.array([value])
        else:
            raise TypeError("expected a scalar or NumPy array as the first "\
                            "argument")
        
        TypeCheck.__init__(self, value, name, symname)

        # Define a 'repr string'
        self._repr_str = "ArrayParam(%s)"

    def setvalue(self, value):
        """
        Set value of ArrayParameter
        """
        if isinstance(value, np.ndarray):
            if len(value) != len(self._value):
                raise TypeError, "expected the passed array to be of "\
                      "size: '%d'"%len(self._value)
            self._value[:] = value
        
        # Set the whole array to value
        elif isinstance(value, scalars):
            self._value[:] = value
        elif isinstance(value, tuple) and len(value)==2:
            if not(isinstance(value[0], int) and \
                   isinstance(value[1], scalars)):
                raise TypeError("expected a size 2 tuple of an int and"\
                                " a scalar")
            self._value[value[0]] = value[1]
        else:
            raise TypeError("expected a scalar, numpy array or a size "\
                            "2 tuple of an int and a scalar")
        
    def resize(self, newsize):
        """
        Change the size of the Array
        """
        if not isinstance(newsize, int):
            raise TypeError("expected newsize argument to be an int")
        if len(self._value) != newsize:
            self._value = np.resize(self._value, newsize)
    
# Test code
if __name__ == "__main__":
    x = TypeCheck(3)
    print x
    print repr(x)
    x.setvalue(5)
    print x
    
    v = RangeCheck(5, 0, 10)
    print v
    print repr(v)
    
    u = OptionCheck("a", ["a", "b", "c", "d"])
    print u
    print repr(u)

    a = ArrayParam(5.,10)
    print a
    print repr(a)
    a.setvalue((0, 6.))
    print a
    a.setvalue(7.)
    print a
    a.setvalue(np.arange(10))
    print a
    a.resize(20)
    print repr(a)
    for atom in (v*x+v*a).atoms():
        print atom.param
    print v*x+v*a
    p = ParSlave(v*x+v*a)
    print p.get()
    print p
    print repr(p)
    a1 = ArrayParam(5.,20)
    print ParSlave(a1*p).get()

    x = ArrayParam(1.)
    y = ArrayParam(1.)
    z = ArrayParam(1.)
    n = ArrayParam(1.)
    
    p = ParSlave(sympy.exp(-z))
    print p
    
    print p.get()
    z.resize(11)
    z.setvalue(np.linspace(0,1,11))
    print p.get()
