__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-16 -- 2011-10-21"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["MeshBoundary", "Boundary", "BoundaryValue", "BoundaryBase", \
           "mesh_boundaries", "none_mesh_boundaries"]

# System imports
import sympy, types

# SubCell imports
from subcell.common import *
from subcellmodelobject import *
from species import *
from domains import *
from fluxes import *
from discretemarkovmodel import *
from variables import *
from connection import Receiver

class BoundaryValue(SubCellModelObject):
    __count = 0
    def __init__(self, species, boundary, domain, ind):
        check_arg(species, Species, 0)
        check_arg(boundary, MeshBoundary, 1)
        check_arg(domain, MeshDomain, 2)
        check_arg(ind, int, 3)
        name = "%s_at_%s%d_in_%s_%d"%(species, boundary, ind, domain, \
                                      BoundaryValue.__count)
        SubCellModelObject.__init__(self, name)
        self.species = species
        self.boundary = boundary
        self.domain = domain
        self.marker = boundary.domain_markers(domain)[ind]
        self.ind = ind
        BoundaryValue.__count += 1

    def __repr__(self):
        return "BoundaryValue(%s, %s, %s, %d)"%(repr(self.species),
                                                repr(self.boundary),
                                                repr(self.domain),
                                                self.ind)

class BoundaryBase(SubCellModelObject):
    def __init__(self, name, domains, dx=1.0, num=1, params=None):
        "Initiate a Boundary"
        params = to_odict(params)
        check_arg(name, str, 0)
        check_arg(domains, (Domain, tuple), 1)
        check_arg(num, int, 2)
        domains = tuplewrap(domains)
        params["num"] = create_parameter(num, "num", ge=1)
        params["dx"] = create_parameter(dx, "dx", gt=0,\
                                        property_param=False, \
                                        symname="%sdx"%name)
        
        # All fluxes deactivated by default
        self.active = [False]*num

        # Reset activation callback
        self.activation_callback = None
        
        SubCellModelObject.__init__(self, name, params)
        
        self._init_domain_info(domains)

        # Register the boundary to the domains
        for domain in domains:
            domain.boundaries.append(self)
        
        self.isgated = False
        self.domains = domains
        self.fluxes = odict()
        self.boundary_values = odict()

    def __repr__(self):
        return "%s('%s', %s%s, %s)"%(type(self).__name__, self.name,\
                                 repr(self.domains), self._param_repr())
        
    # FIXME: Implement limits!
    def _add_flux(self, flux):
        "Add a flux to the Boundary"
        assert(isinstance(flux, Flux))
        
        # Check that any provided species are defined on the domain
        for domain in self.domains:
            if flux.species not in domain.species:
                raise TypeError, "The species '%s' is not defined on the "\
                      "domain: '%s' which is registered in the boundary: "\
                      "'%s'"%(flux.species, domain, self)
            if isinstance(domain.species[flux.species], ScalarFieldSpecies)\
                   and domain.species[flux.species].sigma.get()==0.0:
                raise TypeError, "A flux cannot be added when the diffusion "\
                      "constant of one of the involved ScalarFieldsSpecies is"\
                      " zero: '%s.sigma' == 0"%(domain.species[species])
            
        # Instantiate Flux and register it
        if flux.species not in self.fluxes:
            self.fluxes[flux.species] = [flux]
        else:
            self.fluxes[flux.species].append(flux)
        
        # Generate boundary values based on the used fieldspecies in the flux
        self._generate_boundary_values(flux)

    def _generate_boundary_values(self, flux):
        """
        Generates boundary values for the registered 
        """
        assert(flux.boundary == self)
        
        # Return flux does not connect two FieldSpecies
        if flux.num_scalar_field_species_domains() < 2:
            return
        
        # Iterate over the different instances of the Boundary
        for ind in xrange(self.num):
            # Iterate over the ScalarFieldSpecies and create BoundaryValues
            # for each one of them
            for sfs in flux.scalar_field_species:
                if (sfs, ind) in self.boundary_values:
                    continue
                
                # Create the BoundaryValue
                boundary_value = BoundaryValue(sfs.species, flux.boundary,\
                                               sfs.domain, ind)
                
                # Store the BoundaryValueat the Boundary
                self.boundary_values[(sfs, ind)] = boundary_value

    def _init_domain_info(self, domains):
        "Init domain info"
        self.num_domains = len(domains)
        if self.num_domains > 2:
            raise ValueError, "expected at most two domains"
        self.num_mesh_domains = sum(isinstance(domain, MeshDomain)
                                    for domain in domains)
        self.num_volume_domains = sum(isinstance(domain, VolumeDomain)
                                      for domain in domains)
        self.num_infinite_domains = sum(isinstance(domain, InfiniteDomain)
                                        for domain in domains)
        if self.num_domains == self.num_infinite_domains:
            raise ValueError, "cannot create a Boundary using only "\
                  "'InfiniteDomains'"

    def _receiver(self):
        "Return a Reciver version of the Boundary"
        self.isgated = True
        return Receiver(self)

    def reset_activation(self, value=False):
        "Set the state of the activation of all boundaries to value"
        self.active = [value]*self.num
        
    def receiver(self, ind=0):
        """
        Return the receiver method of Boundary 'ind'

        Usage:
        ------
        >>> sender.connect(flux.receiver(0))
        """
        check_arg(ind, int, 0)
        if ind<0:
            raise ValueError, "expected a positive number for the 'ind' "\
                  "argument"
        if self.num == 1 and ind != 0:
            raise ValueError, "expected '0' for the 'ind' argument because "\
                  "there are only one Boundary"
        if ind>=self.num:
            raise ValueError, "expected 'ind' to be smaller than the number "\
                  "of boundaries in '%s'"%self.boundary.name

        def receive(value):
            """
            A receive function that change the active flag of the Boundary 
            'ind'
            
            The implementation of this function is to call the activate method 
            on all the registered fluxes. The activate method just call a 
            callback function which a particular backend can implement.
            """

            # If not status is changed, return
            if self.active[ind] == bool(value):
                return

            # Change status
            self.active[ind] = bool(value)

            # If no activate callback is register leave
            if self.activation_callback is None:
                warning("*** Warning: No activation callback registered"\
                        " for '%s'"%self.name)
                return 
            
            # Call activate callback function (backend specific)
            self.activation_callback(ind, value)
            
        return receive

class Boundary(BoundaryBase):
    "A Boundary with no MeshDomains"
    def __init__(self, name, domains, area=1.0, dx=1.0, **params):
        domains = tuplewrap(domains)
        try:
            check_arg(domains, tuple, 1, (VolumeDomain, InfiniteDomain))
        except:
            raise TypeError, "Boundary: expected instances of "\
                  "'VolumeDomain' or 'InfiniteDomain' for the 'domain' "\
                  "argument, got: %s"%\
                  (", ".join("'%s'"%(type(domain).__name__) \
                             for domain in domains))
        params = user_defined_parameters(params, symnameprefix=name)
        params["area"] = create_parameter(area, "area", gt=0,\
                                          property_param=False, \
                                          symname="%sarea"%name)
        BoundaryBase.__init__(self, name, domains, dx, 1, params)

class MeshBoundary(BoundaryBase):
    "A Boundary with at least one MeshDomain"
    def __init__(self, name, domains, startmarkers, dx=1.0, num=1, **params):
        domains = tuplewrap(domains)
        if not any(isinstance(domain, MeshDomain) for domain in domains):
            raise TypeError, "MeshBoundary: expected at least one instance of"\
                  "'MeshDomain'for the 'domain' argument, got: %s"%\
                  (", ".join("'%s'"%(type(domain).__name__) \
                             for domain in domains))
        
        if not isinstance(startmarkers, (int, tuple)):
            raise TypeError("Expected an int or a tuple of two ints "\
                            "for the 'start_marker' argument")
        
        startmarkers = tuplewrap(startmarkers)
        check_arg(startmarkers, tuple, 2, int)
        
        # FIXME: Allow ArrayParam in the future?
        params = user_defined_parameters(params, symnameprefix=name)
        
        # FIXME: Remove possibilities to access areas from a subcellmodel file
        ## Create a hidden parameter
        #params["__area"] = create_parameter(1.0, gt=0)
        #
        ## The hidden parameter slaves the exposed one
        #params["area"] = create_parameter(params["__area"][0], gt=0)
        
        BoundaryBase.__init__(self, name, domains, dx, num, params)
        self.startmarkers = _init_markers(self, startmarkers, num)
        
        # Create a placeholder for any Species at the MeshBoundary
        for domain in mesh_domains(domains):
            for scalar_field_species in domain.scalar_field_species:
                setattr(scalar_field_species, name, \
                    MeshBoundarySpecies(scalar_field_species, self))
        
    def __repr__(self):
        return "MeshBoundary('%s', %s, %s%s)"%(self.name,\
                        repr(self.domains), repr(self.startmarkers),\
                        self._param_repr())

    def domain_markers(self, domain):
        "Return the markers of the passed MeshDomain"
        check_arg(domain, MeshDomain, 0)
        if domain not in self.domains:
            raise ValueError, "the domain is not a domain the Boundary "\
                  "connects."
        if not self.num_mesh_domains > 0:
            raise TypeError, "boundary does not connect mesh domains and "\
                  "hence no marker"
        if len(self.markers) == 0:
            return self.markers
        if len(self.markers) == 1:
            return self.markers[0]
        return self.markers[self.domains.index(domain)]
    
    def marker_at_domain(self, marker):
        check_arg(marker, int, 0)
        if self.num_mesh_domains == 0:
            raise ValueError, "Boundary '%s' has no MeshDomains"%self.name
        elif self.num_mesh_domains == 1:
            return mesh_domains(self.domains)[0]
        if marker in self.markers[0]:
            return self.domains[0]
        if marker in self.markers[1]:
            return self.domains[1]
        raise ValueError, "marker %d is not a marker at any domains at "\
              "Boundary '%s'"%(marker, self.name)

def _init_markers(boundary, startmarkers, num):

    # Wrap any passed startmarkers to a tuple
    startmarkers = startmarkers if len(startmarkers) > 0 else (None,)

    # Sanity check for num and area
    if num < 1:
        raise ValueError, "expected a positive value for the 'num' argument"
    
    # Check length of given startmarkers
    if len(startmarkers) > 2 :
        raise ValueError, "expected a list or tuple of maximal length of 2"
    
    # Store boundary type
    boundary.mesh_boundary = True

    if len(startmarkers) != boundary.num_mesh_domains:
        raise ValueError, "expected the same number of startmarkers as"\
              " MeshDomains"
    
    # Check that all start markers are larger than 0
    if not all(startmarker>0 for startmarker in startmarkers):
        raise ValueError, "expected a positive number for the startmarkers"
    
    # Create a list of list of markers related to the MeshDomains
    boundary.markers = [range(startmarkers[i], startmarkers[i]+num)
                        for i, marker in enumerate(startmarkers)]
    
    return startmarkers

def mesh_boundaries(boundaries):
    """
    Return an iterator over the MeshBoundaries in the passed iterator
    """
    return (boundary for boundary in boundaries \
            if isinstance(boundary, MeshBoundary))

def none_mesh_boundaries(boundaries):
    """
    Return an iterator over none MeshBoundaries in the passed iterator
    """
    return (boundary for boundary in boundaries \
            if not isinstance(boundary, MeshBoundary))
    
