__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-11-18 -- 2011-11-02"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["ValueClamp", "EventStopper",  "OneTimeValueCrosser"]

# System imports
import sympy
import numpy as np

# subcell imports
from subcell.common import *
from variables import *

class TimeVariable(ContinuousVariable):
    """
    A Continuous variable representing the present time step
    """
    __num_instances = 0
    def __new__(cls, name=None):
        # FIXME: Dummy name argument....
        if TimeVariable.__num_instances > 0:
            raise RuntimeError("Only one instance of 'TimeVariable' "\
                               "can be created")
        self = ContinuousVariable.__new__(cls, "time")
        TimeVariable.__num_instances += 1
        return self

    def update(self, dt, time):
        """
        Assign the value variables the present time
        """
        self.value = time

class DtTimeVariable(ContinuousVariable):
    """
    A Continuous variable representing the present time step
    """
    __num_instances = 0
    def __new__(cls, name=None):
        # FIXME: Dummy name argument....
        if DtTimeVariable.__num_instances > 0:
            raise RuntimeError("Only one instance of 'DtTimeVariable' "\
                               "can be created")
        self = ContinuousVariable.__new__(cls, "dt")
        DtTimeVariable.__num_instances += 1
        return self

    def update(self, dt, time):
        """
        Assign the value variables from the present dt
        """
        self.value = dt

class TStop(DiscreteVariable):
    """
    A discrete variable taking care of the default stop time
    """
    def __new__(cls, name=None):
        # Initialize base class
        self = DiscreteVariable.__new__(cls, "tstop", True)
        
        self.time_params = None
        self.value = 0.0 
        return self

    def set_timeparams(self, time_params):
        """
        Set timestepper before reseting
        """
        self.time_params = time_params

    @property
    def next_time_step(self):
        assert(self.value > self.time())
        next_dt = self.value - self.time()
        # A fix for numerical impressions
        # FIXME: Parameterize the cutoff level
        if next_dt < 0.0001:
            return 0.0
        return next_dt

    def reset(self):
        assert(self.time_params)
        self.value = self.time_params.tstop

class EventStopper(DiscreteVariable):
    """
    A Discrete Variable that collect discrete events
    
    When no other discrete events has value 1 it will send a
    terminate signal
    """
    def __new__(cls, name="EventStopper"):
        
        # Initialize base class
        self = DiscreteVariable.__new__(cls, name, True)
        self.value = 0.0
        return self

    def reset(self):
        "Reset variable, called once for each simulation"
        self.value = 0

    def receive(self, value):
        if self.time() <= 0.0:
            self.value += value
        else:
            self.value += 1 if value > 0. else -1
            if self.value <= 0:
                self.value = 1.
                self.send()

    def _receiver(self):
        from connection import Receiver
        return Receiver(self)
    
    @property
    def next_time_step(self):
        if self.value > 0.0 or self.time() == 0.0:
            return float('inf')
        
        # Send a dummy timestep. Must be larger than 0
        return 0.1

class OneTimeValueCrosser(DiscreteVariable):
    """
    A Discrete Variable that recieves a continuous value and compare it
    
    When a continuous variable reach a certain cutoff value it will
    send an event
    """
    def __new__(cls, name, cutoff, send_values=[1,0], comp="<", \
                minimize_dt=False):
        check_arg(name, str, 0)
        check_kwarg(send_values, "send_values", list, scalar)
        check_kwarg(comp, "comp", str)
        
        if len(send_values) !=2:
            raise TypeError, "expected a list of length 2 for the "\
                  "'send_values' argument"
        valid_comp = ["<", "<=", ">", ">="]
        if comp not in valid_comp:
            raise TypeError, "expected the 'comp' variable to be one "\
                  "of %s"%(", ".join(valid_comp))
        
        params = odict()
        params["cutoff"] = create_parameter(cutoff, "cutoff")
        params["send_values"] = create_parameter(\
            TypeCheck(send_values), "send_values", only_scalar=False)
        params["comp"] = create_parameter(OptionCheck(comp, valid_comp),\
                                          "comp", only_scalar=False)
        
        # Initialize base class
        self = DiscreteVariable.__new__(cls, name, False, minimize_dt, params)

        self.collected_value = 0.0
        self.first_value_sent = False
        self.second_value_sent = False
        self._connected = False

        return self

    def reset(self):
        "Reset variable, called once for each simulation"
        self.first_value_sent = False
        self.second_value_sent = False
        self.value = self.send_values[0]
        ns = {}
        self.compare = eval("lambda value: value %s %g"%\
                            (self.comp, self.cutoff))
        
    def receive(self, value):
        "Recive the value and check if we will send the second value"
        debug("'%s' received %g which will be compared to %g", \
              self.name, value, self.cutoff)
        if self.first_value_sent and not self.second_value_sent and \
               self.compare(value):
            self.value = self.send_values[1]
            self.second_value_sent = True
            self.send()
            debug("Sending second value of '%s'"%self.name)
    
    def _receiver(self):
        from connection import Receiver
        if self._connected:
            raise TypeError, "'%s' can only be connected to by one sender"%\
                  self.__class__.__name__
        self._connected = True
        return Receiver(self)

    def step(self):
        self.first_value_sent = True
    
    @property
    def next_time_step(self):
        "Make sure the first value get sent at start"
        return float('inf') if self.first_value_sent else 0.0
            
    def __repr__(self):
        return "OneTimeValueCrosser('%s', %f, send_values=%s, , comp='%s')"%\
               (self.name, self.cutoff, repr(self.send_values), self.comp)
    
class ValueClamp(DiscreteVariable):
    "A value clamp discrete variable"
    def __new__(cls, name, times=None, value_list=None, minimize_dt=False):
        times = times or []
        value_list = value_list or []
        check_arg(name, str, 0)
        check_arg(times, list, 1, float)
        check_arg(value_list, list, 2, float)

        params = odict()
        params["value_list"] = create_parameter(\
            TypeCheck(value_list), "value_list", only_scalar=False)
        params["times"] = create_parameter(\
            TypeCheck(times), "times", only_scalar=False)
        
        # Initialize base class
        self = DiscreteVariable.__new__(cls, name, False, minimize_dt, params)

        # Check and store the times and value_list as numpyfied arrays
        self._np_times, self._np_value_list = _check_arguments(times,\
                                                               value_list)

        self.ind = 0

        return self

    def __repr__(self):
        return "ValueClamp('%s', %s, %s, end_sim=%s)"%(self.name, \
                                repr(self.times), repr(self.value_list),\
                                repr(self.end_sim))

    @property
    def next_time_step(self):
        # First check the bounds of the present time
        t = self.time()
        
        if self.ind >= len(self._np_times):
            return float('inf')
        return self._np_times[self.ind] - t
        
    def step(self):
        "Step the variable to the next value"
        assert(self.ind<len(self._np_value_list))
        self.value = self._np_value_list[self.ind]
        self.ind += 1
    
    def reset(self):
        "Reset the value clamp"
        # Turn stored parameter value_list into numpyfied arrays
        self._np_times, self._np_value_list = \
                        _check_arguments(self.times, self.value_list)

        # Reset ind and value
        self.ind = 0
        self.value = self._np_value_list[0]

    def update(self, dt):
        "Update the value based on the present time + dt"
        assert(isinstance(dt, float) and dt>0)
        t = self.time() + dt
        
        # First check the bounds of the present time
        if t >= self._np_times[-1]:
            self.value = self._np_value_list[-1]
            return
        if t <= self._np_times[0]:
            self.value = self._np_value_list[0]
            return
        
        # Get the index for the present time
        index = (self._np_times > t).argmax()-1
        
        # Assign value
        self.value = self._np_value_list[index]

def _check_arguments(times, value_list, start_times_with_zero=True):
    "A help function to check the arguments" 

    # First check that we have an iterable of numbers for the times
    # and value_list arguments
    try:
        times = np.asarray(times, dtype='d')
        if len(times)==0:
            raise TypeError
    except:
        raise TypeError, "expected an iterable of numbers with "\
              "length > 0, for the 'times' argument"
    
    try:
        value_list = np.asarray(value_list, dtype='d')
        if len(value_list)==0:
            raise TypeError
    except:
        raise TypeError, "expected an iterable of numbers with "\
              "length > 0, for the 'value_list' argument"

    # Check that we only have positive numbers in the times argument
    if any(times < 0.0):
        raise TypeError, "expected only positive numbers in the "\
              "'times' argument"

    # If we want times to start with zero, check that
    if  start_times_with_zero and (times[0] != 0.0):
        raise TypeError, "expected 'times' to start with 0"

    # Check for monotone serie of times
    if any(np.diff(times) <= 0):
        raise TypeError, "expected a monoton increasing sequence of "\
              "numbers for the 'times' argument."

    # Check equal length of the times and value_list arguments
    if len(times) != len(value_list):
        raise TypeError, "expected equal length of the 'times' and "\
              "'value_list' arguments."

    return times, value_list
