__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-10-18 -- 2011-08-15"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["Sender", "FieldSender", "Receiver", "Connection"]

# System imports
import types

# SubCell imports
from subcell.common import *

class Receiver(object):
    "A place holder for a receiver function"
    def __init__(self, receiver, variable=None):
        check_arg(variable, (str, types.NoneType), 1)
        self.receiver = receiver
        self.variable = variable

    def __repr__(self):
        return "Receiver(%s, %s)"%(repr(self.receiver), repr(self.variable))

class Sender(object):
    "A place holder for a sender object"
    def __init__(self, sender):
        self.sender = sender

    def __repr__(self):
        return "Sender(%s)"%(repr(self.sender))

class FieldSender(Sender):
    "A place holder for a field sender object"
    def __init__(self, sender, boundary):
        from boundaries import MeshBoundary
        Sender.__init__(self, sender)
        check_arg(boundary, MeshBoundary, 0)
        self.boundary = boundary
    
    def __repr__(self):
        return "FieldSender(%s, %s)"%(repr(self.sender), repr(self.boundary))

class Connection(object):
    "A place holder for a Connection"
    def __init__(self, sender, receiver):
        check_arg(sender, Sender, 0)
        check_arg(receiver, Receiver, 1)
        
        # Check if connetion is allowed
        if not connection_allowed(sender.sender, receiver.receiver):
            raise TypeError, "the connection between a '%s' and '%s' is "\
                  "not allowed"%(type(sender.sender).__name__,\
                                 type(receiver.receiver).__name__)
        self.sender = sender
        self.receiver = receiver
        self._hash = None

        # Check that the number of sender equals the number of receivers 
        # if we have a N-N connection
        # FIXME: Is this nessesary? Should we not check all this when the
        # FIXME: actuall connections are wired?
        if isinstance(sender, FieldSender):
            from discretemarkovmodel import DiscreteMarkovModel
            # If no receiver num is registered we need to figure out how
            # many we need
            if isinstance(receiver, DiscreteMarkovModel) and receiver.num == 0:
                # A SubModel is a slave of a DiscreteMarkovModel so the number
                # of receiver need to be dividable with the number of SubModels
                if isinstance(receiver, DiscreteMarkovSubModel):
                    
                    if sender.boundary.num % receiver.subnum != 0:
                        raise ValueError, "expected the number of senders"\
                              " to be dividable (with no rest) by the "\
                              "number of '%s' SubModels %d % %d != 0"%\
                              (receiver.name, sender.boundary.num,\
                               receiver.subnum)
                    
                    # Assign the number of parents
                    receiver.parent.num = sender.boundary.num / receiver.subnum
                else:
                    # If the receiver is a DiscreteMarkov model (no SubModel)
                    receiver.num = sender.boundary.num
            elif isinstance(receiver, DiscreteMarkovModel) and \
                     receiver.num != sender.boundary.num:
                raise ValueError, "expected the number of sender to be "\
                      "equal to the number of receivers: %d != %d"%\
                      (receiver.num, sender.boundary.num)
        
    def __repr__(self):
        return "Connection(%s, %s)"%(repr(self.sender), repr(self.receiver))

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(repr(self))
        return self._hash

    def __eq__(self, other):
        if isinstance(other, Connection):
            return hash(self) == hash(other)
        return False

# Define allowed receiveres and senders
def connection_allowed(sender, receiver):
    "Put this into a function avoiding problems with circular dependencies"
    from species import DomainSpecificSpecies, MeshBoundarySpecies
    from discretemarkovmodel import DiscreteMarkovModel
    from boundaries import BoundaryBase
    from variables import DiscreteVariable, Variable, ContinuousVariable

    allowed_connections = {
        DiscreteMarkovModel: (BoundaryBase, Variable),
        ContinuousVariable: (Variable, DiscreteMarkovModel),
        DiscreteVariable: (BoundaryBase, Variable, DiscreteMarkovModel),
        DomainSpecificSpecies: (Variable, DiscreteMarkovModel),
        MeshBoundarySpecies: (Variable, DiscreteMarkovModel),
        }

    for sender_type, receiver_types in allowed_connections.iteritems():
        if isinstance(sender, sender_type):
            return isinstance(receiver, receiver_types)
    return False
