__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-10-04 -- 2011-06-22"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["DiscreteMarkovModel", "DiscreteMarkovSubModel"]

# subcell imports
from subcell.common import *

# Local imports
from subcellmodelobject import *

#FIXME: Add end_sim parameter!
class DiscreteMarkovModel(SubCellModelObject):
    _disabled = False
    "A place holder for a discrete Markov model"
    def __init__(self, name, minimize_dt=False, end_sim=False):
        if DiscreteMarkovModel._disabled:
            raise RuntimeError, "The DiscreteMarkovModel has been disabled"\
                  " because no stochastic backend where found."
        params = odict()
        params["minimize_dt"] = create_parameter(minimize_dt, "minimize_dt")
        params["end_sim"] = create_parameter(end_sim, "end_sim")
        
        SubCellModelObject.__init__(self, name, params)
        self.submodels = {}
        self.variables = []
        self.connections =[]

        # Initiate the number of the model to Zero
        # This value is dependent on who connects to it and how it is used 
        # in for example a gated Boundary
        self._num = 0

        # Attributes keeping track of any given init state values or fixed
        # state values
        self._init_state_values = {}
        self._fixed_state_values = []

        # Initiate a dummy receiver func which will be populated by backend
        self.receivers_callback = None
        self.receiver_callback = None
        self.connector_callback = None

    def set_init_state(self, index, state_value):
        "Prescribe the init state value of the model given by the index"
        check_arg(index, int, 0)
        check_arg(state_value, int, 1)
        if index < 0:
            raise ValueError, "expected a positive value for index"
        if state_value < 0:
            raise ValueError, "expected a positive value for state_values"
        self._init_state_values[index] = state_value

    def set_fixed_state(self, index):
        "Fix the state to its init value for the model given by the index"
        check_arg(index, int, 0)
        if index < 0:
            raise ValueError, "expected a positive value for index"
        self._fixed_state_values.append(index)
        
    def _add_submodel(self, submodel, num=1):
        """Creates a place holder of a submodel if not already created.
The submodel is returned"""
        if isinstance(self, DiscreteMarkovSubModel):
            raise ValueError, "Submodels of discrete Markov models cannot "\
                  "have submodels"
        check_arg(submodel, str, 0)

        if submodel in self.submodels:
            raise TypeError, "a submodel of the same name already excists"

        # Add the submodel
        self.submodels[submodel] = DiscreteMarkovSubModel(self, submodel, num)
        
        return self.submodels[submodel]
    
    def _connect(self, receiver):
        "Connect the DiscreteMarkovModel to a receiver"
        from variables import Variable
        check_arg(receiver, Receiver, 0)

        con = Connection(Sender(self), receiver)

        if con in self.connections:
            raise ValueError, "The connection already excists"
        
        # Append connection
        self.connections.append(con)
        
        # Increase the number
        if not isinstance(receiver.receiver, Variable):
            self.num = self.num + receiver.receiver.num

    def _get_num(self):
        return self._num

    def _set_num(self, value):
        assert(isinstance(value, int) and value>=0)
        self._num = value
    
    def iter_receivers(self, var):
        """
        Use the backend specific recevier func to return a receiver to
        the actuall DiscreteMarkovModel"""
        if self.receivers_callback is None:
            raise TypeError, "expected 'receivers_callback' to be "\
                  "populated by a callable function"
        
        return self.receivers_callback(var)

    def get_receiver(self, var, num):
        """
        Use the backend specific recevier func to return a receiver to
        the actuall DiscreteMarkovModel"""
        
        if self.receiver_callback is None or \
               not callable(self.receiver_callback):
            raise TypeError, "expected 'receiver_callback' to be populated"\
                  " by a callable function"
        
        return self.receiver_callback(var, num)

    def get_connector(self, num):
        """
        Use the backend specific sender func to return a sender from
        the actuall DiscreteMarkovModel"""
        if self.connector_callback is None or \
               not callable(self.connector_callback):
            raise TypeError, "expected 'connector_callback' to be populated"\
                  " by a callable function"
        
        return self.connector_callback(num)
    
    # The num property
    num = property(_get_num, _set_num)
    
    def _receiver(self, variable):
        "Returns a Receiver object"
        check_arg(variable, str, 0)
        if variable not in self.variables:
            self.variables.append(variable)

        # Create and return a Receiver object
        return Receiver(self, variable)

    def __repr__(self):
        return "DiscreteMarkovModel('%s')"%self.name
    
    @classmethod
    def disable(cls):
        cls._disabled = True
        
class DiscreteMarkovSubModel(DiscreteMarkovModel):
    "A place holder for a sub model of a Discrete Markov model"
    def __init__(self, parent, name, subnum=1):
        DiscreteMarkovModel.__init__(self, name)
        check_arg(parent, DiscreteMarkovModel, 1)
        check_arg(subnum, int, 1)

        if subnum < 1:
            raise ValueError, "expected a positive number for the "\
                  "'subnum' argument"

        # Store data
        self.parent = parent
        self.subnum = subnum
        
    def __repr__(self):
        return "DiscreteMarkovSubModel(%s, '%s', %d)"%\
               (repr(self.parent), self.name, self.num)
    
    def _get_num(self):
        "return the total number of DiscreteMarkovSubModels"
        return self.parent.num*self.subnum

    def _set_num(self, value):
        assert(isinstance(value, int) and value>=0)
        assert(value%self.subnum==0)
        self.parent._num = value/self.subnum

    def _connect(self, receiver):
        raise TypeError, "cannot connect a submodel to an external receiver"

    def get_connector(self, num):
        raise TypeError, "cannot fetch sender from a submodel"

    num = property(_get_num, _set_num)
        
# Avoid problem with circular dependencies
from connection import *
