__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-16 -- 2011-10-10"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["Domain", "InfiniteDomain", "VolumeDomain", "MeshDomain",
           "Mesh", "mesh_domains", "infinite_domains", "volume_domains",
           "NodeCoordinate", "CellCoordinate"]

# System imports
import sympy

# SubCell imports
from subcell.common import *
from species import *
from subcellmodelobject import *
from variables import *

class NodeCoordinate(NodeParam):
    """A class to represent the coordinates at the nodes in a
    particular MeshDomain"""
    def __init__(self, coord, domain):
        check_arg(coord, str, 1)
        check_arg(domain, MeshDomain, 2)
        
        # Initialize the NodeParam
        NodeParam.__init__(self, domain, 1.0, coord)

        # Define a 'repr string'
        self._repr_str = "NodeCoordinate(%s)"
        
    
class CellCoordinate(CellParam):
    """A class to represent the coordinates at the center of each cell in a
    particular MeshDomain"""
    def __init__(self, coord, domain):
        check_arg(coord, str, 1)
        check_arg(domain, MeshDomain, 2)
        
        # Initialize the CellParam
        CellParam.__init__(self, domain, 1.0, coord)
    
        # Define a 'repr string'
        self._repr_str = "CellCoordinate(%s)"
        
# Some help functions
def add_scalar_species(self, species, init=0.0, **params):
    check_arg(species, Species, 0)
    self._add_species(species, ScalarSpecies(species, self, init, **params))
    
def add_const_species(self, species, init=0.0, **params):
    check_arg(species, Species, 0)
    self._add_species(species, ConstSpecies(species, self, init, **params))

def add_reaction(self, expr, reactants=None, products=None):
    from reactions import Reaction
    reaction = Reaction(self, expr, reactants, products)
    self.reactions.append(reaction)

def _domain_specific_domains(domains, DomainType):
    check_arg(domains, (list, tuple, Domain), 0, Domain)
    if isinstance(domains, Domain):
        domains = [domains]
    return [domain for domain in domains if isinstance(domain, DomainType)]

def mesh_domains(domains):
    return _domain_specific_domains(domains, MeshDomain)

def infinite_domains(domains):
    return _domain_specific_domains(domains, InfiniteDomain)

def volume_domains(domains):
    return _domain_specific_domains(domains, VolumeDomain)
    
class Domain(SubCellModelObject):
    """
    A base class for the different types of Domains possible to define
    in a SubCellModel"""
    def __init__(self, name, params=None):
        check_arg(name, str, 0)
        self.boundaries = []
        self.reactions = []
        self.species = odict()
        self.const_species = []
        self.scalar_species = []
        self.scalar_field_species = []
        self.fixed_scalar_field_species = []
        SubCellModelObject.__init__(self, name, params)

    def _add_species(self, species, domain_specific_species):
        "Protected method to store the domain specific species"
        assert(isinstance(species, Species))
        assert(isinstance(domain_specific_species, DomainSpecificSpecies))
        if species in self.species:
            raise ValueError, "Species: '%s' is already registered in the "\
                  "'%s' domain."%(species, self)
        self.species[species] = domain_specific_species
        if isinstance(domain_specific_species, ConstSpecies):
            self.const_species.append(domain_specific_species)
        elif isinstance(domain_specific_species, ScalarSpecies):
            self.scalar_species.append(domain_specific_species)
        elif isinstance(domain_specific_species, ScalarFieldSpecies):
            self.scalar_field_species.append(domain_specific_species)
        elif isinstance(domain_specific_species, FixedScalarFieldSpecies):
            self.fixed_scalar_field_species.append(domain_specific_species)
        setattr(self, str(species), domain_specific_species)

    def iter_scalar_field_species(self):
        "iterate over not None ScalarFieldSpecies"
        for species in self.scalar_field_species:
            if species is None:
                continue
            yield species

    def num_const_species(self):
        "Return the number of const species declared on the domain"
        return len(self.const_species)

    def num_scalar_species(self):
        "Return the number of scalar species declared on the domain"
        return len(self.scalar_species)
        
    def num_scalar_field_species(self):
        """
        Return the number of scalar field species declared on the domain
        """
        return len(self.scalar_field_species)

    def num_fixed_scalar_field_species(self):
        """
        Return the number of fixed scalar field species declared on the domain
        """
        return len(self.fixed_scalar_field_species)

    def num_field_species(self):
        """
        Return the number of field species declared on the domain
        """
        return len(self.fixed_scalar_field_species + self.scalar_field_species)

class InfiniteDomain(Domain):
    """
    An infinite domain, representing a domain with species that
    can only be const"""
    
    def __init__(self, name, **params):
        params = user_defined_parameters(params, \
                                         symnameprefix=name)
        Domain.__init__(self, name, params)
    
    add_const_species = add_const_species

class VolumeDomain(Domain):
    def __init__(self, name, volume=1.0, **params):
        params = user_defined_parameters(params, \
                                         symnameprefix=name)
        params["volume"] = create_parameter(volume, "volume", \
                                            gt=0, property_param=False,\
                                            symname="%s_volume"%name)
        Domain.__init__(self, name, params)
        
    add_scalar_species = add_scalar_species
    add_const_species = add_const_species
    add_reaction = add_reaction

class Mesh(SubCellModelObject):
    def __init__(self, filename, **params):
        check_arg(filename, str)
        params = user_defined_parameters(params, only_scalar=False, \
                                         symnameprefix=filename)
        params["filename"] = create_parameter(TypeCheck(filename), \
                                              "filename", \
                                              only_scalar=False,\
                                              symname="%s"%filename)
        
        SubCellModelObject.__init__(self, filename, params)
        
        # Populate attributes
        self.domains = []

    def __eq__(self, other):
        if not isinstance(other, Mesh):
            return False
        return self.filename == other.filename

    def __repr__(self):
        return "Mesh('%s')"%self.filename

class MeshDomain(Domain):
    def __init__(self, name, mesh, marker, **params):
        from species import NodeParam, CellParam
        check_arg(mesh, Mesh, 1)
        
        def check_array_params(param):
            "Small help function"
            if isinstance(param, (NodeParam, CellParam)):
                if param.domain is None:
                    param.domain = self
                else:
                    raise RuntimeError("expected the '%s' to "\
                                       "be defined at this domain ('%s')"%\
                                       (param.__class__.__name__, name))
        
        # Check user defined params for CellParams and NodeParams and
        # set domain in any domain-undefined Params
        for paramname, param in params.iteritems():
            
            # Rule out CellParams from any expressions
            if isinstance(param, sympy.Basic):
                for atom in init.atoms:
                    check_array_params(atom)
                check_array_params(param)
        
        # Allow ArrayParam for MeshDomain
        params = user_defined_parameters(params, allow_array=True, \
                                         symnameprefix=name)
        params["marker"] = create_parameter(marker, "marker", gt=0, \
                                            symname="%s_marker"%name)

        # Create a hidden parameter
        params["__volume"] = create_parameter(1.0, "volume", gt=0, \
                                              property_param=False)

        # The hidden parameter slaves the exposed one
        params["volume"] = create_parameter(params["__volume"][0], "volume", \
                                            gt=0, property_param=False,\
                                            symname="%s_volume"%name)

        # Add the six coordinate Parameters
        for coord in "xyz":
            for coord_type, Coordinate in zip(["node", "cell"],\
                                             [NodeCoordinate, CellCoordinate]):
                coord_name = "%s_%s"%(coord, coord_type)
                hidden_name = "__"+coord_name
                params[hidden_name] = \
                        create_parameter(Coordinate(coord_name, self), \
                                         coord_name, property_param=False,\
                                         allow_array=True)
                params[coord_name] = \
                        create_parameter(params[hidden_name][0], \
                                         coord_name, property_param=False, \
                                         allow_array=True, \
                                         symname="%s_%s"%(name, coord_name))
                
        Domain.__init__(self, name, params)

        # A scale dictionary for adding species dependent scale factors
        # used in particulare with electric potentials
        self.scale = odict()
        self.static_electric_potential = None
        self.mesh = mesh
        self.mesh.domains.append(self)

    def __repr__(self):
        return "MeshDomain('%s', %s%s)"%(self.name, repr(self.mesh), \
                                         self._param_repr())

    def add_scalar_field_species(self, species, init=0.0, sigma=0.0, **params):
        "Add a scalar field species to the MeshDomain"
        # Add a default scale factor
        check_arg(species, Species, 0)
        self.scale[species] = 1
        self._add_species(species, ScalarFieldSpecies(species, self, init,\
                                                      sigma, **params))

    def add_fixed_scalar_field_species(self, species, init=0.0, **params):
        "Add a fixed scalar field species to the MeshDomain"
        # Add a default scale factor
        check_arg(species, Species, 0)
        self.scale[species] = 1
        self._add_species(species, FixedScalarFieldSpecies(\
            species, self, init, **params))

    def add_static_electric_potential(self, expr):
        """
        Add a static electric potential to the domain
        
        The expr can either be a NumPy array defining the potential at each
        vertex, or a coordinate dependent expression.
        """
        from staticelectricpotential import StaticElectricPotential
        if self.num_scalar_species() != 0:
            # FIXME: Is this correct?
            raise ValueError, "cannot add an electric potential to a "\
                  "MeshDomain with scalar species defined on it"
        
        if self.static_electric_potential is not None:
            raise ValueError, "the domain '%s' already include a static "\
                  "electric potential"%self.name
        self.static_electric_potential = StaticElectricPotential(expr, self)
    
    add_reaction = add_reaction
    add_scalar_species = add_scalar_species
    add_const_species = add_const_species

