__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-16 -- 2011-10-18"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["Flux", "DiffusionFlux"]

# System imports
import sympy

# SubCell imports
from subcell.common import *
from subcellmodelobject import *
from species import *
from variables import *
from domains import *

class Flux(SubCellModelObject):
    _count = 0
    "The main flux object, used to describe a flux on a Boundary"
    def __init__(self, fluxname, boundary, species, expr, limits=None,
                 epot_scale_all=False, **params):
        from boundaries import BoundaryBase
        check_arg(fluxname, str, 0)
        check_arg(boundary, BoundaryBase, 1)
        check_arg(species, Species, 2)
        check_arg(expr, (sympy.Basic, int, float, TypeCheck), 3)
        
        name = "%s_%s_%s"%(fluxname, species, boundary)
        params = user_defined_parameters(params, symnameprefix=name)
        
        params["epot_scale_all"] = create_parameter(
            epot_scale_all, "epot_scale_all")
        SubCellModelObject.__init__(self, name, params)
        
        # Store fluxname
        self.fluxname = fluxname
        
        # Check sanity for epot_scale_all
        self.epot_domain = None
        if epot_scale_all:
            
            # Extract the mesh_domains with epot added to them
            domains = [domain for domain in  mesh_domains(boundary.domains)\
                       if domain.static_electric_potential is not None]
            
            # If we have two meshdomains both cannot have an epot added to
            # them, while having epot_scale_all True
            if len(domains) > 1:
                raise ValueError, "cannot create a flux with "\
                      "'epot_scale_all' True between two MeshDomains "\
                      "that both adds an electric potential"
            if len(domains) < 1:
                raise ValueError, "cannot create a flux with "\
                          "'epot_scale_all' True when no MeshDomain adds"\
                          " an electric potential"
            
            # Store the epot domain
            self.epot_domain = domains[0]
            
        # Store the Flux count and increase the global Flux count
        self.count = Flux._count
        Flux._count += 1
        
        # Prepare attributes that might get populated
        self.scalar_field_species = []
        self.fixed_scalar_field_species = []
        self.const_species = []
        self.scalar_species = []
        self.all_species = []
        self.variables = []
        self.parameters = []
        
        # Wrap any scalar to a sympy scalar
        if isinstance(expr, scalar):
            expr = sympy.Number(expr)
        elif isinstance(expr, TypeCheck):
            expr = expr.sym

        # Check if all domains have species declared on them
        if not all (species in domain.species \
                    for domain in boundary.domains):
            raise TypeError, "Expected all domains connected by Boundary:"\
                  " '%s' to have a '%s' declared on them."%(boundary, species)

        # Check arguments and extract DomainSpecificSpecies and Variables
        self._extract_domain_specific_species_and_variables(expr, boundary)

        # Check that the flux is not from a fixed_scalar_field_species
        if any(isinstance(domain.species[species], FixedScalarFieldSpecies) \
               for domain in mesh_domains(boundary.domains)):
            raise TypeError, "the flux cannot be applied to a "\
                  "FixedScalarFieldSpecies"
        
        # Check that the flux connects at least one Scalar(Field)Species
        if not any(isinstance(domain.species[species], \
                              (ScalarFieldSpecies, ScalarSpecies)) \
                   for domain in boundary.domains):
            raise TypeError, "The flux need to connect at least one "\
                  "Scalar- or ScalarFieldSpecies"
        
        # A channel can have four different type flags that can be either
        # True or False:
        #   robin, constant, variable_dependent, gated
        #
        # NOTE: constant means constant jacobian contribution
        # NOTE: (basically linear)
        #
        # A general flux can be described by:
        # 
        #    J = alpha*u + c0
        # 
        # If the flux is:
        #   robin, alpha is nonzero
        #   constant and robin, alpha is constant
        #   constant but variable dependent it is semi constant
        #   (constant during a Newton iteration)
        #   gated it resides on a Gated Boundary and can be turned on and
        #   off during simulation
        
        # Store arguments
        self._expr = expr
        self.boundary = boundary
        self.species = species

        # Set flags determining the type the flux is
        self._set_flux_types()
        
        # Add itself to the boundary
        self.boundary._add_flux(self)

    def _set_flux_types(self):
        """
        Set the type of the flux
        """
        self.robin = any(self.species == species.species for species in \
                         self.scalar_field_species + self.scalar_species)
        self.constant, self.variable_dependent = \
                       self._dependencies(self.expr())
        
    def expr(self, to_number=False):
        "Return the expr with parameters evaluated"
        if to_number:
            return self._expr.subs(odict((param,
                                          sympy.Number(param.param.get()))\
                                         for param in self.parameters))
        else:
            return self._expr
    
    def scalar_field_species_domains(self):
        """
        Return the domains where the receiving species is a scalar
        field speices.
        """
        return [domain for domain in self.boundary.domains \
                if isinstance(domain.species[self.species], \
                              ScalarFieldSpecies)]

    def scalar_species_domains(self):
        """
        Return the domains where the receiving species is a scalar
        speices.
        """
        return [domain for domain in self.boundary.domains \
                if isinstance(domain.species[self.species], ScalarSpecies)]
    
    def num_scalar_field_species_domains(self):
        """
        Return the number of domains with scalar fields species
        """
        return len(self.scalar_field_species_domains())

    def num_scalar_species_domains(self):
        """
        Return the number of domains with scalar fields species
        """
        return len(self.scalar_species_domains())

    def _extract_domain_specific_species_and_variables(self, expr, boundary):
        
        # Exchange any Symbols generated through SymPy magic back to
        # ParSlaves
        expr = exchange_symbols_with_ParSlaves(expr)

        # If we got a Sympy expression check that the provided atoms are valid
        allowed_types = (int, float, sympy.Number, DomainSpecificSpecies,
                         Variable, ParSymbol)
        if not all(isinstance(atom, allowed_types) for atom in expr.atoms()):
            # FIXME: Add unit test for when this fail!
            raise TypeError, "expected a SymPy expression consisting of only "\
                  "scalars and domain specific species or variables"
        
        # Iterate over the DomainSpecificSpecies
        debug("Analyse Flux on boundary %s:", boundary)
        for atom in expr.atoms():
            if isinstance(atom, (ConstSpecies, ScalarSpecies, \
                                 ScalarFieldSpecies)):
                self.all_species.append(atom)
            if isinstance(atom, ConstSpecies):
                debug("  Register %s: %s", atom.__class__.__name__, atom)
                self.const_species.append(atom)
            elif isinstance(atom, ScalarSpecies):
                debug("  Register %s: %s", atom.__class__.__name__, atom)
                self.scalar_species.append(atom)
            elif isinstance(atom, FixedScalarFieldSpecies):
                debug("  Register %s: %s", atom.__class__.__name__, atom)
                self.fixed_scalar_field_species.append(atom)
            elif isinstance(atom, ScalarFieldSpecies):
                debug("  Register %s: %s", atom.__class__.__name__, atom)
                self.scalar_field_species.append(atom)
            elif isinstance(atom, Variable):
                debug("  Register %s: %s", atom.__class__.__name__, atom)
                self.variables.append(atom)
            elif isinstance(atom, ParSymbol):
                if isinstance(atom.param, ParSlave):
                    debug("  Register %s: %s", atom.__class__.__name__, \
                          atom.param.symbols)
                else:
                    debug("  Register %s: %s", atom.__class__.__name__, atom)
                self.parameters.append(atom)
                
        # FIXME: Consider implementing the use of FixScalarFieldSpecies in
        # FIXME: Fluxes
        if self.fixed_scalar_field_species:
            raise NotImplementedError("FixedScalarFieldSpecies cannot be "\
                                      "used to declare Fluxes.")
        
        # Check for undefined domains in Cell and Node Params
        # and that domain for Node and Cell params are defined on domain
        # connected by the Boundary
        for parsymbol in self.parameters:
            param = parsymbol.param
            if not param.is_array:
                continue
            what = "NodeParam" if param.is_node_param else "CellParam"
            if self.num_scalar_field_species_domains() == 2:
                raise RuntimeError("Only scalar parameters can be used in"\
                                   " a FluxBoundary between two "\
                                   "ScalarFieldSpecies.")
            
            if param.domain is None:
                raise RuntimeError("Found a %s in the Flux expression "\
                                   "which has no domain defined. See "\
                                   "dosctring of '%s' for more info."%\
                                   (what, what))
            
            if param.domain not in boundary.domains:
                raise RuntimeError("Found a %s in the Flux expression "\
                                   "which is not defined on a MeshDomain "\
                                   "connected by the Boundary the Flux is"\
                                   " defined on"%(what))
        
        # Check that all passed DomainSpecificSpecies are defined at a domain
        # which the Boundary connects
        allowed_species = []
        for domain in boundary.domains:
            allowed_species.extend(domain.species.values())
        
        domain_species = self.scalar_field_species + self.const_species + \
                         self.scalar_species + self.fixed_scalar_field_species
        for species in domain_species:
            if species not in allowed_species:
                # FIXME: Add unit test for when this fail!
                raise ValueError, "found a DomainSpecificSpecies which is not"\
                      " declared in a domain connected by the Boundary %s: "\
                      "%s"%(str(self), str(species))
    
    def _dependencies(self, expr):
        "Check for dependencies in the Robin term"

        # Default values
        constant = True
        variable_dependent = False
        
        # If not Robin just return default values
        if not self.robin:
            return constant, variable_dependent

        # A flux can contain a lot of different possible combinations of
        # species when differentiated wrt to the receiving species we can
        # reveal if it is linear, constant Neuman or Robin. I need to figure
        # the different combinations and get this going!
        # I probably need to figure out how to include the ScalarSpecies too, 
        # as this will play an important role in all of this.
        # 
        # Collect all variables that we solve for
        all_scalar_species = self.scalar_field_species + self.scalar_species

        # If the flux is of robin type we only check the alpha part for
        # constantness
        #     J = alpha*u+u0
        variable_species = [species for species in all_scalar_species
                            if self.species == species.species]

        # Iterate over the variables
        for scalar in variable_species:
            
            # Differentiate the expr wrt to each variable
            differentiated = expr.diff(scalar)
            for a in differentiated.atoms():

                # If there still excist a scalar the expr is nonlinear
                if a in all_scalar_species:
                    constant = False
                
                # If there still excist a variable the expr is variable
                # dependent
                if a in self.variables:
                    variable_dependent = True
        
        # Return dependencies
        return constant, variable_dependent

    def sign(self, domain):
        "Returns the sign of the flux given the domain"
        if domain not in self.boundary.domains:
            raise ValueError, "'%s' is not part of the boundary which this "\
                  "flux is defined on"%domain.name
        if len(self.boundary.domains)==1:
            return "+"
        return "-" if self.boundary.domains.index(domain) == 0 else "+"
    
    def __repr__(self):
        return "Flux(%s, %s, %s)"%(repr(self.boundary), repr(self.species), \
                                   str(self._expr))

class DiffusionFlux(Flux):
    """
    A linear diffusion flux between two boundaries defined by the distance
    between them together with a diffusion constant sigma
    """
    def __init__(self, boundary, species, sigma, epot_scale_all):
        from boundaries import BoundaryBase
        check_arg(boundary, BoundaryBase, 0)
        check_arg(species, Species, 1)
        
        if not (isinstance(sigma, scalar) or \
                (isinstance(sigma, TypeCheck) and sigma.is_scalar)):
            raise TypeError("Expected a scalar value for the 'sigma' argument")

        # Sanity checks
        if len(boundary.domains)!=2:
            raise ValueError("A DiffusionFlux need to be declared on a "\
                             "Boundary that connects two domains.")
        if not all(species in domain.species for domain in boundary.domains):
            raise ValueError("Expected '%s' to be declared on both domains"\
                             " of the Boundary."%species)

        # Check that we only have one DiffusionFlux for each species on a
        # Boundary
        if species in boundary.fluxes and \
           any(isinstance(flux, DiffusionFlux) \
               for flux in boundary.fluxes[species]):
            raise TypeError("A DiffusionFlux is already defined for "\
                            "'%s' at '%s'"%(species, boundary))
        
        # Init Base class with phony  expr
        Flux.__init__(self, "diffusion", boundary, species, 1, \
                      epot_scale_all=epot_scale_all, sigma=sigma)
        
        # Create flux expression from now registered sigma
        expr = (boundary.domains[0].species[species]-\
                boundary.domains[1].species[species])*self.sigma/boundary.dx
        
        # Check arguments and extract DomainSpecificSpecies and Variables
        self._extract_domain_specific_species_and_variables(expr, boundary)

        # Store arguments
        self._expr = expr
        
        # Set flux type based on the flux expression
        self._set_flux_types()

        # Re-generate the bondary values
        self.boundary._generate_boundary_values(self)
