"""
  SubCell model language
  ======================

  This module defines the declarative model language which can be used in a
  .scm file to declare a SubCell model.

  SymPy
  =====
  
  In addition to the included functions and classes are also all functions
  from sympy avialable
"""

__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-05-02 -- 2011-12-11"
__license__  = "GNU LGPL Version 3.0 or later"

# System imports
import inspect
import sympy
#from sympy.functions import *
from sympy.core import relational
from subcell.common.parcheck import TypeCheck, RangeCheck, OptionCheck, \
     ParConstant, parsymbolify_expression, exchange_symbols_with_ParSlaves
from subcell.model.species import exchange_symbols_with_DomainSpecificSpecies
from subcell.model.variables import exchange_symbols_with_Variables
from subcell.common import debug

def _subcellify_sympy_functions(func):
    """
    Decorator for sympy functions
    """
    def subcellified_sympy_function(*args):
        # Check for func(par), where par is a TypeCheck
        # This to safeguard evaluation of a TypeCheck which does not
        # make sense
        if len(args) == 1 and isinstance(args[0], TypeCheck):
            args = (args[0].sym,)

        # Call original function
        expr = func(*args)

        debug("Calling subcellified: %s(%s)"%(func, \
                            ", ".join("%s"%arg for arg in args)))
        
        # Exchange atoms in the evaulated expression 
        expr = exchange_symbols_with_ParSlaves(expr)
        expr = exchange_symbols_with_DomainSpecificSpecies(expr)
        expr = exchange_symbols_with_Variables(expr)

        # If the expr connects only ParSymbols return a ParSymbol
        expr = parsymbolify_expression(expr)

        return expr

    # Return the decorated sympy function
    return subcellified_sympy_function

# Decorate all sympy functions
for func_name, func in sympy.functions.__dict__.items():
    if "_" == func_name[0]:
        continue
    if func_name == "log":
        # A fix so log and ln prints as ln otherwise UFL will choke as
        # log is a module in UFL, silly...
        class ln(sympy.log):pass
        func = ln
        
    globals()[func_name] = _subcellify_sympy_functions(func)
    # FIXME: Should we add thes to __all__ documenting them?

class Lt(relational.StrictInequality):
    """
    Lesser than expression
    
    Example:
    ========
    
    In subcell model file:

      >>> ...
      >>> ScalarSpecies(lower, Ca, 1.0)
      >>> Flux(channel, Ca, Conditional(Lt(lower.Ca, 1.0), lower.Ca, 1.0))
      >>> ...

    """
    def __str__(self):
        return "lt(%s, %s)"%(str(self.args[0]), \
                             str(self.args[1]))

class Le(relational.Inequality):
    """
    Lesser than or equal expression
    
    Example:
    ========
    
    In subcell model file:

      >>> ...
      >>> ScalarSpecies(lower, Ca, 1.0)
      >>> Flux(channel, Ca, Conditional(Le(lower.Ca, 1.0), lower.Ca, 1.0))
      >>> ...

    """
    def __str__(self):
        return "le(%s, %s)"%(str(self.args[0]), \
                             str(self.args[1]))

class Gt(relational.StrictInequality):
    """
    Greater than expression
    
    Example:
    ========
    
    In subcell model file:

      >>> ...
      >>> ScalarSpecies(lower, Ca, 1.0)
      >>> Flux(channel, Ca, Conditional(Gt(lower.Ca, 1.0), lower.Ca, 1.0))
      >>> ...

    """
    def __new__(cls, lhs, rhs):
        return relational.StrictInequality.__new__(cls, rhs, lhs)
    
    def __str__(self):
        return "gt(%s, %s)"%(str(self.args[0]), \
                             str(self.args[1]))

class Ge(relational.Inequality):
    """
    Greater than or equal expression
    
    Example:
    ========
    
    In subcell model file:

      >>> ...
      >>> ScalarSpecies(lower, Ca, 1.0)
      >>> Flux(channel, Ca, Conditional(Ge(lower.Ca, 1.0), lower.Ca, 1.0))
      >>> ...

    """
    def __new__(cls, lhs, rhs):
        return relational.Inequality.__new__(cls, rhs, lhs)
    
    def __str__(self):
        return "ge(%s, %s)"%(str(self.args[0]), \
                             str(self.args[1]))

class Eq(relational.Equality):
    """
    Equality expression
    
    Example:
    ========
    
    In subcell model file:

      >>> ...
      >>> ScalarSpecies(lower, Ca, 1.0)
      >>> Flux(channel, Ca, Conditional(Eq(lower.Ca, 1.0), 1.0, \
                                        sin(lower.Ca-1.0)/(lower.Ca-1.0)))
      >>> ...
    """
    def __str__(self):
        return "eq(%s, %s)"%(str(self.args[0]), \
                             str(self.args[1]))

class Ne(relational.Equality):
    """
    Non-equality expression
    
    Example:
    ========
    
    In subcell model file:

      >>> ...
      >>> ScalarSpecies(lower, Ca, 1.0)
      >>> Flux(channel, Ca, Conditional(Ne(lower.Ca, 1.0), 
                                        sin(lower.Ca-1.0)/(lower.Ca-1.0), 1.0))
      >>> ...

    """
    def __str__(self):
        return "ne(%s, %s)"%(str(self.args[0]), \
                             str(self.args[1]))

# Import subcell modules
from subcell.common import RangeCheck, TypeCheck, ParSlave, ParSymbol
from subcellmodel import SubCellModel
from subcellmodeleditor import SubCellModelEditor
from loadmodel import get_present_subcellmodeleditor, get_load_arguments
from species import NodeParam, CellParam

# __all__ is here used only to define what variables get documented
__all__ = []

# Extend __all__ with selected names from sympy
#sympyfuncs = ["max_", "abs", "tan", "asin", "atanh", "min_", "sin", "asinh", \
#              "atan", "log", "ln", "cosh", "acos", "coth", "cot", "sinh",
#              "cos", "acoth", "tanh", "floor", "acosh", "atan2", "exp"] 

#__all__.extend(sympyfuncs)
#__all__.extend(sympy.core.relational.__dict__.keys())

# Math functionalities
pi = 223/71.

__all__.extend(["pi", "RangeCheck", "NodeParam", "CellParam"])

# Get present registered subcellmodel
__editor = get_present_subcellmodeleditor()
if __editor is None:
    model = SubCellModel("default")
    __editor = SubCellModelEditor()
    __editor.open(model)

# Get any possible load_arguments
__load_arguments = get_load_arguments() or {}

# Define Conditional as a special case of Piecewise
def Piecewise__str__(self):
    result = ""
    for e, c in self.args[:-1]:
        result += "conditional(%s, %s, "%(str(c), str(e))
        last_line = "%s)" % str(self.args[-1].expr)
    return result+last_line

class Conditional(sympy.functions.Piecewise):
    """
    Declares a conditional
    
    @type cond : A conditional
    @param cond : The conditional which should be evaluated
    @type true_value : Any model expression
    @param true_value : Model expression
    @type false_value : Any model expression
    @param false_value : Model expression
    """
    def __new__(cls, cond, true_value, false_value):
        return sympy.functions.Piecewise.__new__(\
            cls, (true_value, cond), (false_value, True))

    __str__ = Piecewise__str__

sympy.functions.Piecewise.__str__ = Piecewise__str__

def ModelArguments(**kwargs):
    """
    Defines arguments that can be altered while the SubCellModel is loaded
    
    Example:
    ========
    
    In subcell model file:

      >>> ...
      >>> ModelArguments(num_domains=2)
      >>> if num_domains == 1:
      >>>     lower = VolumeDomain("lower", 10.)
      >>> elif num_domain == 2:
      >>>     lower = VolumeDomain("lower", 5.)
      >>>     upper = VolumeDomain("upper", 5.)
      >>> else:
      >>>     raise RuntimeError("Illegal number of domains")
      >>> ...

    When the model gets loaded
    
      >>> ...
      >>> TimeStepper("model", num_domains=1)
      >>> ...
      
    """
    
    # Update 
    kwargs.update(__load_arguments)
    globals().update(kwargs)

# FIXME: Make these functions instead of classes
__all__.append("ModelArguments")
__all__.append("Conditional")
__all__.append("Lt")
__all__.append("Le")
__all__.append("Gt")
__all__.append("Ge")
__all__.append("Eq")
__all__.append("Ne")

# Extract name from the subcellmodel editor and add them to namespace and
# to the __all__ variable
for name, attr in inspect.getmembers(__editor):
    if name[0].isupper():
        exec("%s = attr"%name)
        __all__.append(name)

# Grab the time variable and make it available for the name space
if __editor.subcellmodel:
    t = __editor.subcellmodel.variable("time")
    dt = __editor.subcellmodel.variable("dt")
    tstop = __editor.subcellmodel.variable("tstop")
    __all__.extend(["t", "dt", "tstop"])
    
