__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-11-29 -- 2011-09-14"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["Reaction"]

# System imports
import sympy
import types

# SubCell imports
from subcell.common import *
from subcellmodelobject import *
from variables import *
from domains import *
from species import *

class Reaction(SubCellModelObject):
    """The main reaction object, used to describe a reaction on a domain"""
    def __init__(self, domain, expr, reactants=None, products=None):
        """
        Instantiate a Reaction object
        """
        check_arg(domain, (VolumeDomain, MeshDomain), 0)
        check_arg(expr, (sympy.Basic, int, float), 1)
        
        # Reactants and product attributes. They will be populated with Species
        # as keys and the stochiometric value as values
        self.reactants = odict()
        self.products = odict()
        
        # Check reactants and products kwargs
        self._check_reactants_and_products(reactants, products, domain)
        
        # Call base class constructor
        SubCellModelObject.__init__(self, "%s=>%s reaction in %s"%\
                (", ".join(str(species) for species in self.reactants.keys()),\
                 ", ".join(str(species) for species in self.products.keys()),\
                                    domain))
        
        # Prepare attributes that might get populated
        self.scalar_field_species = []
        self.fixed_scalar_field_species = []
        self.const_species = []
        self.scalar_species = []
        self.variables = []
        self.parameters = []
        
        # Wrap any scalar to a sympy scalar
        if not isinstance(expr, sympy.Basic):
            expr = sympy.Number(expr)
        elif isinstance(expr, TypeCheck):
            expr = expr.sym
        
        # Check arguments and extract DomainSpecificSpecies and Variables
        self._extract_domain_specific_species_and_variables(expr, domain)
        
        # A channel can have four different type flags that can be either
        # True or False:
        #   robin, constant, variable_dependent, gated
        # A general flux can be described by:
        # 
        #    J = alpha*u + c0
        # 
        # If the flux is:
        #   robin, alpha is nonzero
        #   constant and robin, alpha is constant
        #   constant but variable dependent it is semi constant
        #   (constant during a Newton iteration)
        #   gated it resides on a Gated Boundary and can be turned on and
        #   off during simulation
        
        # Store arguments
        self._expr = expr
        self.domain = domain
        
        # Set flags determining the type the reaction
        self.robin = any(species.species in self.products.keys()+\
                         self.reactants.keys() for species in \
                         self.scalar_field_species + self.scalar_species)
        self.constant, self.variable_dependent = \
                       self._dependencies(self.expr())

        reactants_and_products = self.reactants.keys() + self.products.keys()
        self.only_field_species = all(isinstance(domain.species[species], \
                                                 FieldSpecies) for species \
                                      in reactants_and_products)

        # Add itself to the domain
        self.domain.reactions.append(self)
        
    def expr(self, to_number=False):
        "Return the expr with parameters evaluated"
        if to_number:
            return self._expr.subs(odict((param, param.to_Number())\
                                         for param in self.parameters))
        else:
            return self._expr
    
    def _check_reactants_and_products(self, reactants, products, domain):
        # Check reactants and products kwargs
        if products is None and reactants is None:
            raise TypeError, "expected at least one of 'reactants' and "\
                  "'products' to be defined"

        for kwarg, name in [(reactants, "reactants"), (products, "products")]:
            check_kwarg(kwarg, name, (types.NoneType, Species, tuple, list))

            # If no reactant or product is provided just continue
            if kwarg is None:
                continue
            
            # If a list is provided
            if isinstance(kwarg, list):
                for inst in kwarg:
                    # Check each item in the passed list
                    if isinstance(inst, tuple):
                        if len(inst)!=2 or \
                           (not isinstance(inst[0], Species)\
                           or not isinstance(inst[1], int) or inst[1]>0):
                            raise TypeError, "expected elements of '%s' "\
                              "kwarg to be a 'Species' or a tuples of length "\
                              "2 with a Species as its first element, and a "\
                              "positive number as the second element, "\
                              "defining the stochiometri"%name
                        getattr(self, name)[inst[0]] = inst[1]
                    # If no stochiometric information is provided, assume 1
                    elif isinstance(inst,  Species):
                        getattr(self, name)[inst] = 1
                    else:
                        raise TypeError, "expected elements of '%s' kwarg to "\
                              "be a 'Species' or a tuples of length 2 with a "\
                              "Species as its first element, and a positive "\
                              "number as the second element, defining the "\
                              "stochiometri"%name
            
            # Else if a Species is provided     
            elif isinstance(kwarg, Species):
                getattr(self, name)[inst] = 1

            # Else a tuple is provided
            else:
                if len(kwarg)!=2 or \
                    (not isinstance(kwarg[0], Species) or \
                     not isinstance(kwarg[1], int) or kwarg[1]>0):
                    raise TypeError, "expected the tuple provided to the '%s'"\
                          " kwarg to have length 2 with a Species as its "\
                          "first element, and a positive number as the second"\
                          " element, defining the stochiometri"%name
                getattr(self, name)[kwarg[0]] = kwarg[1]

            # Check that the reactants and products are defined on the domain
            for species in getattr(self, name).keys():
                if not species in domain.species:
                    raise ValueError, "the %s '%s' is not declared as a "\
                          "Species in the %s domain"%(name, str(species),\
                                                      str(domain))

    def _extract_domain_specific_species_and_variables(self, expr, domain):
        
        # Exchange any Symbols generated through SymPy magic back to
        # ParSlaves
        expr = exchange_symbols_with_ParSlaves(expr)

        # If we got a Sympy expression check that the provided atoms are valid
        allowed_types = (int, float, sympy.Number, DomainSpecificSpecies,
                         Variable, ParSymbol)
        if not all(isinstance(atom, allowed_types) for atom in expr.atoms()):
            # FIXME: Add unit test for when this fail!
            raise TypeError, "expected a SymPy expression consisting of only "\
                  "scalars and domain specific species or variables"
        
        # Iterate over the DomainSpecificSpecies
        debug("Analyse Reaction on domain %s (%s):", domain, expr)
        for atom in expr.atoms():
            if isinstance(atom, ConstSpecies):
                self.const_species.append(atom)
                debug("  Register %s: %s", atom.__class__.__name__, atom)
            elif isinstance(atom, ScalarSpecies):
                self.scalar_species.append(atom)
                debug("  Register %s: %s", atom.__class__.__name__, atom)
            elif isinstance(atom, ScalarFieldSpecies):
                self.scalar_field_species.append(atom)
                debug("  Register %s: %s", atom.__class__.__name__, atom)
            elif isinstance(atom, FixedScalarFieldSpecies):
                self.fixed_scalar_field_species.append(atom)
                debug("  Register %s: %s", atom.__class__.__name__, atom)
            elif isinstance(atom, Variable):
                self.variables.append(atom)
                debug("  Register %s: %s", atom.__class__.__name__, atom)
            elif isinstance(atom, ParSymbol):
                if isinstance(atom.param, ParSlave):
                    debug("  Register %s: %s (%s)", atom.__class__.__name__, \
                          atom, atom.param.symbols)
                else:
                    debug("  Register %s: %s", atom.__class__.__name__, atom)
                self.parameters.append(atom)
        
        # Check included parameters
        for parsymbol in self.parameters:
            param = parsymbol.param
            if not (param.is_scalar or param.is_node_param):
                raise RuntimeError("parameters included in a reaction need to"\
                        " be scalars or NodeParam")

            # Check that only scalar parameters are used for a Reaction on
            # a VolumeDomain
            if isinstance(domain, VolumeDomain) and not param.is_scalar:
                raise RuntimeError("Only scalar parameters allowed in "\
                                   "Reactions on VolumeDomains")
            
            # Check that any NodeParam is defined on the same domain as
            # the reaction
            if param.is_node_param and domain != param.domain:
                raise RuntimeError("Node parameters included in a rection "\
                                   "need to be declared on the same domain "\
                                   "as the reaction.")
        
        # Check that all passed DomainSpecificSpecies are defined at the domain
        allowed_species = domain.species.values()
        
        domain_species = self.scalar_field_species + self.const_species + \
                         self.scalar_species
        for species in domain_species:
            if species not in allowed_species:
                # FIXME: Add unit test for when this fail!
                raise ValueError, "found a DomainSpecificSpecies: '%s'"\
                      " which is not declared on the domain for this "\
                      "reaction"%str(species)
        
    def _dependencies(self, expr):
        "Check for dependencies in the Robin term"
        
        # Default values
        constant = True
        variable_dependent = False
        
        # If not Robin just return default values
        if not self.robin:
            return constant, variable_dependent

        # Collect all variables that we solve for
        all_scalar_species = self.scalar_field_species + self.scalar_species

        # If the reaction is of robin type we only check the alpha part for
        # constantness
        #     R = alpha*u+u0
        variable_species = self.reactants.keys() + self.products.keys()

        # Iterate over the variables
        for scalar in variable_species:
            
            # Differentiate the expr wrt to each variable
            differentiated = expr.diff(self.domain.species[scalar])
            for a in differentiated.atoms():
                
                # If there still excist a scalar the expr is nonlinear
                if a in all_scalar_species:
                    constant = False
                
                # If there still excist a variable the expr is variable
                # dependent
                if a in self.variables:
                    variable_dependent = True
        
        # Return dependencies
        return constant, variable_dependent
    
    def __repr__(self):
        if not hasattr(self, "_repr"):
            reactants = "" if not self.reactants else \
                ", reactants=[%s]"%(", ".join(\
                repr(item) for item in self.reactants.iteritems()))
            products = "" if not self.products else \
                ", products=[%s]"%(", ".join(\
                repr(item) for item in self.products.iteritems()))
            self._repr = "Reaction(%s, %s%s%s)"%\
                (repr(self.domain), str(self._expr), reactants, products)
        return self._repr
