__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-16 -- 2011-10-21"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["DomainSpecificSpecies", "ScalarFieldSpecies", "ConstSpecies",\
           "FixedScalarFieldSpecies", "ScalarSpecies", "Species",\
           "MeshBoundarySpecies", "FieldSpecies", "NodeParam", "CellParam"]

# System imports
import sympy
import types

# SubCell imports
from subcell.common import *
from subcellmodelobject import *
from connection import *

# Sub classes of ArrayParam
class NodeParam(ArrayParam):
    """
    A class to declare parameters at the nodes in a Mesh
    """
    def __init__(self, domain=None, value=1, name=None):
        from subcell.model.domains import MeshDomain
        check_arg(domain, (MeshDomain, types.NoneType))
        ArrayParam.__init__(self, value, name=name)
        self.is_node_param = True
        self.domain = domain

class CellParam(ArrayParam):
    """
    A class to declare parameters at the Cell in a Mesh
    """
    def __init__(self, domain=None, value=1, name=None):
        from subcell.model.domains import MeshDomain
        check_arg(domain, (MeshDomain, types.NoneType))
        ArrayParam.__init__(self, value, name=name)
        self.is_cell_param = True
        self.domain = domain

def enforce_and_check_node_param(param, name, modelname, domain, \
                                 expr_param=False):
    if param.is_cell_param:
        raise TypeError("CellParams cannot be used in an "\
                        "expression to described a param ('%s')"\
                        "value to a %s. Use "\
                        "NodeParams instead"%(name, modelname))
    if param.is_node_param:
        if param.domain is None:
            if expr_param:
                raise RuntimeError("NodeParam '%s' in expression for '%s' "\
                                   "has no Domain"%(str(param), name))
            param.domain = domain
        elif param.domain != domain:
            raise RuntimeError("expected a NodeParam to be defined "\
                               "at the same domain as the %s"%modelname)
    
def enforce_and_check_cell_param(param, name, modelname, domain, \
                                 expr_param=False):
    if param.is_node_param:
        raise TypeError("NodeParams cannot be used in an "\
                        "expression to described a param ('%s')"\
                        "value to a %s. Use "\
                        "CellParams instead"%(name, modelname))
    if param.is_cell_param:
        if param.domain is None:
            if expr_param:
                raise RuntimeError("CellParam '%s' in expression for '%s' "\
                                   "has no Domain"%(str(param), name))
            param.domain = domain
        elif param.domain != domain:
            raise RuntimeError("expected a CellParam to be defined "\
                               "at the same domain as the %s"%modelname)

def exchange_symbols_with_DomainSpecificSpecies(expr):
    """
    Substitution dict to remove all generated sympy.Symbols which
    does not play with DomainSpecificSpecies
    """
    d = {}
    for atom in expr.atoms():
        if str(atom) in DomainSpecificSpecies._all_objects:
            d[atom] = DomainSpecificSpecies._all_objects[str(atom)]
    return expr.subs(d)

class DomainSpecificSpecies(SubCellModelObject, sympy.Symbol):
    _all_objects = {}
    "Creates a domain specific specie"
    def __new__(cls, species, domain, params):
        from domains import Domain
        assert(isinstance(species, Species))
        assert(isinstance(domain, Domain))
        assert("init" in params)
        assert(isinstance(params, dict))
        
        # Init sympy object
        self = sympy.Symbol.__new__(cls, domain.name+"."+species.name)
        DomainSpecificSpecies._all_objects[str(self)] = self
        
        # Store attributes
        self.domain = domain
        self.species = species
        species.domains.append(domain)
        self.fluxes = []
        self.connections = []

        # Init base class
        SubCellModelObject.__init__(self, domain.name+"."+species.name, \
                                    params)
        
        return self

    def __repr__(self):
        return "%s('%s', %s, %s%s)"%(type(self).__name__, self.name, \
                                   repr(self.species), repr(self.domain), \
                                   self._param_repr())

    def flux_sign(self, flux):
        "Return the sign of the flux as a string"
        check_arg(flux, Flux)
        assert(flux in self.fluxes)
        return flux.sign(self.domain)

    def _connect(self, receiver):
        "Store connection information between the species and a reciever"
        check_arg(receiver, Receiver, 0)
        con = Connection(Sender(self), receiver)
        if con in self.connections:
            raise ValueError, "The connection already excists"
        
        # Store the connection
        self.connections.append(con)

class FieldSpecies(DomainSpecificSpecies):
    """
    A domain specific species declared to be a field defined on a mesh
    """
    def __new__(cls, species, domain, init, params):
        from boundaries import mesh_boundaries
        from domains import MeshDomain
        check_arg(species, Species, 0)
        check_arg(domain, MeshDomain, 1)

        # Handle Init param
        symname = "%s_%sinit"%(str(domain), str(species))
        if isinstance(init, scalar):
            init = create_parameter(init, "init", ge=0, property_param=False, \
                                    symname=symname)
        elif isinstance(init, TypeCheck) and init.is_scalar:
            init = create_parameter(init, "init", ge=0, property_param=False, \
                                    symname=symname)
        elif isinstance(init, TypeCheck) and init.is_node_param:
            enforce_and_check_node_param(init, "init", "FieldSpecies",\
                                         domain)
            init = create_parameter(init, "init", property_param=False, \
                                    allow_array=True, symname=symname)
        elif isinstance(init, sympy.Basic):
            for atom in init.atoms():
                if not isinstance(atom, ParSymbol):
                    continue
                enforce_and_check_node_param(atom.param, "init", \
                                             "FieldSpecies", domain)
            init = create_parameter(init, "init", property_param=False, \
                                    allow_array=True, symname=symname)
        else:
            raise TypeError("expected a scalar parameter or a NodeParam "\
                            "for the 'init' parameter.")
        
        params["init"] = init

        self = DomainSpecificSpecies.__new__(cls, species, domain, params)

        # Create a placeholder for Species at a MeshBoundary
        for boundary in mesh_boundaries(domain.boundaries):
            setattr(self, str(boundary), MeshBoundarySpecies(self, boundary))
        
        return self

    def _connect(self, receiver, boundary=None):
        "Store connection information between the species and a reciever"
        from boundaries import MeshBoundary
        check_arg(receiver, Receiver, 0)

        # If no boundary we connect to the average value instead of
        # BoundaryValue
        if boundary is None:
            DomainSpecificSpecies._connect(self, receiver)
            return 
        
        check_arg(boundary, MeshBoundary, 1)
        
        if boundary not in self.domain.boundaries:
            raise ValueError, "the boundary is not in the '%s' "\
                  "domain"%str(self.domain)
        
        # Store the connection
        self.connections.append(Connection(FieldSender(self, boundary),\
                                           receiver))

class ScalarFieldSpecies(FieldSpecies):
    """
    A domain specific species declared to be a scalar field defined on a mesh
    """
    def __new__(cls, species, domain, init=0.0, sigma=0.0, **params):
        # If sigma is a TypeCheck check that it is either a scalar or array
        symname = "%s_%s_sigma"%(str(domain), str(species))
        if isinstance(sigma, scalar):
            sigma = create_parameter(sigma, "sigma", ge=0, \
                                     property_param=False, symname=symname)
        elif isinstance(sigma, TypeCheck) and sigma.is_scalar:
            sigma = create_parameter(sigma, "sigma", ge=0,\
                                     property_param=False, symname=symname)
        elif isinstance(sigma, TypeCheck) and sigma.is_cell_param:
            enforce_and_check_cell_param(sigma, "sigma", "ScalarFieldSpecies",\
                                         domain)
            sigma = create_parameter(sigma, "sigma", property_param=False, \
                                     allow_array=True, symname=symname)
        elif isinstance(sigma, sympy.Basic):
            for atom in sigma.atoms():
                if not isinstance(atom, ParSymbol):
                    continue
                enforce_and_check_cell_param(atom.param, "sigma", \
                                             "ScalarFieldSpecies", domain)
            sigma = create_parameter(sigma, "sigma", property_param=False,\
                                     allow_array=True, symname=symname)
                    
        else:
            raise TypeError("expected a scalar parameter, a parameter "\
                            "expression or a CellParam for the 'sigma' "\
                            "parameter.")
        
        # Check user defined params for CellParams and set domain in
        # any domain-undefined NodeParams
        for paramname, param in params.iteritems():
            
            # Rule out CellParams from any expressions
            if isinstance(param, sympy.Basic):
                for atom in init.atoms():
                    if not isinstance(atom, ParSymbol):
                        continue
                    enforce_and_check_node_param(atom.param, paramname, \
                                                 "ScalarFieldSpecies", \
                                                 domain)
            
            # Rule out the use of CellParams directly as a parameter
            enforce_and_check_node_param(param, paramname, \
                                         "ScalarFieldSpecies", domain)

        params = user_defined_parameters(params, allow_array=True, \
                                         symnameprefix="%s_%s"%\
                                         (str(domain), str(species)))
        params["sigma"] = sigma
        
        return FieldSpecies.__new__(cls, species, domain, init, params)

class FixedScalarFieldSpecies(FieldSpecies):
    """
    A domain specific species declared to be a fixed scalar
    field defined on a mesh"""
    def __new__(cls, species, domain, init=0.0, **params):
        
        # Check user defined params for CellParams and set domain in
        # any domain-undefined NodeParams
        for paramname, param in params.iteritems():
            
            # Rule out CellParams from any expressions
            if isinstance(param, sympy.Basic):
                for atom in init.atoms():
                    if not isinstance(atom, ParSymbol):
                        continue
                    enforce_and_check_node_param(atom, paramname, \
                                                 "FixedFieldSpecies", \
                                                 domain)
            
            # Rule out the use of CellParams directly as a parameter
            enforce_and_check_node_param(param, paramname, \
                                         "FixedFieldSpecies", domain)
        
        params = user_defined_parameters(params, allow_array=True, \
                                         symnameprefix="%s%s"%\
                                         (str(domain), str(species)))
        
        return FieldSpecies.__new__(cls, species, domain, init, params)

class ConstSpecies(DomainSpecificSpecies):
    "A domain specific species declared to be a constant"
    # FIXME: Should we turn this into a Parmeter? It will then be "folded"
    # as other parameters
    def __new__(cls, species, domain, init=0.0, **params):
        check_arg(species, Species, 0)
        params = user_defined_parameters(params, \
                                         symnameprefix="%s_%s"%\
                                         (str(domain), str(species)))
        symname = "%s_%s_init"%(str(domain), str(species))
        params["init"] = create_parameter(init, "init", ge=0, \
                                          property_param=False, \
                                          symname=symname)

        return DomainSpecificSpecies.__new__(cls, species, domain, params)

class ScalarSpecies(DomainSpecificSpecies):
    "A domain specific species declared to be a scalar"
    def __new__(cls, species, domain, init=0.0, **params):
        check_arg(species, Species, 0)
        params = user_defined_parameters(params, \
                                         symnameprefix="%s_%s"%\
                                         (str(domain), str(species)))
        symname = "%s_%s_init"%(str(domain), str(species))
        params["init"] = create_parameter(init, "init", ge=0, \
                                          property_param=False,
                                          symname=symname)
        return DomainSpecificSpecies.__new__(cls, species, domain, params)

class Species(SubCellModelObject):
    def __init__(self, name, valence=1, **params):
        check_arg(name, str, 0)
        params = user_defined_parameters(params, symnameprefix=name)
        params["valence"] = create_parameter(valence, "valence", ge=-3, le=3)
        self.domains = []
        SubCellModelObject.__init__(self, name, params)

class MeshBoundarySpecies(SubCellModelObject):
    """
    A Class for making it more easy to establish connection between a
    ScalarFieldSpecies residing on a MeshBoundary
    """
    def __init__(self, domain_specific_species, boundary):
        from boundaries import MeshBoundary
        check_arg(domain_specific_species, DomainSpecificSpecies, 0)
        check_arg(boundary, MeshBoundary, 1)
        self.domain_specific_species = domain_specific_species
        self.species = domain_specific_species.species
        self.boundary = boundary
        SubCellModelObject.__init__(self, "%s-%s-%s"%(\
            domain_specific_species.domain, boundary, self.species))
