__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-16 -- 2011-10-04"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["SubCellModel"]

# SubCell imports
from subcell.common import *
from domains import *
from boundaries import *
from species import *
from discretemarkovmodel import *
from variables import *
from builtinvariables import TimeVariable, DtTimeVariable, TStop

class SubCellModel(object):
    def __init__(self, name):
        self._finalized = False

        check_arg(name, str)
        self.name = name
        self.meshes = []
        self.domains = []
        self.boundaries = []
        self.boundary_values = []
        self.all_values_at_boundaries = []
        self.fluxes = []
        self.reactions = []
        self.species = odict()
        self.containing_species = odict()
        self.const_species = []
        self.scalar_species = []
        self.scalar_field_species = []
        self.field_species = []
        self.fixed_scalar_field_species = []
        self.discrete_markov_models = []
        self.continuous_variables = []
        self.discrete_variables = []
        self.discrete_markov_model_connections = odict()
        self.species_connections = []
        self.variable_connections = []
        self.params = odict()
        self.marker_to_mesh_domain = odict()
        self.distance_to_boundaries = []

        # Instantiate the time variables
        self.continuous_variables.append(TimeVariable("time"))
        self.continuous_variables.append(DtTimeVariable("dt"))
        self.discrete_variables.append(TStop("tstop"))

    def finalized(self):
        """
        Return True if the SubCellModel is finalized and ready to
        be passed to a TimeStepper. Only SubCellModelEditors can
        properly finalize a SubCellModel.
        """
        
        return self._finalized
        
    def __eq__(self, other):
        assert(isinstance(other, SubCellModel))
        return self.name == other.name
    
    def _list_of_domains(self, DomainType):
        "Returns a list of a particulare Domain type"
        assert(issubclass(DomainType, Domain))
        return [domain for domain in self.domains if isinstance(domain, \
                                                                DomainType)]

    def _num_of_domains(self, DomainType):
        "Return the number of a particulare Domain type"
        return len(self._list_of_domains(DomainType))

    def iter_gated_boundaries(self):
        "Return an iterator over gated Boundaries"
        return (boundary for boundary in self.boundaries \
                if boundary.isgated)

    def iter_gated_mesh_boundaries(self):
        "Return an iterator over gated mesh Boundaries"
        return (boundary for boundary in self.iter_mesh_boundaries() \
                if boundary.isgated)

    def iter_none_gated_mesh_boundaries(self):
        "Return an iterator over the mesh Boundaries that is not gated"
        return (boundary for boundary in self.iter_mesh_boundaries() \
                if not boundary.isgated)

    def iter_none_mesh_boundaries(self):
        "Return an iterator over the non mesh Boundaries"
        return none_mesh_boundaries(self.boundaries)        

    def iter_mesh_boundaries(self):
        "Return an iterator over all mesh Boundaries"
        return mesh_boundaries(self.boundaries)
    
    def iter_all_values_at_boundaries(self):
        for domain in self.mesh_domains():
            for boundary in domain.boundaries:
                for species in domain.iter_scalar_field_species():
                    for ind in xrange(boundary.num):
                        yield BoundaryValue(species.species, boundary, \
                                            domain, ind)
                for species in domain.fixed_scalar_field_species:
                    for ind in xrange(boundary.num):
                        yield BoundaryValue(species.species, boundary, \
                                            domain, ind)

    # FIXME: Make this the method to acces MixedFunctionSpaces variables
    # FIXME: Use this method to control how the different species get allocated
    # FIXME: dofs in the MixedFunction
    def iter_scalar_field_species(self):
        """
        Iterate over all scalar field species

        We first iterate over all species declared using the first sub space
        of the MixedFunctionSpace. This means we iterate over all domains whos
        first subspace is used to declare ScalarFieldSpecies.

        Either None or a ScalarFieldSpecies is yeilded
        """
        for domain in self.mesh_domains():
            for species in domain.iter_scalar_field_species():
                yield species
        
    def max_scalar_fields(self):
        "Return the maximal number of scalar fields in any single MeshDomain"
        max_scalar_fields = 0
        for domain in self.mesh_domains():
            max_scalar_fields = max(domain.num_scalar_field_species(), \
                                    max_scalar_fields)
        return max_scalar_fields

    def mesh_domains(self):
        "Returns a list of all registered MeshDomains"
        return self._list_of_domains(MeshDomain)
    
    def volume_domains(self):
        "Returns a list of all registered VolumeDomains"
        return self._list_of_domains(VolumeDomain)

    def infinite_domains(self):
        "Returns a list of all registered InifiniteDomains"
        return self._list_of_domains(InfiniteDomain)

    def num_mesh_domains(self):
        "Returns the number of registered MeshDomains"
        return self._num_of_domains(MeshDomain)

    def num_volume_domains(self):
        "Returns the number of registered VolumeDomains"
        return self._num_of_domains(VolumeDomain)

    def num_infinite_domains(self):
        "Returns the number of registered InfiniteDomains"
        return self._num_of_domains(InfiniteDomain)

    def variable(self, name):
        """
        Return a variable with the name
        """
        check_arg(name, str)
        all_variables = self.continuous_variables + self.discrete_variables
        variables = [var for var in all_variables if var.name == name]
        if not variables:
            raise RuntimeError("No variable with name '%s' registered"%name)
        return variables[0]

    def _list_of_fluxes(self, SpeciesSubclass):
        "Returns a list of all fluxes flowing to species of a particular class"
        assert(issubclass(SpeciesSubclass, DomainSpecificSpecies))
        fluxes = []
        
        for flux in self.fluxes:
            if any(isinstance(domain.species[flux.species], SpeciesSubclass)
                   for domain in flux.boundary.domains):
                fluxes.append(flux)
        return fluxes
        
    def scalar_fluxes(self):
        "Return all fluxes that flows to scalar species"
        return self._list_of_fluxes(ScalarSpecies)

    def scalar_field_fluxes(self):
        "Return all fluxes that flows to scalar field species"
        return self._list_of_fluxes(ScalarFieldSpecies)

    def add_discrete_markov_model(self, name):
        "Add a DiscreteMarkovModel to the cell model"

        self._add_check()
            
        mm = DiscreteMarkovModel(self, name)
        self.discrete_markov_models.append(mm)
        return mm

    def __repr__(self):
        return "SubCellModel('%s')"%self.name

