__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-01-14 -- 2011-10-26"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["SubCellModelEditor"]

# System imports
import inspect
import types
import copy

# SubCell imports
from subcell.common import *
from subcell.common.logger import *
from subcell.model.subcellmodel import SubCellModel
from subcell.model.domains import *
from subcell.model.boundaries import *
from subcell.model.fluxes import *
from subcell.model.species import *
from subcell.model.reactions import Reaction
from subcell.model.discretemarkovmodel import *
from subcell.model.builtinvariables import *
from subcell.model.variables import *
from subcell.model.distancetoboundaries import *

def _register_name(name, value=None, frames=1):
    "Register a name in the namespace if not that name is already declared."
    assert(isinstance(name, str))
    
    value = value or name

    # We go back some frames
    frame = inspect.currentframe()
    for i in range(frames):
        frame = frame.f_back
    
    if str(name) in frame.f_globals:
        raise ValueError, "Cannot declare a variable with name '%s'. The "\
              "name already exists."%str(name)
    
    try:
        debug("adding '%s' to namespace", name)
        frame.f_globals[str(name)] = value
    finally:
        del frame

class SubCellModelEditor(object):
    def __init__(self, model=None):
        """
        Create a SubCellModelEditor
        """

        self.subcellmodel = None
        
        if model is not None:
            if isinstance(model, str):
                subcellmodel = SubCellModel(model)
            if isinstance(model, SubCellModel):
                if model.finalized():
                    raise TypeError, "expected the SubCell model to be not "\
                          "finalized"
                subcellmodel = model
            self.open(subcellmodel)

    def open(self, subcellmodel):
        """
        Open a SubCellModel for editing
        """
        check_arg(subcellmodel, SubCellModel)
        if self.subcellmodel is not None and not \
               self.subcellmodel.finalized():
                raise TypeError, "A subcellmodel is still open. Close that "\
                      "before opening another one."
        
        if subcellmodel.finalized():
            raise TypeError, "expected the SubCellModel to not be finalized"
        self.subcellmodel = subcellmodel
        
    def close(self):
        """
        Close the SubCellModel for editing
        """
        self._finalize()
        
    def __register_domain(self, domain):
        "A helper method to register a specific domain"
        assert(isinstance(domain, Domain))
        if domain in self.subcellmodel.domains:
            raise ValueError, "domain with name '%s' already exists"%domain
        self.subcellmodel.domains.append(domain)
        return domain

    def _add_check(self):
        "To be called when adding new object to check for is finalized"
        if self.subcellmodel.finalized():
            raise RuntimeError("Subcell model is finalized and no extra "\
                               "objects can be added.")

    def ModelParameters(self, **params):
        """
        Declare parameters to the model

        Example:
        ========

        >>> param1, param2 = Parameters(param1=1, \\
        >>>                             param2=RangeCheck(0.1, ge=0, le=10))

        @return: The declared parameters as a list
        
        """

        self._add_check()
        
        for name, value in params.iteritems():
            if name in self.subcellmodel.params:
                raise ValueError, "parameter '%s' is already a registered"\
                      " parameter"%name
            
            if not isinstance(value, TypeCheck):
                value = TypeCheck(value)
            
            if isinstance(value, (NodeParam, CellParam)):
                raise TypeError("NodeParams or CellParams cannot be "\
                                "declared using ModelParameters")
            
            # Assign a name to the variable
            value.name=(name, name)
            
            # Create a sympy Parameter and register it in the model namespace
            _register_name(name, value, 2)

            # Store parameter in global dict
            self.subcellmodel.params[name] = value

    def InfiniteDomains(self, *args):
        """
        Declare several InfiniteDomains

        Example:
        ========

        >>> Domain1, Domain2 = InfiniteDomains("Domain1", "Domain2")

        @rtype: InfiniteDomains
        @return: A list of the declared InfiniteDomains
        """

        self._add_check()
        
        domains = []
        for i, name in enumerate(args):
            check_arg(name, str, i)
            domains.append(self.__register_domain(InfiniteDomain(name)))
        if len(domains) ==1:
            return domains[0]
        return tuple(domains)

    def InfiniteDomain(self, name, **params):
        """
        Declare an InfiniteDomain with optional parameter associated to
        the Domain
        
        Example:
        ========
        
        >>> domain = InfiniteDomain("domain", param=1)
        >>> param = domain.param

        @type name : str
        @param name : The name of the domain
        @rtype: InfiniteDomain
        @return: The declared domain
        
        """

        self._add_check()
        
        return self.__register_domain(InfiniteDomain(name, **params))

    def VolumeDomain(self, name, volume, **params):
        """
        Declare a VolumeDomain with optional parameter associated to the Domain
        
        Example:
        ========
        
        >>> domain = VolumeDomain("domain", 1000, param=1)
        >>> param = domain.param

        @type name : str
        @param name : The name of the domain
        @type volume : A positive scalar
        @param volume : The volume of the domain
        @rtype: VolumeDomain
        @return: The declared domain
        
        """

        self._add_check()
        
        return self.__register_domain(VolumeDomain(name, volume, **params))

    def Mesh(self, filename):
        """
        Declare a Mesh, which is associated with a file

        Example:
        ========
        
        >>> mesh = Mesh("mesh")

        @type filename : str
        @param filename : The basename of mesh filename
        @rtype: Mesh
        @return: The declared Mesh
        
        """

        self._add_check()
        
        if self.subcellmodel.meshes:
            raise NotImplementedError, "can only have one mesh per model"
        mesh = Mesh(filename)
        self.subcellmodel.meshes.append(mesh)
        return mesh
    
    def MeshDomain(self, name, marker, mesh, **params):
        """
        Declare a MeshDomain, which is associated with a Mesh

        Example:
        ========

        >>> domain = MeshDomain("domain", 2, mesh, param=1)
        >>> param = domain.param

        @type name : str
        @param name : The name of the domain
        @type marker : A positive integer
        @param marker : The marker defining a domain in the mesh
        @type mesh : A Mesh
        @param mesh : The mesh associated to the domain
        @rtype: MeshDomain
        @return: The declared domain
        
        """

        # FIXME: Find a way to parametrize the mesh...
        assert(marker>=0)
        
        return self.__register_domain(MeshDomain(name, marker, mesh, \
                                                 **params))
        
    def Species(self, name, valence=1, containing=None, **params):
        """
        Declare a Species

        Example:
        ========

        >>> BCa = Species("BCa", 2, Ca, kon=.1, koff=1.)
        >>> koff = BCa.koff

        @type name : str
        @param name : The name of the species
        @type valence : A positive integer
        @param valence : The valence of the species
        @type containing : Species
        @param containing : The Species contained in this species. This can be
        used to calculate conservation of Species. I{Optional argument}
        @rtype: Species
        @return: The declared species
        
        """

        self._add_check()
        species = Species(name, valence, **params)
        if species in self.subcellmodel.species:
            raise ValueError, "a species with name '%s' is already"\
                  " added in the '%s' model."%(name, self.subcellmodel.name)
        
        # Check for containing species (Used for conservation log)
        if containing is not None:
            check_kwarg(containing, "containing", (Species, tuple))
            containing = tuplewrap(containing)
            for cont in containing:
                if not isinstance(cont, Species):
                    raise TypeError, "expected each entry in the passed tuple"\
                          " of 'containing' to be a Species"
                if cont not in self.subcellmodel.species:
                    raise TypeError, "expected Species in the 'containing' "\
                          "argument to be an already declared Species"
                self.subcellmodel.containing_species[cont].append(species)
                
        # Register species
        self.subcellmodel.species[species] = []
        self.subcellmodel.containing_species[species] = [species]
        
        return species

    def Boundary(self, name, domain, area=1.0, dx=1.0, **params):
        """
        Declare a Boundary B{not} associated with a MeshDomain.
        
        Example:
        ========

        >>> channel = Boundary("channel", (domain0, domain1), area, 2.0)

        Flux direction:
        ===============

        If the boundary connects two domains, will the order of the domains
        define the direction of a flux, which is declared on the boundary.
        In the above example will the direction of a flux go from C{domain0} to
        C{domain1}.

        @type name : str
        @param name : The name of the Boundary
        @type domain : An InfiniteDomain or a VolumeDomain
        @param domain : One or two domains which the Boundary connects
        @type area : A positive scalar
        @param area : The area of this boundary
        @type dx : A positive scalar
        @param dx : The distance between two domains
        @rtype: Boundary
        @return: The declared boundary
        
        """
        
        self._add_check()
        
        # FIXME: Change order of num and area, following the changes in
        # editor method
        boundary = Boundary(name, domain, area, dx, **params)
        self.subcellmodel.boundaries.append(boundary)
        return boundary
    
    def MeshBoundary(self, name, domain, startmarkers, dx=1.0, num=1, \
                     **params):
        """
        Declare a MeshBoundary associated with a MeshDomain
        
        Example:
        ========
        
        >>> channel = MeshBoundary("channel", (domain0, domain1), \\
        >>>                        (10, 20), 2.0, 3)
        
        Flux direction:
        ===============
        
        If the boundary connects two domains, will the order of the domains
        define the direction of a flux, which is declared on the boundary.
        In the above example will the direction of a flux go from C{domain0} to
        C{domain1}.

        @type name : str
        @param name : The name of the boundary
        @type domain : An InfiniteDomain, VolumeDomain or a MeshDomain, at
        least one need to be a MeshDomain
        @param domain : One or two domains which the Boundary connects
        @type startmarkers : A positive integer or a tuple of two positive
        integers
        @param startmarkers : The markers associated to the Boundaries. 
        @type dx : A positive scalar
        @param dx : The distance between two domains
        @type num : int
        @param num : The number of Boundaries associated. I{Optional argument}
        with this boundary
        @rtype: MeshBoundary
        @return: The declared boundary
        """

        self._add_check()
        
        boundary = MeshBoundary(name, domain, startmarkers, dx, num, **params)
        self.subcellmodel.boundaries.append(boundary)
        return boundary

    def DistanceToBoundaries(self, domain, *args):
        """
        Create an object which can be used to define a closest distance to 
        a set of MeshBoundaries

        Example:
        ========
        
        >>> channel0 = MeshBoundary("channel0", domain, 10)
        >>> channel1 = MeshBoundary("channel1", domain, 20)
        >>> boundary_distance = DistanceToBoundaries(domain, channel0, \
                                                     channel1)
        >>> boundary_distance.node_dist
        
        @rtype: DistanceToBoundaries
        @return: A DistanceToBoundaries object
        """
        
        self._add_check()
        obj = DistanceToBoundaries(domain, *args)
        
        self.subcellmodel.distance_to_boundaries.append(obj)
        return obj
        
    def DiscreteMarkovModel(self, name, minimize_dt=False, \
                                  end_sim=False):
        """
        Declare a DiscreteMarkovModel

        Example:
        ========

        >>> activator = DiscreteMarkovModel("activator", True)

        @type name : str
        @param name : The name of the discrete Markov model. This name need to
        correspond to a model defined elsewhere, for example a Gillstep file.
        @type minimize_dt : bool
        @param minimize_dt : If True, minimizes the time step whenever the
        markov model change value. I{Optional argument} 
        @type end_sim : bool
        @param end_sim : If True, the simulation will come to a stop, before
        tstop. I{Optional argument}
        @rtype: DiscreteMarkovModel
        @return: The discrete Markov model
        """

        self._add_check()
            
        mm = DiscreteMarkovModel(name, minimize_dt, end_sim)
        self.subcellmodel.discrete_markov_models.append(mm)
        return mm

    def DiscreteMarkovSubModel(self, discrete_markov_model, name, num=1):
        """
        Declare a DiscreteMarkovSubModel on a parent DiscreteMarkovModel

        Example:
        ========

        >>> activator = DiscreteMarkovModel("activator", True)
        >>> sub_activator = DiscreteSubMarkovModel(activator, \\
        >>>                 "sub_activator", 2)

        @type discrete_markov_model : DiscreteMarkovModel
        @param discrete_markov_model : The sub model will get declared within
        this parent MarkovModel.
        @type name : str
        @param name : The name of the discrete sub Markov model. This name
        need to correspond to a sub model of a discrete markov model defined
        elsewhere, for example in a Gillstep file.
        @type num : int
        @param num : The number of sub models for each parent model
        @rtype: DiscreteMarkovSubModel
        @return: The discrete Markov sub model
        """

        self._add_check()
        check_arg(discrete_markov_model, DiscreteMarkovModel, 0)
        check_arg(name, str, 1)
        check_arg(num, int, 2)
        return discrete_markov_model._add_submodel(name, num)

    def ScalarFieldSpecies(self, domain, species, init, sigma=1.0, \
                           **params):
        """
        Declare a scalar field species in a MeshDomain.
        
        A scalar field species is a diffusive species who's value varies in
        space and time. The species is added as an attribute to the domain,
        and the initial value is available as an attribute to the species.

        Example:
        ========

        >>> mesh = Mesh("mesh")
        >>> domain = MeshDomain(mesh, 2)
        >>> Ca = Species("Ca", 2)
        >>> ScalarFieldSpecies(domain, Ca, 1.0, 5.0, kon=5)
        >>> Ca_init = domain.Ca.init
        >>> Ca_kon = domain.Ca.kon

        @type domain : MeshDomain
        @param domain : The domain which the scalar field species is
        declared in
        @type species : Species
        @param species : The Species which are declared on the domain.
        @type init : Positive scalar or Parameter
        @param init : The initial value of the scalar field. This parameter
        can be a parameter previously defined.
        @type sigma : Positive scalar or Parameter
        @param sigma : The diffusion constant of the scalar field. This
        parameter can be a parameter previously defined.
        @param params: Optional parameters are declared as attributes to the
        species, see example above.

        """
        # Add a default scale factor
        check_arg(domain, MeshDomain, 0)
        check_arg(species, Species, 1)
        domain.scale[species] = 1
        domain._add_species(species, ScalarFieldSpecies(species, domain, init,\
                                                        sigma, **params))
        
    def FixedScalarFieldSpecies(self, domain, species, init, **params):
        """
        Declare a fixed field species in a MeshDomain.
        
        A scalar field species is a species who's value varies in space and
        time. It is fixed and does not diffuse. The species is added as an
        attribute to the domain, and the initial value is available as an
        attribute to the species.

        Example:
        ========

        >>> mesh = Mesh("mesh")
        >>> domain = MeshDomain("domain", mesh, 2)
        >>> BCa = Species("BCa")
        >>> FixedScalarFieldSpecies(domain, BCa, 4.0, kon=5)
        >>> BCa_init = domain.BCa.init
        >>> BCa_kon = domain.BCa.kon

        @type domain : MeshDomain
        @param domain : The domain which the scalar field species is
        declared in
        @type species : Species
        @param species : The Species which are declared on the domain
        @type init : Positive scalar or Parameter
        @param init : The initial value of the scalar field. This parameter can
        be a parameter previously defined
        can be a parameter previously defined
        @param params: Optional parameters are declared as attributes to the
        species, see example above.

        """

        # Add a default scale factor
        check_arg(domain, MeshDomain, 0)
        check_arg(species, Species, 1)
        domain.scale[species] = 1
        domain._add_species(species, FixedScalarFieldSpecies(\
            species, domain, init, **params))
        
    def ScalarSpecies(self, domain, species, init, **params):
        """
        Declare a scalar species in a MeshDomain or a VolumeDomain.
        
        A scalar species is a species who's value varies in time. The species
        is added as an attribute to the domain, and the initial value is
        available as an attribute to the species.

        Example:
        ========

        >>> domain = VolumeDomain("domain", 1000)
        >>> Ca = Species("Ca", 2)
        >>> ScalarSpecies(domain, Ca, 1.0, tot=5)
        >>> Ca_init = domain.Ca.init
        >>> Ca_tot = domain.Ca.tot

        @type domain : MeshDomain, VolumeDomain
        @param domain : The domain which the scalar field species is
        declared in
        @type species : Species
        @param species : The Species which are declared on the domain
        @type init : Positive scalar or Parameter
        @param init : The initial value of the scalar field. This parameter can
        be a parameter previously defined
        @param params: Optional parameters are declared as attributes to the
        species, see example above.
        
        """
        check_arg(domain, (VolumeDomain, MeshDomain), 0)
        check_arg(species, Species, 1)
        domain._add_species(species, ScalarSpecies(species, domain,\
                                                   init, **params))
        
    def ConstSpecies(self, domain, species, init, **params):
        """
        Declare a const species in a MeshDomain, a VolumeDomain or in an
        InfiniteDomain
        
        A const species is a species who's value do not varies in
        time. The species is added as an attribute to the domain, and
        the value is available as an attribute to the species. 

        Example:
        ========

        >>> domain = VolumeDomain("domain", 1000)
        >>> Mg = Species("Mg", 2)
        >>> ConstSpecies(domain, Mg, 100.0, tot=5)
        >>> Mg_init = domain.Mg.init
        >>> Mg_tot = domain.Mg.tot

        @type domain : MeshDomain, VolumeDomain, InfiniteDomain
        @param domain : The domain which the const species is declared in
        @type species : Species
        @param species : The Species which are declared on the domain
        @type init : Positive scalar or Parameter
        @param init : The constant value of the species. This parameter can
        be a parameter previously defined.
        @param params: Optional parameters are declared as attributes to the
        species, see example above.
        
        """
        check_arg(domain, (InfiniteDomain, VolumeDomain, MeshDomain), 0)
        check_arg(species, Species, 1)
        domain._add_species(species, ConstSpecies(species, domain,\
                                                  init, **params))

    def Reaction(self, domain, expr, reactants=None, products=None):
        """
        Declare a reaction on a domain
        
        A reaction can be defined between species declared on a volume
        or a mesh domain.
        
        Example:
        ========
        
        >>> domain = VolumeDomain("domain", 1000)
        >>> Ca = Species("Ca", 2)
        >>> BCa = Species("BCa", kon=5., koff=1., tot=10)
        >>> ScalarSpecies(domain, Ca, 1.0, 5.0)
        >>> ScalarSpecies(domain, BCa, 2.0, 1.0)
        >>> Reaction(domain, BCa.kon*domain.Ca*(BCa.tot-domain.BCa)\\
        >>>          -BCa.koff*BCa, reactants=Ca, products=BCa)
        
        @type domain : MeshDomain, VolumeDomain
        @param domain : The domain which the reaction is declared in
        @type expr : A math expression
        @param expr : The expression which defines the reaction. Domain
        specific species which are declared on the domain can be used in
        the expression.
        @type reactants : Species, list
        @param reactants : The reactants of the reaction. These can be defined
        using:
         - a single species: C{reactants=Ca}, or
         - a list of species C{reactants=[Ca, Mg]}, or
         - a list of tuples each with a length 2, where the first element
           defines the species and the second a positive integer which defines
           the stoichimoetry, C{reactants=[(Ca, 1), (Mg, 1)]}
        @type products : Species, list
        @param products : The products of the reaction. These can be defined
        using:
         - a single species: C{products=Ca}, or
         - a list of species C{products=[Ca, Mg]}, or
         - a list of tuples each with a length 2, where the first element
           defines the species and the second a positive integer which defines
           the stoichimoetry, C{products=[(Ca, 1), (Mg, 1)]}
        """
        check_arg(domain, (VolumeDomain, MeshDomain), 0)
        reaction = Reaction(domain, expr, reactants, products)
        
    def Flux(self, fluxname, boundary, species, expr, limits=None, \
             epot_scale_all=False):
        """
        Declare a flux on a boundary
        
        A flux can be defined using species declared on the domains the
        boundary connects.
        
        Example:
        ========
        
        >>> domain0 = VolumeDomain("domain0", 1000)
        >>> domain1 = VolumeDomain("domain1", 1000)
        >>> boundary = Boundary("boundary", (domain0, domain1), 1, 10.,\\
        >>>                     Kd=RangeCheck(1., gt=0))
        >>> Ca = Species("Ca", 2)
        >>> ScalarSpecies(domain0, Ca, 1.0, 5.0)
        >>> ScalarSpecies(domain1, Ca, 10.0, 5.0)
        >>> Flux("Pump", boundary, Ca, 5/(1+(boundary.Kd/domain0.Ca)**2)
        
        @type fluxname : str
        @param fluxname : The name of the flux
        @type boundary : Boundary, MeshBoundary
        @param boundary : The boundary which the flux is defined on
        @type species : Species
        @param species : The species the flux involves
        @type expr : A math expression
        @param expr : The expression which defines the reaction. Domain
        specific species which are declared on the domains connected by
        the boundary can be used in the expression.
        @type limits : None
        @param limits : A limit which can be provided for singular values of
        the flux (Not implemented)
        @type epot_scale_all : bool
        @param epot_scale_all : If True will species used in the expression,
        which are defined on neighboring domain to a domain using a electric
        potential, be scaled according to the potential.
        @rtype: Flux
        @return: The flux
        """
        check_arg(fluxname, str, 0)
        check_arg(boundary, BoundaryBase, 1)
        check_arg(species, Species, 2)
        if limits is not None:
            raise NotImplemented("'limits' are not implemented")
        return Flux(fluxname, boundary, species, expr, limits, epot_scale_all)
        
    def DiffusionFlux(self, boundary, species, sigma, epot_scale_all=False):
        """
        Declare a diffusion flux on a boundary
        
        A diffusion flux defines a flux of linear diffusion between two domains
        
        Example:
        ========
        
        >>> domain0 = VolumeDomain("domain0", 1000)
        >>> domain1 = VolumeDomain("domain1", 1000)
        >>> boundary = Boundary("boundary", (domain0, domain1), 1, 10.,\\
        >>>                     Kd=RangeCheck(1., gt=0))
        >>> Ca = Species("Ca", 2)
        >>> ScalarSpecies(domain0, Ca, 1.0, 5.0)
        >>> ScalarSpecies(domain1, Ca, 10.0, 5.0)
        >>> DiffusionFlux(boundary, Ca, 5.0)
        
        @type boundary : Boundary, MeshBoundary
        @param boundary : The boundary which the flux is defined on
        @type species : Species
        @param species : The species the flux involves
        @type sigma : Scalar, Parameter
        @param sigma : The sigma of the flux
        @type epot_scale_all : bool
        @param epot_scale_all : If True will species used in the expression,
        which are defined on neighboring domain to a domain using a electric
        potential, be scaled according to the potential.
        @rtype: DiffusionFlux
        @return: The flux
        """
        check_arg(boundary, BoundaryBase, 0)
        check_arg(species, Species, 1)
        return DiffusionFlux(boundary, species, sigma, epot_scale_all)
        
    def StaticElectricPotential(self, domain, expr):
        """
        Declare a static electric potential to a domain

        Example:
        ========
        
        >>> mesh = Mesh("mesh")
        >>> domain = MeshDomain(mesh, 2)
        >>> StaticElectricPotential(domain, exp(-z))
        
        @type domain : MeshDomain
        @param domain : The domain the potential is defined on
        @expr : A mathematical expression defining the static potential.
        Coordinates, such as, x, y, z can be used to define the potential.
        The expression can also be a NumPy array defining the potential at
        each vertex.
        
        """
        check_arg(domain, MeshDomain, 0)
        domain.add_static_electric_potential(expr)
    
    def ValueClamp(self, name, times, value_list, minimize_dt=False):
        """
        Declare a ValueClamp variable to the cell model
    
        Example:
        ========
        
        >>> domain = VolumeDomain("domain", 1000.)
        >>> boundary = Boundary("boundary", domain, 1, 10)
        >>> activator = ValueClamp("activator", [0, 5, 10], [0, 1, 0])
        >>> Connection(activator, boundary)
        >>> Flux(boundary, 2.0)
        
        @type name : String
        @param name : The name of the ValueClamp
        @type times : list of positive scalars
        @param times : A list of times that corresponds to the time the value
        clamp will change
        @type value_list : list of scalars
        @param value_list : A list of values that corresponds to the values the
        value clamp will change to
        @rtype: ValueClamp
        @return: The value clamp
        """
        
        self._add_check()
    
        variable = ValueClamp(name, times, value_list, minimize_dt)
        self.subcellmodel.discrete_variables.append(variable)
        return variable
    
    def EventStopper(self): 
        """
        Declare an EventStopper to the cell model. The EventStopper will end
        a simulation if all connected discrete variables (for example a
        DiscreteMarkovModel) are zeros.
    
        Example:
        ========

        >>> markov_model = DiscreteMarkovModel("markov_model")
        >>> event_stopper = EventStopper()
        >>> Connection(markov_model, event_stopper)
        
        @rtype: EventStopper
        @return: The event stopper
        """
        
        self._add_check()
    
        variable = EventStopper("EventStopper")
        self.subcellmodel.discrete_variables.append(variable)
        return variable

    def OneTimeValueCrosser(self, name, cutoff, send_values=[1,0], comp="<", \
                            minimize_dt=False):
        """
        Declare an OneTimeValueCrosser to the cell model

        OneTimeValueCrosser will send two values during a simulation.
        These values are given in the values argument. The first is sent
        at start, the second is sent whenever a receiving evaluate to true
        when compared to the cutoff value. The comp argument is a str which
        the user can define.
    
        Example:
        ========

        >>> value_crosser = OneTimeValueCrosser("value_crosser", 5)
        >>> channel = Boundary("channel", (upper, lower), 10)
        >>> ScalarSpecies(upper, Ca, 10)
        >>> Connection(value_crosser, channel)
        >>> Connection(upper.Ca, value_crosser)
        
        @rtype: OneTimeValueCrosser
        @return: The value crosser
        """
        self._add_check()
    
        variable = OneTimeValueCrosser(name, cutoff, send_values, comp, \
                                       minimize_dt)
        self.subcellmodel.discrete_variables.append(variable)
        return variable
    
    def Connection(self, sender, receiver):
        """
        Create a connection between a sender and a receiver
        
        Example:
        ========

        >>> mesh = Mesh("mesh")
        >>> domain = MeshDomain("domain", mesh, 2)
        >>> boundary0 = MeshBoundary("boundary0", domain, 10)
        >>> boundary1 = MeshBoundary("boundary1", domain, 20)
        >>>
        >>> activator0 = ValueClamp("activator0", [0, 5, 10], [0, 1, 0])
        >>> activator1 = MarkovModel("activator1")
        >>>
        >>> Ca = Species("Ca", 2)
        >>> ScalarFieldSpecies(domain, Ca, 10., 5.)
        >>>
        >>> Flux(boundary0, 2.0)
        >>> Flux(boundary1, 1.0)
        >>> 
        >>> Connection(activator0, boundary0)
        >>> Connection(activator1, boundary1)
        >>> Connection(domain.Ca.boundary1, activator1)
        
        In the example above we create three Connections. Two which controls
        the openness of two boundaries, and one which connect the value of the
        scalar field species at boundary1 to the Markov model. It is assumed
        that the Markov model can receive a variable called "Ca".
        
        """
        # FIXME: Put all connection logic in this method.
        # FIXME: Now it is all around the place...
        self._add_check()
        
        check_arg(sender, (MeshBoundarySpecies, DomainSpecificSpecies, \
                           DiscreteMarkovModel, Variable))
        check_arg(receiver, (MeshBoundary, Boundary, Variable,
                             DiscreteMarkovModel))
        
        if isinstance(sender, (MeshBoundarySpecies, DomainSpecificSpecies)):
            allowed_receivers = (DiscreteMarkovModel, Variable)
            if not isinstance(receiver, allowed_receivers):
                raise TypeError, "A 'DiscreteMarkovModel' or a 'Variable' is"\
                      " expected to be the receiver when the sender is a "\
                      "DomainSpecificSpecies."
            
            receiver = receiver._receiver(str(sender.species)) \
                       if isinstance(receiver, DiscreteMarkovModel) else \
                       receiver._receiver()
            
            if isinstance(sender, MeshBoundarySpecies):
                boundary = sender.boundary
                sender = sender.domain_specific_species
                sender._connect(receiver, boundary)
            else:
                sender._connect(receiver)
        elif isinstance(receiver, DiscreteMarkovModel):
            sender._connect(receiver._receiver(str(sender)))
        else:
            sender._connect(receiver._receiver())

    def __repr__(self):
        return "SubCellModelEditor('%s')"%self.subcellmodel.name

    def _remove_hidden_params(self, param_dict):

        assert(isinstance(param_dict, dict))
        
        # If a parameter starts with a "__" it is a hidden parameter
        for pname in param_dict.keys():
            if len(pname)<2:
                continue
            if pname[:2] == "__":
                param_dict.pop(pname)

                # If the hidden parameter is followed by a slave remove
                # that one too
                if pname[2:] in param_dict:
                    param_dict.pop(pname[2:])
            
        return param_dict

    
    def _rearange_field_species_order(self):
        """
        Re-arrange order of scalar field species
        """
        # Check that all MeshDomains include at least one ScalarFieldSpecies
        for domain in mesh_domains(self.subcellmodel.domains):
            if not any(isinstance(species, ScalarFieldSpecies) \
                       for species in domain.species.itervalues()):
                raise TypeError, "expected at least one ScalarFieldSpecies "\
                      "declared in MeshDomain: '%s'"%domain

        # Grab a list of all species. Use this as a template for the
        # species order for all domains
        def species_comp(species0, species1):
            return cmp(len(self.subcellmodel.species[species0]), \
                       len(self.subcellmodel.species[species1]))

        def domain_comp(domain0, domain1):
            return cmp(all_domain_species[domain0].count(None), \
                       all_domain_species[domain1].count(None))

        def print_list(l):
            return  [str(val) for val in l] 
        
        # Sort the species with the most frequent first
        canonical_order = [species for species, domain_species in \
                           self.subcellmodel.species.iteritems() 
                           if any(isinstance(dspecies, ScalarFieldSpecies) \
                                  for dspecies in domain_species)]
        debug("All species before ordering: %s", print_list(canonical_order))
        canonical_order.sort(species_comp, reverse=True)
        all_domain_species = {}
        all_domains = []
        for domain in self.subcellmodel.mesh_domains():
            
            # Fill a list corresponding to this domain
            if not domain.scalar_field_species:
                continue
            
            # Create a list of all domains with ScalarFieldSpecies
            all_domains.append(domain)
            all_domain_species[domain] = [None]*len(canonical_order)
            for species in domain.scalar_field_species:
                ind = canonical_order.index(species.species)
                all_domain_species[domain][ind] = species.species

            self.subcellmodel.marker_to_mesh_domain[domain.marker] = domain
        
        # Sort all domains so the one with most species gets first.
        all_domains.sort(domain_comp)
        largest_domain = all_domains[0]
        rest_domains = all_domains[1:]

        debug("All domains ordered after most species: %s", \
              print_list(all_domains))
        debug("Domain with most species: %s", largest_domain)
        debug("All species after ordering: %s", print_list(canonical_order))
        
        # If the domain with most species include a None then we check if
        # we can reduce the number of Species towards max_scalar_fields()
        if None in all_domain_species[largest_domain]:
            
            # Grab the not None Species 
            all_not_None_species = [species for species in all_domain_species[\
                largest_domain] if species is not None]
            
            all_None_species = [canonical_order[ind] for ind, species in \
                                enumerate(all_domain_species[largest_domain]) \
                                if species is None]
            
            # Iterate over the not None species and see if some of these can be
            # inserted into the None places, reducing the amount of space
            # need to resolve the Species
            # For this to happen that species need not to be included in all
            # other domains. If it is then the None_Species cannot be included
            # in the particular domain
            species_exchange = {}
            for not_None_species in reversed(all_not_None_species):
                for None_species in all_None_species:
                    # If we already have found an exchanger for this species
                    # continue
                    if None_species in species_exchange:
                        continue
                    for domain in rest_domains:
                        # If both species are in the domain then this
                        # combination of species is not mixable
                        if not_None_species in all_domain_species[domain] and \
                               None_species in all_domain_species[domain]:
                            break
                    # If for loop exits with no break we have a viable
                    # combination
                    else:
                        species_exchange[None_species] = not_None_species
                        # Call break so no more None_species can be combined
                        # with the not_None_species
                        break
            
            debug("Species exchange: %s", dict((str(key), str(value)) \
                                               for key, value in \
                                               species_exchange.items()))
            
            new_species_order = all_domain_species[largest_domain]
            new_canonical_order = canonical_order[:]
            for None_species, not_None_species in species_exchange.iteritems():
                not_None_ind = new_canonical_order.index(not_None_species)
                None_ind = new_canonical_order.index(None_species)
                
                # Pop the not None from canonical order to update for
                # further species
                new_canonical_order.pop(not_None_ind)
                for domain, domain_species in all_domain_species.items():
                    
                    # Replace the None Species with the not None one
                    if None_species not in domain_species:
                        domain_species[None_ind] = not_None_species
                    domain_species.pop(not_None_ind)
                    
                    # Update the domains
                    all_domain_species[domain] = domain_species

        # Update the order of the scalar_field_species for each domain
        for domain, domain_species in all_domain_species.items():
            debug("%s : %s", str(domain), print_list(domain_species))
            scalar_field_species = []
            for species in domain_species:
                if species in domain.species and \
                   isinstance(domain.species[species], ScalarFieldSpecies):
                    scalar_field_species.append(domain.species[species])
                else:
                    scalar_field_species.append(None)
                domain.scalar_field_species = scalar_field_species
        
    def _finalize(self):
        """
        Finalizes the model. No object can be added after this is called
        """
        self.subcellmodel._finalized = True

        # Collect stuff relating to boundaries
        # ------------------------------------
        boundary_params = {}
        for boundary in self.subcellmodel.boundaries:
            
            # Collect boundary values
            if boundary.boundary_values:
                self.subcellmodel.boundary_values.extend(\
                    boundary.boundary_values.values())
            
            # Collect fluxes
            if boundary.fluxes:
                for fluxes in boundary.fluxes.values():
                    self.subcellmodel.fluxes.extend(fluxes)
                    
            # Collect boundary params
            bound_params = self._remove_hidden_params(dict(**boundary.params))
            
            # Collect flux parameters
            for species, fluxes in boundary.fluxes.items():
                species_param = {}
                for flux in fluxes:
                    if flux.params:
                        species_param[str(flux)] = ParameterDict(**flux.params)
                if species_param:
                    bound_params[str(species)] = ParameterDict(**species_param)

            # Create a ParameterDict out of the collected dicts
            boundary_params[boundary.name] = ParameterDict(**bound_params)

        # Collect stuff relating to domains
        # ---------------------------------
        domain_params = {}
        for domain in self.subcellmodel.domains:

            # Collect domain params
            domain_params[domain.name] = self._remove_hidden_params(\
                dict(**domain.params))

            # Collect species
            for species, domain_specific_species in \
                    domain.species.iteritems():

                # Collect all species
                self.subcellmodel.species[species].append(\
                    domain_specific_species)

                # Collect species connections
                self.subcellmodel.species_connections.extend(\
                    domain_specific_species.connections)

                # Collect species params
                domain_params[domain.name][str(species)] = \
                    ParameterDict(**domain_specific_species.params)

                if isinstance(domain_specific_species, ConstSpecies):
                    self.subcellmodel.const_species.append(\
                        domain_specific_species)
                elif isinstance(domain_specific_species, ScalarSpecies):
                    self.subcellmodel.scalar_species.append(\
                        domain_specific_species)
                elif isinstance(domain_specific_species, FieldSpecies):
                    self.subcellmodel.field_species.append(\
                        domain_specific_species)
                    if isinstance(domain_specific_species, ScalarFieldSpecies):
                        pass # These will be populated later
                    elif isinstance(domain_specific_species, \
                                    FixedScalarFieldSpecies):
                        self.subcellmodel.fixed_scalar_field_species.append(\
                            domain_specific_species)
                        # Check that the FixedSpecies is used in a reaction
                        for reaction in domain.reactions:
                            if domain_specific_species.species in \
                               reaction.reactants.keys() + \
                               reaction.products.keys():
                                break
                        else:
                            raise RuntimeError(\
                                "expected a declared FixedScalarFieldSpecies "\
                                "to be used in a reaction. Use '%s' in a "\
                                "reaction or remove it."%\
                                str(domain_specific_species))
            
            # Convert the parameters to a ParameterDict
            domain_params[domain.name] = ParameterDict(\
                **domain_params[domain.name])
                
            # Collect reactions
            if domain.reactions:
                self.subcellmodel.reactions.extend(domain.reactions)

        # Rearrange order of field species
        self._rearange_field_species_order()
        
        # Assign the scalar_field_species using the iter_scalar_field_species
        # which now relies on the new order
        for field_species in self.subcellmodel.iter_scalar_field_species():
            self.subcellmodel.scalar_field_species.append(field_species)
        
        # Check that the user included at least one flux or a reaction
        if not any(self.subcellmodel.reactions + self.subcellmodel.fluxes):
            raise TypeError, "expected at least one reaction or flux to "\
                  "be defined in the subcell model"
        
        # Collect species params
        species_params = {}
        for species in self.subcellmodel.species:
            species_params[str(species)] = ParameterDict(**species.params)

        # Collect variable params and connections
        variable_params = {}
        for variable in self.subcellmodel.continuous_variables + \
            self.subcellmodel.discrete_variables:
            variable_params[str(variable)] = ParameterDict(**variable.params)
            self.subcellmodel.variable_connections.extend(variable.connections)

        # Collect discrete markov model parameters
        dmm_params = {}
        for dmm in self.subcellmodel.discrete_markov_models:
            dmm_params[str(dmm)] = ParameterDict(**dmm.params)
        
        # Collect connections
        for markov_model in self.subcellmodel.discrete_markov_models:
            self.subcellmodel.discrete_markov_model_connections\
                            [markov_model.name] = markov_model.connections
        
        # Create all boundary values
        self.subcellmodel.all_values_at_boundaries.extend(value for value in \
                            self.subcellmodel.iter_all_values_at_boundaries())

        def boundary_value_cmp(x, y):
            "Used to sort the boundary values" 
            if x.boundary != y.boundary:
                return cmp(str(x.boundary), str(y.boundary))
            if x.domain != y.domain:
                return cmp(str(x.domain), str(y.domain))
            if x.species != y.species:
                return cmp(str(x.species), str(y.species))
            return cmp(x.ind, y.ind)

        # Sort the boundary values
        self.subcellmodel.all_values_at_boundaries.sort(boundary_value_cmp)

        # Create the final parameters for the subcellmodel
        params = ParameterDict(
            variables = ParameterDict(**variable_params),
            species = ParameterDict(**species_params),
            domains = ParameterDict(**domain_params),
            boundaries = ParameterDict(**boundary_params),
            discrete_markov_models = ParameterDict(**dmm_params),
            # FIXME: Assume only one mesh!
            mesh = ParameterDict(**self.subcellmodel.meshes[0].params),
            **self.subcellmodel.params
            )
            
        # FIXME: Move a ParSlave to TypeCheck, and make it a flag instead
        # FIXME: also add a check when getting a value from a ParSlave so it
        # FIXME: can be range checked!
        
        # Create a ParameterDict out of the dictionary of Parameters
        self.subcellmodel.params = params

