__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-23 -- 2011-09-21"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["SubCellModelObject", "create_parameter",\
           "user_defined_parameters", "to_odict"]

# System imports
import types
import sympy

# subcell imports
from subcell.common import *

_property_storage = {}

class SubCellModelObject(object):
    __count = 0
    __initalized = False
    __all_names = []
    _all = []
    def __init__(self, name, params=None, *args, **kwargs):
        # Check if initalized already, if not the class variable will return
        # False. This is nessesary when a SubCellModelObject is initialized in
        # a __new__ method because the __init__ method is called a second
        # time...
        if self.__initalized:
            # If so just return
            return

        if name in self.__all_names:
            raise RuntimeError("Expects uniqe names for all model object: "\
                               "'%s' already excist."%name)
        self.__all_names.append(name)

        # Assign default arguments
        self.params = params or {}
        
        # Check arguments
        check_arg(name, str, 0)
        check_kwarg(self.params, "params", dict)

        # Flag the object as initialized
        self.__initalized = True

        # Store attributes
        self.name = name
        
        # Register passed parameters
        self.params, self.builtin_params, self.property_params = \
                     self._register_parameters(self.params)

        # Keep an uniqe identifier to the object
        # FIXME: Why is this nessesary?
        self._count = SubCellModelObject.__count
        SubCellModelObject.__count += 1
        
        self._hash = hash(name+str(self._count))

        # Store SubCellObject
        #SubCellModelObject._all.append(self)
        
        #_property_storage[self._hash] = 

    def _register_parameters(self, params):
        """
        Register any parameters domain/species/boundary specific
        parameters
        """

        if not params:
            return {}, [], []
        
        assert((isinstance(params, dict)))
        
        builtin_params = []
        property_params = []
        
        # Iterate over the parameters and register them
        for parname, (value, builtin, property_param) in \
                reversed(params.items()):
            
            if builtin:
                builtin_params.append(parname)
            
            if property_param:
                property_params.append(parname)
            
            if not isinstance(value, (int, float, bool, sympy.Basic, \
                                      TypeCheck)):
                raise TypeError("expected a scalar, TypeCheck as the "\
                                "parameter argument")
            elif isinstance(value, scalar):
                value = TypeCheck(value)
            elif isinstance(value, TypeCheck):
                value = value
            elif isinstance(value, sympy.Basic):
                if not all(isinstance(atom, (sympy.Number, ParSymbol))\
                           for atom in value.atoms()):
                    raise TypeError, "expected parameter expression "\
                          "to consist of scalars or other parameters"
                value = ParSlave(value)
            
            # Update parameters
            params[parname] = value
            
            if not property_param:
                self.__dict__[parname] = value

        return params, builtin_params, property_params

    def __getattribute__(self, param):
        try:
            property_params = object.__getattribute__(self, "property_params")
        except:
            property_params = []
        
        if param in property_params:
            return self.params[param].get()
        return object.__getattribute__(self, param)
    
    def __eq__(self, other):
        if not isinstance(other, SubCellModelObject):
            return False
        return self._hash == other._hash

    def __str__(self):
        return self.name

    def _param_repr(self):
        # Create a repr str based on the registered parameters
        # First register the parameters in _param_names
        params = [", %s=%s"%(name, repr(self.params[name])) \
                  for name in self.builtin_params if "__" not in name]
        params.extend([", %s=%s"%(name, repr(value)) \
                       for name, value in self.params.iteritems() \
                       if name not in self.builtin_params])
        
        return "".join(params)
        
    def __repr__(self):
        return "%s('%s'%s)"%(type(self).__name__, self.name, \
                             self._param_repr())

    def __hash__(self):
        return self._hash

def user_defined_parameters(params, only_scalar=True, allow_array=False, \
                            symnameprefix=None):
    #debug("Creating parameters with symname prefix: '%s'", symnameprefix)
    assert(isinstance(params, dict))
    ret = odict()
    for name, value in params.iteritems():
        symname = None if symnameprefix is None else \
                  "%s_%s"%(symnameprefix, name)
        ret[name] = create_parameter(value, name, False, False, \
                                     only_scalar=only_scalar, \
                                     allow_array=allow_array,
                                     symname=symname)
    
    return ret

def create_parameter(value, name, built_in=True, property_param=True,\
                     ge=None, le=None, gt=None, lt=None, only_scalar=True,\
                     allow_array=False, symname=None):
    "Create a RangeCheck or a TypeCheck from either a scalar, TypeCheck "
    # FIXME: Fix a way to check Parameter types for the passed value
    check_arg(value, (scalar)+(TypeCheck, sympy.Basic), 0)
    check_arg(name, str, 1)
    
    # If we have a sympy.Basic we assume it is the result of an expression
    # of TypeChecks and send it to ParSlave
    #debug("Creating param '%s' with symname '%s'", name, symname)
    if isinstance(value, sympy.Basic):
        return (ParSlave(value), built_in, property_param)

    # If a TypeCheck is passed just return this
    if isinstance(value, TypeCheck):
        # FIXME: Clean this logic...
        scalar_fail = only_scalar and not value.is_scalar
        if not allow_array:
            if scalar_fail:
                raise TypeError("expected a scalar parameter for '%s'"%name)
        elif scalar_fail and not value.is_array:
            raise TypeError("expected a scalar parameter or an "\
                            "ArrayParam for '%s'"%name)
        
        # If a ParSlave is passed just return it
        if isinstance(value, ParSlave):
            return (value, built_in, property_param)

        # Re-create parameter symbol with potential symname
        value.sym = ParSymbol(value, symname)

    # If a scalar is passed create a TypeCheck wrapper around it
    else:
        # If no range check variables are given, then just create a TypeCheck
        if all(arg is None for arg in [ge, le, gt, lt]):
            ParType = TypeCheck
            kwargs = dict(name=name, symname=symname)
        else:
            ParType = RangeCheck
            kwargs = dict(name=name, symname=symname, ge=ge, le=le, \
                          gt=gt, lt=lt)
        
        # Create parameter
        value = ParType(value, **kwargs)
        
    # Assign the name of the parameter
    if value.name is None:
        if symname is None:
            value.name = name
        else:
            value.name = name, symname
    
    return (value, built_in, property_param)
    
def to_odict(params=None):
    if params is None:
        return odict()
    assert(isinstance(params, dict))
    if isinstance(params, odict):
        return params
    ret = odict()
    for key, value in params.iteritems():
        ret[key] = value

    return ret
