__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-09-16 -- 2011-10-21"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["DiscreteVariable", "Variable", "ContinuousVariable",\
           "create_parameter"]

# System imports
import sympy

# subcell imports
from subcell.common import *
from subcellmodelobject import *    

def exchange_symbols_with_Variables(expr):
    """
    Substitution dict to remove all generated sympy.Symbols which
    does not play with DomainSpecificSpecies
    """
    d = {}
    for atom in expr.atoms():
        if str(atom) in Variable._all_objects:
            d[atom] = Variable._all_objects[str(atom)]
    return expr.subs(d)

class Variable(SubCellModelObject, sympy.Symbol):
    """
    Base class for variables that are not solved for, but rather prescribed
    by some defined rule"""
    
    _all_objects = {}
    
    # A Dummy function which return the present time
    time = lambda : 0.0
    
    def __new__(cls, name, params=None):
        check_arg(name, str, 0)
        self = sympy.Symbol.__new__(cls, name)
        Variable._all_objects[str(self)] = self
        
        # FIXME: When is this populated?
        self.receivers = []
        self.connections = []
        
        # Initiate value to zero
        self.value = 0.0
        self._func = None
        
        # Initiate the base class passing any specific parameters along
        SubCellModelObject.__init__(self, name, params)
        return self

    def _connect(self, receiver):
        "Store connection information between the variable and a reciever"
        from connection import Sender, Receiver, Connection
        check_arg(receiver, Receiver)

        con = Connection(Sender(self), receiver)

        if con in self.connections:
            raise ValueError, "The connection already excists"
        
        # Store the connection
        self.connections.append(con)
        
    def send(self):
        "Send its value to all its receivers"
        for receiver in self.receivers:
            receiver(self.value)

    @property
    def time(self):
        "Return the present time as seen from the registered function"
        return Variable.time()

    def reset(self):
        "Rest variable, called once for each simulation"
        # Implement in subclass
        pass

    def _receive(self, value):
        # Implement in subclass
        pass

class DiscreteVariable(Variable):
    "Base class for discrete variables"
    def __new__(cls, name, end_sim=False, minimize_dt=False, params=None):
        check_arg(name, str, 0)
        params = to_odict(params)
        params["minimize_dt"] = create_parameter(minimize_dt, "minimize_dt")
        params["end_sim"] = create_parameter(end_sim, "end_sim")
        return Variable.__new__(cls, name, params)
    
    @property
    def next_time_step(self):
        return float('inf')
    
    def update(self, dt):
        pass

    def step(self):
        pass

class ContinuousVariable(Variable):
    "Base class for continuous variables"
    def __new__(cls, name, params=None):
        return Variable.__new__(cls, name, params)

    def step_back(self, offset):
        assert(isinstance(offset, float))
        assert(offset>0)
        
    def update(self, dt, time):
        assert(isinstance(dt, float))
        assert(isinstance(time, float))
        assert(dt>0)
