__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2011-03-17 -- 2011-10-31"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["SubCellData"]

import os

from subcell.common import check_arg
from subcell.common.disk import load
from subcell.common.logger import info

class SubCellData(object):
    """
    A class to retreive data stored from a subcell simulations
    """
    
    def __init__(self, casename, latest_timestamp=False, collect=False):
        """
        Create a SubCellData instance, holding data from subcell simulation
        
        @type casename : str
        @param casename : The name of the cPickle file from which the data are 
        read from
        @type latest_timestamp : bool
        @param latest_timestamp : If true return the data from latest version
        of saved data with the same basename
        @type collect : bool
        @param collect : If True collect all data with the same basename and
        the same parameters
        
        """
        
        check_arg(casename, str, 0)
        self.data = load(casename, latest_timestamp, collect)
        self.new_data_structure = isinstance(self.data["diffreact"], list)

        # Strip casename from .cpickle and store it for use in field
        # data extractor
        if ".cpickle" in casename:
            casename = casename.replace(".cpickle", "")
        
        self.casename = casename
        
    def params(self):
        """
        Return the parameters for a registered run
        """
        return self.data["params"]
        
    def current(self, boundary, species, fluxname, num=0, run=0):
        """
        Return the recorded current from a boundary with a declared flux.
        
        The species asked for need to have a nonzero valence, otherwise
        the current will equal 0.
        
        @type boundary : str
        @param boundary : The boundary which the flux are declared on
        @type species : str
        @param species : The name of the species which the fluxed described.
        @type fluxname : str
        @param fluxname : The name of the flux
        @type num : scalar
        @param num : The boundary number from which the current is
        returned from.
        @type run : scalar
        @param run : The run number from which the current is returned from.
        @rtype: numpy array
        @return: The current given in pA
       
        """
        
        value = self.scalar_flux(boundary, species, fluxname, num, run)
        valence = self.data["params"]["subcellmodel"]["species"][species]\
                  ["valence"]
        
        return value*1e12*valence*0.096*1e3

    def scalar_flux(self, boundary, species, fluxname, num=0, run=0):
        """
        Return the recorded scalar_flux (umoles/ms) from a boundary with a
        declared flux.
        
        @type boundary : str
        @param boundary : The boundary which the flux are declared on
        @type species : str
        @param species : The name of the species which the fluxed described.
        @type fluxname : str
        @param fluxname : The name of the flux
        @type num : scalar
        @param num : The boundary number from which the flux number is returned
        from.
        @type run : scalar
        @param run : The run number from which the flux number is
        returned from.
        @rtype: numpy array
        @return: The flux number given in number per ms
        """
        check_arg(boundary, str, 0)
        check_arg(species, str, 1)
        check_arg(fluxname, str, 2)
        check_arg(num, int, 3)
        check_arg(run, int, 4)

        num_runs = self.num_runs()
        
        if run >= num_runs:
            raise IndexError, "The run num '%d' is too large. Only '%d' "\
                  "runs are registered."%(run, num_runs)

        if not "scalar_fluxes" in self.data["diffreact"][run]:
            raise KeyError, "No scalar fluxes collected"

        data = self.data["diffreact"][run]["scalar_fluxes"]
        
        if not boundary in data:
            raise KeyError, "The boundary '%s' is not a declared boundary in"\
                  " the data or there are no flux declared on the "\
                  "boundary."%boundary

        if not species in data[boundary]:
            raise KeyError, "The species '%s' is not declared as a flux on "\
                  "boundary '%s'"%(species, boundary)

        if not fluxname in data[boundary][species]:
            raise KeyError, "The flux '%s' is not declared as a flux on "\
                  "boundary '%s'"%(fluxname, boundary)
            
        num_bound = len(data[boundary][species][fluxname])
        
        if num >= num_bound:
            raise IndexError, "The boundary num '%d' is too large. Only '%d' "\
                  "boundaries are registered."%(num, num_bound)

        return data[boundary][species][fluxname][num]
    
    def flux_numbers(self, boundary, species, fluxname, num=0, run=0):
        """
        Return the recorded flux number from a boundary with a declared flux.
        
        @type boundary : str
        @param boundary : The boundary which the flux are declared on
        @type species : str
        @param species : The name of the species which the fluxed described.
        @type fluxname : str
        @param fluxname : The name of the flux
        @type num : scalar
        @param num : The boundary number from which the flux number is returned
        from.
        @type run : scalar
        @param run : The run number from which the flux number is
        returned from.
        @rtype: numpy array
        @return: The flux number given in number per ms
        
        """
        
        value = self.scalar_flux(boundary, species, fluxname, num, run)

        return value*6e23*1e-6
    
    def time(self, run=0):
        """
        Returns the time for a certain run
    
        @type run : scalar
        @param run : The run number from which the time is returned from.
        @rtype: numpy array
        @return: The time of the n:th run
    
        """
        
        check_arg(run, int, 0)

        num_runs = self.num_runs()
        
        if run >= num_runs:
            raise IndexError, "The run num '%d' is too large. Only '%d' "\
                  "runs are registered."%(run, num_runs)
        
        if self.new_data_structure:
            return self.data["diffreact"][run]["times"]["times"]

        return self.data["diffreact"]["times"]["times"][run]
        
    def dt(self, run=0):
        """
        Returns the size of the time steps for a certain run
    
        @type run : scalar
        @param run : The run number from which the time is returned from.
        @rtype: numpy array
        @return: The size of the steps of the n:th run
    
        """
        
        check_arg(run, int, 0)

        num_runs = self.num_runs()
        
        if run >= num_runs:
            raise IndexError, "The run num '%d' is too large. Only '%d' "\
                  "runs are registered."%(run, num_runs)
        
        if self.new_data_structure:
            return self.data["diffreact"][run]["times"]["dt"]

        return self.data["diffreact"]["times"]["dt"][run]
        
    def concentration(self, domain, species, run=0):
        """
        Returns the average concentration of a given species from a domain
    
        @type domain : str
        @param domain : The domain from where the concentration is
        collected from
        @type species : str
        @param species : The name of the species
        @rtype: numpy array
        @return: The concentration
        """
    
        check_arg(domain, str, 0)
        check_arg(species, str, 1)
        check_arg(run, int, 2)

        num_runs = self.num_runs()
        
        if run >= num_runs:
            raise IndexError, "The run num '%d' is too large. Only '%d' "\
                  "runs are registered."%(run, num_runs)
        
        data = self.data["diffreact"][run] if self.new_data_structure else \
               self.data["diffreact"]

        if domain not in data["domains"]:
            raise KeyError, "The domain '%s' is not a declared domain in the "\
                  "model"%domain
    
        if species not in data["domains"][domain]:
            raise KeyError, "The species '%s' is not a registered species"\
                  " in the '%s' domain"%(species, domain)

        if self.new_data_structure:
            return data["domains"][domain][species]
    
        return self.data["diffreact"]["domains"][domain][species][run]
    
    def number(self, domain, species, run=0):
        """
        Returns the average number of a given species from a domain
    
        @type domain : str
        @param domain : The domain from where the number is collected from
        @type species : str
        @param species : The name of the species
        @rtype: numpy array
        @return: The number of the species
        """
    
        check_arg(domain, str, 0)
        check_arg(species, str, 1)
        check_arg(run, int, 2)

        conc = self.concentration(domain, species, run)
        data = self.data["geometry"] if self.new_data_structure else \
               self.data["diffreact"]["geometry"] 
        
        volume = data["domains"][domain]
        
        return conc*volume/1e24*6e23*1e-6
    
    def boundary_value(self, domain, boundary, species, num=0, run=0):
        """
        Depricated function
        """
        print "Depricated: Use 'conc_at_boundary' instead"
        return self.conc_at_boundary(domain, boundary, species, num, run)
        
    def conc_at_boundary(self, boundary, species, num=0, run=0):
        """
        Returns the average concentration at a given mesh boundary and domain
    
        @type boundary : str
        @param boundary : The boundary which the concentration are returned
        from.
        @type species : str
        @param species : The name of the species
        @rtype: numpy array
        @return: The number of the species
        """
        
        check_arg(boundary, str, 0)
        check_arg(species, str, 1)
        check_arg(num, int, 2)
        check_arg(run, int, 3)
        
        num_runs = self.num_runs()
        
        if run >= num_runs:
            raise IndexError, "The run num '%d' is too large. Only '%d' "\
                  "runs are registered."%(run, num_runs)
        
        if not self.new_data_structure:
            raise NotImplementedError("not accessible for old data structure")
        data = self.data["diffreact"][run]

        # Keeping backward compatability
        for variable_name in ["conc_at_boundary", "value_at_boundaries", \
                              "boundary_values"]:
            if variable_name in data:
                break
        
        conc_boundaries = self.conc_boundaries()
        if not boundary in conc_boundaries:
            raise KeyError, "The boundary '%s' is not a registered boundary."\
                  %(boundary)

        ret = {}
        for domain, values in conc_boundaries[boundary].iteritems():
            if not species in values:
                raise KeyError, "The species '%s' is not a registered "\
                      "species in a domain adjacent to '%s'"%(species, \
                                                              boundary)
            local_data = data[variable_name][domain][species][boundary]
            if num >= len(local_data):
                raise IndexError, "The boundary num '%d' is too large. "\
                      "Only '%d' boundaries are registered."%(num, \
                                                              len(local_data))
                
            ret[domain] = local_data[num]

        if len(ret) == 1:
            return ret[domain]

        return ret

    def markov_model_values(self, markov_model, num=0, run=0):
        """
        Returns registered Markov model values
        """
        
        check_arg(markov_model, str, 0)
        check_arg(num, int, 1)
        check_arg(run, int, 2)
        
        if self.new_data_structure:
            data = self.data["stochastic"][run][markov_model]
            return data["value_times"][num], data["values"][num]
        
        data = self.data["stochastic"][markov_model]
        return data["value_times"][num][run], data["values"][num][run]
    
    
    def variable(self, variable, run=0):
        """
        Return the values of a continuous variable
        """
        
        check_arg(variable, str, 0)
        check_arg(run, int, 1)
        
        if not self.new_data_structure:
            raise NotImplementedError("continuous_variable is not accessible "\
                                      "for old data structure")
        num_runs = self.num_runs()
        
        if run >= num_runs:
            raise IndexError, "The run num '%d' is too large. Only '%d' "\
                  "runs are registered."%(run, num_runs)
        
        data = self.data["diffreact"][run]
        
        for vartype in ["continuous_variables", "discrete_variables"]:
            if variable in data[vartype]:
                return data[vartype][variable]
                
        raise KeyError("The variable '%s' is not registered.")

    def variables(self):
        """
        Return the names of all variables registered
        """
        if not self.new_data_structure:
            raise NotImplementedError("variables are not accessible "\
                                      "for old data structure")
        
        data = self.data["diffreact"][0]
        ret = {}
        for vartype in ["continuous_variables", "discrete_variables"]:
            if vartype in data and len(data[vartype]) > 0:
                ret[vartype] = [variable for variable in data[vartype]]
                
        return ret
    
    def markov_model_states(self, markov_model, num=0, run=0):
        """
        Returns registered Markov model states
        """
        
        check_arg(markov_model, str, 0)
        check_arg(num, int, 1)
        check_arg(run, int, 2)

        if self.new_data_structure:
            data = self.data["stochastic"][run][markov_model]
            return data["state_times"][num], data["states"][num]
        
        data = self.data["stochastic"][markov_model]
        return data["state_times"][num][run], data["states"][num][run]
    
    def volume(self, domain):
        """
        Returns the registered volume of a domain
        """
        
        check_arg(domain, str, 0)

        data = self.data["geometry"] if self.new_data_structure else \
               self.data["diffreact"]["geometry"] 
        if domain in data["volume domains"]:
            return data["volume domains"][domain]
        if domain in data["mesh domains"]:
            return data["mesh domains"][domain]
        
        raise RuntimeError("No domain with name '%s' is registered"%domain)
        
    def area(self, boundary):
        """
        Return the registered area of a boundary
        """
        
        check_arg(boundary, str, 0)
        
        data = self.data["geometry"] if self.new_data_structure else \
               self.data["diffreact"]["geometry"]

        if boundary in data["boundaries"]:
            return data["boundaries"][boundary]["area"]

        if boundary in data["mesh boundaries"]:
            return data["mesh boundaries"][boundary]["area"]

        raise RuntimeError("No boundary with name '%s' is registered"%boundary)

    def dx(self, boundary):
        """
        Return the registered dx of a boundary
        """
        
        check_arg(boundary, str, 0)
        
        data = self.data["geometry"] if self.new_data_structure else \
               self.data["diffreact"]["geometry"]

        if boundary in data["boundaries"]:
            return data["boundaries"][boundary]["dx"]

        if boundary in data["mesh boundaries"]:
            return data["mesh boundaries"][boundary]["dx"]

        raise RuntimeError("No boundary with name '%s' registered"%boundary)
    def mesh_area(self, boundary, domain="", num=0):
        """
        Return the registered area of a mesh boundary
        """
        check_arg(boundary, str, 0)
        check_arg(domain, str, 1)
        check_arg(num, int, 2)
        
        data = self.data["geometry"]["boundaries"][boundary][domain] \
               if self.new_data_structure else \
               self.data["diffreact"]["geometry"]["boundaries"][boundary]

        return data[num]

    def markov_models(self):
        """
        Return names of all Markov models which data has been collected for.
        
        """
        if "stochastic" not in self.data:
            return []
        
        if self.new_data_structure:
            return self.data["stochastic"][0].keys()

        return self.data["stochastic"].keys()
    
    def domains(self):
        """
        Return names of all domains together with available species which
        data has been collected for.
        
        """
        if self.new_data_structure:
            if "domains" not in self.data["diffreact"][0]:
                return {}
            return dict((domain,\
                         self.data["diffreact"][0]["domains"][domain].keys())\
                        for domain in self.data["diffreact"][0]\
                        ["domains"].keys())
        
        if "domains" not in self.data["diffreact"]:
            return {}
        return dict((domain,\
                     self.data["diffreact"]["domains"][domain].keys())\
                    for domain in self.data["diffreact"]["domains"].keys())
    
    def num_boundaries(self, boundary):
        """
        Return the number of boundaries from a particular boundary
        """
        
        check_arg(boundary, str, 0)

        if boundary not in self.params()["subcellmodel"]["boundaries"]:
            return 0
        return self.params()["subcellmodel"]["boundaries"][boundary]["num"]
        
    def boundaries(self):
        """
        Return all registered Boundaries with their domains
        """
        boundaries = {}

        data = self.data["geometry"] if self.new_data_structure else \
               self.data["diffreact"]["geometry"]

        if "boundaries" in data:
            for boundary in data["boundaries"]:
                boundaries[boundary] = data["boundaries"][boundary]["domains"]

        if "mesh boundaries" in data:
            for boundary in data["mesh boundaries"]:
                boundaries[boundary] = \
                                data["mesh boundaries"][boundary]["domains"]

        return boundaries
    
    def scalar_flux_boundaries(self):
        """
        Return names on all boundaries which available scalar flux data has
        been collected for.
        """
        if "scalar_fluxes" not in self.data["diffreact"][0]:
            return {}
        data = self.data["diffreact"][0]["scalar_fluxes"]
        return dict((boundary,
                     dict((species, data[boundary][species].keys())
                          for species in data[boundary]))
                    for boundary in data)

    def conc_boundaries(self):
        """
        Return names on all boundaries which available conc data has
        been collected for.
        """
        if self.new_data_structure:
            if "conc_at_boundary" not in self.data["diffreact"][0]:
                return {}
            data = self.data["diffreact"][0]["conc_at_boundary"]
            domains = data.keys()
            boundaries = set()
            for domain in domains:
                for species in data[domain]:
                    for boundary in data[domain][species]:
                        boundaries.add(boundary)
            boundaries = dict((boundary, {}) for boundary in boundaries)
            for domain in domains:
                for species in data[domain]:
                    for boundary in data[domain][species]:
                        if domain not in boundaries[boundary]:
                            boundaries[boundary][domain] = []
                        boundaries[boundary][domain].append(species)
            
            return boundaries
        
        raise NotImplementedError("Not implemented for old data structure")

    def num_runs(self):
        """
        Return the number of runs
        """
        if self.new_data_structure:
            return len(self.data["diffreact"])
        
        return len(self.data["diffreact"]["times"]["times"])

    def tstops(self):
        """
        Return a list of all tstop
        """
        if self.new_data_structure:
            return [data_run["times"]["times"][-1] for data_run in \
                    self.data["diffreact"]]
        return [times[-1] for times in \
                self.data["diffreact"]["times"]["times"]]

    def field_concentration(self, domain, species, run=0, mesh=None):
        """
        Returns an iterator over the field concentration of a given species
        from a domain
    
        @type domain : str
        @param domain : The domain from where the concentration is
        collected from
        @type species : str
        @param species : The name of the species
        @rtype: iterator
        @return: Tuple of the time step and the concentration
        """
        from dolfin import Mesh, Function, FunctionSpace, TimeSeries

        filename = "%s/%d/%s%s"%(self.casename, run,\
                                 domain, species)
        if not os.path.isfile(filename+"_mesh_0.bin.gz"):
            raise IOError("No field data for run %d of %s in %s for %s."%(\
                run, species, domain, self.casename))

        # Load TimeSeries
        info("Loading field data from: %s", filename)
        ts = TimeSeries(filename, True)

        # Check data
        if len(ts.vector_times())==0:
            raise RuntimeError("No field data stored in '%s'"%filename)
        
        # Load mesh if not given
        if not mesh:
            mesh = Mesh()
            ts.retrieve(mesh, 0.0)
        
        # Create Function
        V = FunctionSpace(mesh, "CG", 1)
        u = Function(V)

        # Iterator over vector times
        times = ts.vector_times()
        for t in times:
            ts.retrieve(u.vector(), t)
            yield t, u
        
