__author__ = "Johan Hake (hake.dev@gmail.com)"
__copyright__ = "Copyright (C) 2010 " + __author__
__date__ = "2010-10-05 -- 2011-11-07"
__license__  = "GNU LGPL Version 3.0 or later"

__all__ = ["TimeStepper"]

# System imports
from operator import attrgetter
import os
import numpy as np
import time
import copy

# SubCell imports
from subcell.common import *
from subcell.common.logger import *
from subcell.common.disk import *
from subcell.backends import *
from subcell.model import SubCellModel, load_model
from subcell.model.discretemarkovmodel import DiscreteMarkovModel
from subcell.model.variables import *

# FIXME: Convert to runner parameters?
_params = dict(
    # General time parameters
    time=ParameterDict(
        tstart=RangeCheck(0.0, gt=-1e6, lt=1e5),
        tstop=RangeCheck(10.0, gt=0, lt=1e5),
        dt_max_values=TypeCheck([]),
        dt_max_times=TypeCheck([]),
        dt_max=RangeCheck(1.0, gt=0, lt=100),
        dt_start=RangeCheck(.1, gt=0, lt=100),
        dt_min=RangeCheck(.1, gt=0, lt=100),
        num_runs=RangeCheck(1, ge=0),
        dt_expand=RangeCheck(2., ge=1),
        transient=TypeCheck(True),
        initial_steady_state_solve=False,
        ),
    
    # General io parameters
    io=ParameterDict(\
        casename=TypeCheck("default"),
        datadir=TypeCheck("data"),
        logdir=TypeCheck("log"),
        use_time_stamp=TypeCheck(False),
        use_logfile=False,
        supress_output=False,
        ),
    )

class DtExpander(DiscreteVariable):#
    "A discrete variable taking care of an expanding time step"
    def __new__(cls, timestepper):
        assert(isinstance(timestepper, TimeStepper))
        
        # Initialize base class
        self = DiscreteVariable.__new__(cls, "dt_expander")

        # Save the time params from the timestepper
        self.time_params = timestepper.params.time
        self.value = self.time_params.dt_start
        self.scale = self.time_params.dt_expand
        
        return self
    
    @property
    def next_time_step(self):
        return self.value

    def reset(self):
        self.value = self.time_params.dt_start
        self.scale = self.time_params.dt_expand

    def minimize(self):
        self.value = self.time_params.dt_min
        
    def step(self):
        self.value *= self.scale

class DtMax(DiscreteVariable):
    "A discrete variable taking care of the maximal timestep"
    def __new__(cls, timestepper):
        assert(isinstance(timestepper, TimeStepper))
        
        # Initialize base class
        self = DiscreteVariable.__new__(cls, "dt_max")

        # Save the time params from the timestepper
        self.time_params = timestepper.params.time
        self.value = 0.0
        self._ind = 0

        return self
    
    @property
    def next_time_step(self):
        return min(self.value, self._times[self._ind+1] - self.time())

    def reset(self):
        # Check if we have a vector for the dt_max
        if self.time_params.dt_max_values:
            dt_max_values = self.time_params.dt_max_values[:]
            dt_max_times = self.time_params.dt_max_times[:]
            error_str = "Expected 'dt_max_times' and 'dt_max_values' to "\
                        "be two list of equally size. The first a monotonic "\
                        "assending list with time values. The second with "\
                        "positive maximum values for dt, corresponding to "\
                        "the values in the first list. Got '%s'/'%s'"
            if len(dt_max_times)!=len(dt_max_values):
                raise ValueError(error_str%(str(dt_max_times), \
                                            str(dt_max_values)))
            if not np.all(np.diff(dt_max_times)>0):
                raise ValueError(error_str%(str(dt_max_times), \
                                            str(dt_max_values)))
            if not np.all(dt_max_values>0):
                raise ValueError(error_str%(str(dt_max_times), \
                                            str(dt_max_values)))
            if dt_max_times[0] != self.time_params.tstart:
                raise ValueError("The first time value need to be equal "\
                                 "to 'tstart': %f!=%f"%\
                                 (dt_max_times[0], self.time_params.tstart))
            if dt_max_times[-1] >= self.time_params.tstop:
                raise ValueError("The last value of the time values need to"\
                                 " be smaller than 'tstop'")
            
            # Save lists
            self._times = dt_max_times
            self._dt_max_vec = dt_max_values
            self._times.append(self.time_params.tstop)
            self._dt_max_vec.append(self._dt_max_vec[-1])
        else:
            
            # Create lists from time params
            self._times = [self.time_params.tstart, self.time_params.tstop]
            self._dt_max_vec = [self.time_params.dt_max]*2
        
        self._ind = 0
        self.value = self._dt_max_vec[0]

    def step(self):
        if self._times[self._ind+1] <= self.time() and \
               not self.time() == self.time_params.tstop:
            self._ind += 1
            self.value = self._dt_max_vec[self._ind]

class TimeStepper(object):
    """
    A class that takes care of the time and the stepping of a SubCell model
    """
    def __init__(self, subcellmodel, diffusion_reaction_backend="dolfin",
                 stochastic_backend="gillstep", **kwargs):
        check_arg(subcellmodel, str, 0)
        check_arg(diffusion_reaction_backend, str, 1)
        check_arg(stochastic_backend, str, 2)
        
        # Get the backends
        DiffusionReactionBackend = load_diffusion_reaction_backend_class(\
            diffusion_reaction_backend)
        try:
            StochasticBackend = load_stochastic_backend_class(\
                stochastic_backend)
        except Exception, e:
            warning(str(e))
            warning("*** Warning: No stochastic backend installed, disabling "\
                    "the use of DiscreteMarkovModels")
            DiscreteMarkovModel.disable()
            StochasticBackend = lambda x : None
            
            # When we do not have a stochastic backend to keep track of time we
            # need to establish one here
            self._time = 0.0
        
        # Load subcell model
        self.subcellmodel = load_model(subcellmodel, **kwargs)
        
        # Load the backends and wire up the connections
        self.stoch_backend = StochasticBackend(self.subcellmodel)
        self.diffreact_backend = DiffusionReactionBackend(self.subcellmodel)
        
        # Set the time for the timestepper
        # (depends on whether we have a stochastic backend or not)
        self.step_dt, self.time = self._time_handlers()
        
        # Set the present_time for the Variables
        Variable.time = self.time
        
        # Store variables
        self.continuous_variables = self.subcellmodel.continuous_variables
        self.discrete_variables = self.subcellmodel.discrete_variables

        # Extract params
        self.params = self._extract_params(kwargs)
        """@ivar: Parameters"""
        
        # Create the time dt_max and tstop discrete variables
        self._dt_max = DtMax(self)
        self._tstop = self.subcellmodel.variable("tstop")
        self._tstop.set_timeparams(self.params.time)
        self.discrete_variables.append(self._dt_max)
        self._dt_expander = DtExpander(self)
        self.discrete_variables.append(self._dt_expander)
        
        # Collect end_sim senders
        # NOTE: (highjacking the connection system here...)
        for var in self.discrete_variables:
            if var.end_sim:
                var.receivers.append(self.end_sim)
            if var.minimize_dt:
                var.receivers.append(self.minimize_dt)
        
        # Set callbacks for minimize dt and end sim for stochastic backend
        if self.stoch_backend is not None:
            self.stoch_backend.set_end_sim_callback(self.end_sim)
            self.stoch_backend.set_minimize_dt_callback(self.minimize_dt)
        
        # Init the next_discrete_object
        self._next_discrete_object = []

        self.reset_runtime_params()
        
    def reset_runtime_params(self):
        # Runtime parameters
        self.runtime_params = ParameterDict(
            logfile=TypeCheck(""),
            run_num=RangeCheck(0, ge=0),
            datafile=TypeCheck(os.path.join(self.params.io.datadir,
                                             self.params.io.casename)),
            step=RangeCheck(0, ge=0),
            present_time_str="",
            time_at_start=0.0,
            )
        
    def _extract_params(self, model_arguments):
        "Extract parameters stored in the Backends"
        
        # Create a shallow copy of the module parameters
        params = dict(**_params)

        # Collect parameter from objects
        if hasattr(self.stoch_backend, "params"):
            params["stochastic"] = self.stoch_backend.params

        if hasattr(self.diffreact_backend, "params"):
            params["diffreact"] = self.diffreact_backend.params

        params["subcellmodel"] = self.subcellmodel.params
        params["model_arguments"] = ParameterDict(\
            **dict((key, ParConstant(value)) \
            for key, value in model_arguments.items()))
        
        return ParameterDict(**params)

    def _time_handlers(self):
        "A help method to assign a step_dt function"
        if self.stoch_backend is None:
            
            # Create a step dt function
            def step_dt(dt):
                assert(isinstance(dt, scalar))
                assert(dt>0)
                self._time += dt
                
                # Return zero offset
                return 0.0

            def time(*arg):
                return self._time
            
            return step_dt, time
        else:
            return self.stoch_backend.step_dt, \
                   self.stoch_backend.time

    @property
    def all_variables(self):
        """
        Return all variables
        """
        return self.discrete_variables + self.continuous_variables
    
    def next_time_step(self):
        """
        Return the next time step based on the registered discrete variables
        """
        # Get the discrete object with the smallest next_time_step
        self.discrete_variables.sort(key=attrgetter("next_time_step"))
        next_time_step = self.discrete_variables[0].next_time_step
        for var in self.discrete_variables:
            
            # If any subsequent timestep is smaller than the smallest timestep
            # as set by dt_min we include that discrete variable
            if abs(var.next_time_step - next_time_step) < \
                   self.params.time.dt_min:
                debug("Adding %s to next_discrete_objects", var)
                self._next_discrete_object.append(var)
        
        # Return the value of the smallest next_time_step value
        return next_time_step
    
    def end_sim(self, value):
        """
        Call this method to end the simulation
        """
        self._end_sim = bool(value)

    def minimize_dt(self, value):
        """
        Called when the time step is minimized.
        """
        if self.time() != self.params.time.tstart:
            info("Minmize time step")
            self._dt_expander.minimize()
    
    def reset_before_set_of_runs(self):
        """
        Reset data structures for a set of runs
        """
        
        # Get time when run is called
        self.runtime_params.present_time_str = present_time_str()
        self.runtime_params.time_at_start = time.time()
        
        # Initialize runtime parameters
        datafile = os.path.join(self.params.io.datadir,
                                self.params.io.casename)

        # Create data dir if not excisting
        if not os.path.isdir(self.params.io.datadir):
            os.makedirs(self.params.io.datadir)

        if self.params.io.use_time_stamp:
            datafile = datafile+"-"+self.runtime_params.present_time_str
        
        self.runtime_params.datafile = datafile
        
        info("Storing data in '%s'", datafile)
        
        # Check for log file
        if self.params.io.use_logfile:
            logdir = self.params.io.logdir
            if not os.path.isdir(logdir):
                os.makedirs(logdir)
            logfile = os.path.join(logdir, self.params.io.casename)
            if self.params.io.use_time_stamp:
                logfile = logfile+"-"+self.runtime_params.present_time_str
            
            logfile += ".log"
            
            add_logfile(logfile, "w")
            self.runtime_params.logfile = logfile
            info("Using log file'%s'", logfile)
        
        # Dictionary for collected values
        self.data = dict(diffreact=[], stochastic=[], \
                         params=self.params.copy(True))

        # Reset backends before set of runs
        self.diffreact_backend.reset_before_set_of_runs()
        if self.stoch_backend is not None:
            self.stoch_backend.reset_before_set_of_runs()
        
    def run(self):
        """
        Run simulations. This is the only method a user need to care about.
        """

        # Reset data structures before a set of run
        self.reset_before_set_of_runs()
        
        # If no runs just call reset system
        if self.params.time.num_runs == 0:
            self.reset_before_single_run()
        else:
            # Iterate over runs and execute them
            for num in xrange(self.params.time.num_runs):

                # Update runtime parameters
                self.runtime_params.run_num = num
                
                # Print 
                t = self.formated_real_time()
                info(wrap_log_message("Run number:%3d, real time: %s"%\
                                      (num+1, t)))
                flush_logger()
                
                # Do one run
                self._one_run()
                
        t = self.formated_real_time()
        info(wrap_log_message("Simulation finished. Real time: %s"%t))
        flush_logger()
        self.finalize()
        
        # Reset runtime parameters
        self.reset_runtime_params()

    def _one_run(self):
        """
        Make one run
        """
        self.reset_before_single_run()
        self.runtime_params.step = 0
        while not self._end_sim:
            self.runtime_params.step += 1
            self.step_system()
        
        self.collect_data_after_single_run()
        
    def reset_before_single_run(self):
        """
        Reset the whole system and start time from 'time'
        """
        
        tstart = self.params.time.tstart
        
        self._end_sim = False

        # Reset the time
        # We need to reset the time before the stochastic_backend is reseted
        # as some variables might rely on the present time, which is governed
        # by the stochastic_backend, _and_ the stochastic backend might be
        # dependent on the value of some variables...
        if self.stoch_backend is None:
            self._time = tstart
        else:
            # Set time of stochastic backend
            self.stoch_backend.reset_time(tstart)

        # Set runtime parameters for the diffusion reaction solver
        p = self.diffreact_backend.runtime_params
        p.update(dict(time=tstart, tstart=tstart,
                      dt_expand=self.params.time.dt_expand,
                      initial_steady_state_solve=\
                          self.params.time.initial_steady_state_solve,
                      run_num=self.runtime_params.run_num,
                      
                      # The data dir for the diffreaction solver is the same as
                      # the data file for the timestepper
                      datadir=self.runtime_params.datafile,
                      num_runs=self.params.time.num_runs,
                      ))
        
        # Reset reaction diffusion backend and send its values
        self.diffreact_backend.reset_before_single_run()
        
        # Reset all variables
        for var in self.all_variables:
            var.reset()
        
        # Send all values of the continuous variables to its receivers
        for var in self.continuous_variables:
            var.send()

        # Update the values the diffreaction backend experience
        self.diffreact_backend.update_variables(self.all_variables)
        
        # Send values of discrete variables that has an intial time
        if self.next_time_step()<=0.:
            while self._next_discrete_object:
                var = self._next_discrete_object.pop()
                var.step()
                var.send()
        
        # Reset stochastic backend
        if self.stoch_backend is not None:
            self.stoch_backend.reset_before_single_run(tstart)

    def collect_data_after_single_run(self):
        """
        Collects data from the diffreact and stochastic backends
        """
        
        # Get backend data
        diffreact_data = self.diffreact_backend.\
                         get_data_after_single_run()
        stochastic_data = self.stoch_backend.get_data_after_single_run()\
                          if self.stoch_backend is not None else None
        
        # Append data
        self.data["diffreact"].append(diffreact_data)
        if stochastic_data is not None:
            self.data["stochastic"].append(stochastic_data)
        
        # Pop any initialized data
        elif "stochastic" in self.data:
            self.data.pop("stochastic")

        if "geometry" not in self.data:
            self.data["geometry"] = self.diffreact_backend.get_geometry()
            
        # Save data for each run.
        # NOTE: Additional data will just overwrite previous data. In this way
        # NOTE: we can break a running simulation and still have some data!
        info("Save data to '%s'", self.runtime_params.datafile)
        save(self.runtime_params.datafile, **self.data)
        
    def finalize(self):
        """
        Finalize system. This method is called when the simulation is finished
        """
        
        # Remove any potential logfile handler from logger
        if self.params.io.use_logfile:
            remove_logfile(self.runtime_params.logfile)
        
        # Go into interactive mode if any plotted solutions
        self.diffreact_backend.interactive()
        
    def step_system(self):
        """
        Step the system one discrete time step
        """

        time = self.time()

        # Get the next time step by evaluating the DiscreteVariables
        dt = self.next_time_step()

        # Evaluate and send the ContinuousVariables at the next time step
        self.update_continuous_variables(dt, time)

        # Solve the reaction diffusion reaction equations using the
        # next time step
        self.diffreact_backend.runtime_params.time = time
        self.diffreact_backend.runtime_params.dt = dt
        debug("Solving diffreaction solver with dt = %.3f at t = %.2f ms",\
              dt, time)
        self.diffreact_backend.step()
        
        # Step the stochastic backend
        offset = self.step_dt(dt)
        
        # If we got an offset from the stochastic backend
        if offset>0:

            debug("DiscreteMarkovModel offset: %.3f", offset)

            # Backstep the continuous variables
            self.step_back_continuous_variables(offset/dt)
            
            # Backstep the reaction diffusion backend
            self.diffreact_backend.step_back(offset/dt)
            
            # Pop the collected next timestep variables
            while self._next_discrete_object:
                self._next_discrete_object.pop()
                
        else:
            # Step and send the discrete variables that was responsible for 
            # the size of the dt
            assert(self._next_discrete_object)
            info("")
            begin_log("Discrete variables:")
            while self._next_discrete_object:
                var = self._next_discrete_object.pop()
                var.step()
                info("Sending: %s"%var.name)

                # Call backend to update backend specific things
                self.diffreact_backend.update_variables([var])
                
                var.send()
            end_log()
        
        # Get time after possible 
        time = self.time()
        dt -= offset
        info(wrap_log_message("run:%3d, after step: %3d, time: %6.2f "\
                              "(%.2f) ms, dt taken: %5.2f (%.2f) ms, "\
                              "real time: %s"%\
                              (self.runtime_params.run_num+1,\
                               self.runtime_params.step, \
                               time, self._tstop.value, dt, \
                               self._dt_max.value, self.formated_real_time())))
        flush_logger()
        
        # Update time before handling io
        self.diffreact_backend.runtime_params.time = time
        self.diffreact_backend.runtime_params.dt = dt
        
        # Do backend specific io
        self.diffreact_backend.handle_io()

    def step_back_continuous_variables(self, offset):
        """
        Update the continuous variables by stepping the time back
        with a given offset
        """
        assert(isinstance(offset, float))
        assert(offset>0)
        for cont in self.continuous_variables:
            cont.step_back(offset)
        
        # Call backend to update backend specific things
        self.diffreact_backend.update_variables(self.continuous_variables)

    def update_continuous_variables(self, dt, time):
        """
        Update the continuous variables by steping it with time dt
        """
        
        # Update the values
        for cont in self.continuous_variables:
            cont.update(dt, time)

        # Call backend to update backend specific things
        self.diffreact_backend.update_variables(self.continuous_variables)

        if self.continuous_variables:
            begin_log("Continuous variables:")
            
        # Send the updated values 
        for cont in self.continuous_variables:
            info("Sending: %s"%cont.name)
            cont.send()

        if self.continuous_variables:
            end_log()

    def formated_real_time(self):
        """
        Return the time from start of simulation
        """
        return format_time(time.time() - self.runtime_params.time_at_start)
        
